#include <stdio.h>
#include <assert.h>
#include "doublink.h"

static doublink_node_t*
last_from_head(doublink_head_t* head) {
  return head->last;
}

static doublink_node_t*
first_from_last(doublink_node_t* last) {
  return last ? last->next : NULL;
}

static doublink_node_t*
node_from_head(doublink_head_t* head) {
  return (doublink_node_t*) head;
}

static void
singleton(doublink_head_t* head,
          doublink_node_t* node) {
  head->last = node;
  node->prev = node_from_head(head);
  node->next = node;
}

static int
isfirst(doublink_node_t* node) {
  return (node->prev->next->next == node);
}

static int
islast(doublink_node_t* node) {
  return (node->next->prev->next == node);
}

void
doublink_init(doublink_head_t* head) {
  head->last = NULL;
}

doublink_node_t*
doublink_get_first(doublink_head_t* head) {
  return first_from_last(last_from_head(head));
}

void
doublink_push_first(doublink_head_t* head, doublink_node_t* node) {
  doublink_node_t* last = last_from_head(head);

  if(!last) {
    singleton(head, node);
  }
  else {
    doublink_node_t* new_first = node;
    doublink_node_t* old_first = last->next;

    /* update last: */
    last->next = new_first;

    /* update old_first */
    old_first->prev = new_first;

    /* update new_first: */
    new_first->prev = node_from_head(head);
    new_first->next = old_first;
  }
}

void
doublink_push_last(doublink_head_t* head, doublink_node_t* node) {
  doublink_node_t* last = last_from_head(head);

  if(!last) {
    singleton(head, node);
  }
  else {
    doublink_node_t* first = last->next;
    doublink_node_t* new_last = node;
    doublink_node_t* old_last = last;

    /* update new_last: */
    new_last->next = first;
    new_last->prev = old_last;
    
    /* update old_last: */
    old_last->next = new_last;
    
    /* update head: */
    head->last = new_last;    
  }  
}

void
doublink_remove(doublink_node_t* node) {
  int is_first = isfirst(node);
  int is_last  = islast(node);
  
  if(!is_first && !is_last) {
    /* Case: remove internal node. */
    node->prev->next = node->next;
    node->next->prev = node->prev;
  }
  else {
    if(is_first && is_last) {
      /* Case: remove sole node. */
      node->prev->next = NULL;
    }
    else if(is_first) {
      /* Case: remove first node. */
      node->next->prev = node->prev;
      node->prev->next->next = node->next;
    }
    else if(is_last) {
      /* Case: remove last node. */
      node->prev->next = node->next;
      node->next->prev->next = node->prev;
    }
  }
}

void
doublink_replace(doublink_node_t* node0, doublink_node_t* node1) {
  int is_first = isfirst(node0);
  int is_last  = islast(node0);

  /* node1 has same prev as node0, the next field depends on the cases below. */
  node1->prev = node0->prev;
  node1->next = NULL;

  /* Make incoming pointers to node0 now refer to node1, including the
     next field of node1 (if applicable).  We do this by case: */
  if(!is_first && !is_last) {
    /* Case: replace internal node */
    node1->next = node0->next;
    node1->prev->next = node1;
    node1->next->prev = node1;
  }
  else {
    if(is_first && is_last) {
      /* Case: replace sole node */
      node1->prev->next = node1;
      node1->next = node1;
    }
    else if(is_first) {
      /* Case: replace first node */
      node1->next = node0->next;
      node1->next->prev = node1;
      node1->prev->next->next = node1;
    }
    else if(is_last) {
      /* Case: replace last node */
      node1->next = node0->next;
      node1->prev->next = node1;
      node1->next->prev->next = node1;
    }    
  }

  node0->prev = NULL;
  node0->next = NULL;
}


doublink_node_t*
doublink_seq_remove(doublink_head_t* head, doublink_node_t** _last) {
  doublink_node_t* last = last_from_head(head);
  doublink_node_t* first = first_from_last(last);

  if(last) {
    /* Have: first != NULL. */
    /* Update first and last: */
    first->prev = NULL;
    last->next = NULL;    
  }

  /* Update head: */
  head->last = NULL;
  
  if(_last) {*_last = last;}  
  return first;
}

void
doublink_seq_append(doublink_head_t* head,
                    doublink_node_t* seq_first,
                    doublink_node_t* seq_last)
{
  doublink_node_t* old_last = last_from_head(head);
 
  if(old_last) {
    /* Link: seq_last to old_first */
    seq_last->next = first_from_last(old_last);

    /* Link: old_last with seq_first */
    old_last->next = seq_first;
    seq_first->prev = old_last;
  }
  else {
    /* Link: make sequence circular. */
    seq_first->prev = node_from_head(head);
    seq_last->next = seq_first;
  }

  head->last = seq_last;
}

void
doublink_verify(doublink_head_t* head) {
  doublink_node_t* last = head->last;
  
  if(last) {
    doublink_node_t* first = last->next;
    doublink_node_t* fing1 = first;
    doublink_node_t* fing2 = last;

    /* Reach head from first */
    assert(first->prev == node_from_head(head));

    while(fing1 != last &&
          fing2 != first) {
      fing1 = fing1->next;
      fing2 = fing2->prev;
    }

    /* The forward (next path) is the same length as the backward
       (prev path) */
    assert(fing1 == last &&
           fing2 == first);
  }
}
