#ifndef __DOUBLINK_H__
#define __DOUBLINK_H__

#include "adt_directives.h"

ABSTYPE(doublink_head_s, doublink_head_t); // 1 Word.
ABSTYPE(doublink_node_s, doublink_node_t); // 2 Words.


/* Initialize the list as empty. */
void
doublink_init(doublink_head_t* head);

/* Verifies some properties about the doublink list, does abort() if
   any checked properties fail. */
void
doublink_verify(doublink_head_t* head);


/* Insert a node at the beggining of list, given the list head */

/* Used in:
   -- hashtbl_move
   -- actk_alloc_fresh
   -- actk_alloc_garb
   -- actk_alloc_reuse
*/
void
doublink_push_first(doublink_head_t* head,
                    doublink_node_t* node);

/* Insert a node at the end of a list, given the list head */

/* Used in:
   -- hashtbl_insert
   -- hashtbl_move
*/
void
doublink_push_last(doublink_head_t* head,
                   doublink_node_t* node);

/* Gives back NULL or the first node.  one can iterate through the
   nodes by using next pointers until one reaches first again (next
   pointers are circular). */
doublink_node_t*
doublink_get_first(doublink_head_t* head);

/* Unlink a single node from it's current sequence */

/* Used in:
   -- hashtbl_remove
   -- actk_alloc_garb
   -- actk_alloc_reuse
*/
void
doublink_remove(doublink_node_t* node);


void
doublink_replace(doublink_node_t* node0, doublink_node_t* node1);

/* Given the first and last node (which may be the same node) of a
   sequence, insert the sequence as a prefix of the list with the
   given head */

/* Used in:
   -- actk_alloc_garball
*/
void
doublink_seq_append(doublink_head_t* head,
                    doublink_node_t* node_first,
                    doublink_node_t* node_last);

/* Given the head of a list remove and get the sequence (first and
   last node).  If the first node is non-NULL then it will be
   connected through next pointers until the last node, which will
   have a NULL next pointer.  first and last may be equal for
   singleton lists.  The first node is the return value. */

/* Used in:
   -- actk_garb_all
   -- actk_free_all
   -- hashtbl_move
*/
doublink_node_t*
doublink_seq_remove(doublink_head_t* head,
                    doublink_node_t** last);
  

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Implementation */

/* General doublink list topology: */
/*
     .------.     .------------------ - - ------------------------.
    \|/     |    \|/                                              |
     `      |     `                                               |
  +----+    |   +----+     +----+              +----+     +----+  |
  |last|-.  |   |next|---> |next|--> . . .---> |next|---> |next|--`
  +----+ |  |   |----|     |----|              |----|     |----| 
   head  |  `---|prev| <---|prev| <--. . . <---|prev| <---|prev| <-.
         |      +----+     +----+              +----+     +----+   |
         |       first      second             2nd-last    last    |
         |                                                         |
         `---------------------------------------------------------`
*/

/* doublink list with only a single node */
/*
     .------.     .-----.
    \|/     |    \|/    |
     `      |     `     |
  +----+    |   +----+  |
  |last|-.  |   |next|--`
  +----+ |  |   |----|   
   head  |  `---|prev| <--.
         |      +----+    |
         |    first/last  |
         |                |
         `----------------`
*/
         
struct doublink_head_s {
  doublink_node_t* last;
};

struct doublink_node_s {
  doublink_node_t* next;
  doublink_node_t* prev;
};

#endif
