/* Matthew A Hammer   */
/* <hammer@tti-c.org> */

/* Get the header if we don't already have it. */
#ifndef __DOUBLIST_H__
#define __DOUBLIST_C__
#include "doublist.h"
#undef __DOUBLIST_C__
#endif

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

DOUBLIST_FUN void DOUBLIST(_init)(DOUBLIST(_t)* l) {
#if DOUBLIST_HEAPHEAD
  l->headptr.head = NULL;
#else
  l->node.next = NULL;
  l->node.prev = NULL;
#endif
}

DOUBLIST_FUN DOUBLIST(_node_t*) DOUBLIST(_peek)(DOUBLIST(_t)* l) {
#if DOUBLIST_HEAPHEAD
  return l->headptr.head;
#else    
  if(l->node.prev == NULL)
    return l->node.next;
  else
    return &(l->node);
#endif
}

DOUBLIST_FUN int DOUBLIST(_isempty)(DOUBLIST(_t)* l) {
  return DOUBLIST(_peek)(l) == NULL;
}

DOUBLIST_FUN void
DOUBLIST(_pushhandle)(DOUBLIST(_t)* l, DOUBLIST(_handle_t) h) {
  DOUBLIST(_node_t)* node = h;

#if DOUBLIST_HEAPHEAD  
  DOUBLIST(_node_t)* head = (DOUBLIST(_node_t)*) &(l->headptr.head);
  DOUBLIST(_node_t)* next = l->headptr.head;
  l->headptr.head = node;
#else
  DOUBLIST(_node_t)* head = &(l->node);
  DOUBLIST(_node_t)* next = l->node.next;
  l->node.next = node;
#endif

  if(next)
    next->prev = node;

  node->next = next;
  node->prev = head;
}

DOUBLIST_FUN
DOUBLIST(_handle_t) DOUBLIST(_push)(DOUBLIST(_t)* l, DOUBLIST(_val_t) val) {
  DOUBLIST(_handle_t) h;

#if !DOUBLIST_HEAPHEAD
  if(l->node.prev == NULL) {
    /* SPECIAL CASE: l points to a node, not a headptr, and the node
       isn't being used -- if this holds, then we can fill this node
       without allocating a new node */ 
    DOUBLIST(_node_t)* head = &(l->node);
    head->prev = head;
    head->val = val;
    h = head;
    return h;
  }
#endif
  DOUBLIST(_node_t)* node = DOUBLIST_MALLOC(sizeof(DOUBLIST(_node_t)));
  node->val = val;
  
  /* the handle is just the new node (handles are just abstractions
     for list-nodes). */
  h = node;
  
  DOUBLIST(_pushhandle(l, h));
  return h;
}

DOUBLIST_FUN
DOUBLIST(_handle_t) DOUBLIST(_find)(DOUBLIST(_t)* l, DOUBLIST(_val_t) val) {
  DOUBLIST(_handle_t) h = NULL;
  DOUBLIST(_node_t)*  n = DOUBLIST(_peek)(l);
  while(n) {
    if(n->val == val) {h = n; return h;}
    else n = n->next;
  }
  return h;
}

DOUBLIST_FUN
DOUBLIST(_val_t) DOUBLIST(_remove)(DOUBLIST(_handle_t) h) {
  DOUBLIST(_node_t)* node = h;
  DOUBLIST(_val_t)    val = node->val;
  if(node->prev == node) {
    /* SPECIAL CASE: handle is for first node, which cannot be freed
       since there's no headptr (first-node is acting as the list's
       head).  So put this node into the "empty-head-node" state. */
    node->prev = NULL;
  }
  else {
    DOUBLIST(_node_t)* prev = node->prev;
    DOUBLIST(_node_t)* next = node->next;
    if(prev) prev->next = next;
    if(next) next->prev = prev;

    DOUBLIST_FREE(sizeof(DOUBLIST(_node_t)), node);
  }
  return val;
}

DOUBLIST_FUN DOUBLIST(_val_t) DOUBLIST(_pop)(DOUBLIST(_t)* l) {
  DOUBLIST(_handle_t) h = DOUBLIST(_peek)(l);
  DOUBLIST(_val_t)    v = h->val;
  DOUBLIST(_remove)(h);
  return v;
}

DOUBLIST_FUN DOUBLIST(_val_t) DOUBLIST(_val)(DOUBLIST(_handle_t) h) {
  DOUBLIST(_node_t)* node = h;
  return node->val;
}

/* --- Iteration --- */

DOUBLIST_FUN void DOUBLIST(_iter)(DOUBLIST(_t)* l, DOUBLIST(_iter_t)* i) {
  i->node = DOUBLIST(_peek)(l);
}

DOUBLIST_FUN void DOUBLIST(_next)(DOUBLIST(_iter_t)* i) {
  assert(i->node);
  i->node = i->node->next;
}

DOUBLIST_FUN int DOUBLIST(_isend)(DOUBLIST(_iter_t)* i) {
  return !(i->node);
}

DOUBLIST(_handle_t) DOUBLIST(_currenthandle)(DOUBLIST(_iter_t)* i) {
  assert(i->node);
  return i->node;
}

DOUBLIST(_val_t) DOUBLIST(_currentval)(DOUBLIST(_iter_t)* i) {
  assert(i->node);
  return i->node->val;
}
