/* Matthew A Hammer   */
/* <hammer@tti-c.org> */

#define __DOUBLIST_H__
#include "adt_directives.h"

/* The user defines the following before including this header:
 *
 * -- DOUBLIST(suffix): as prefix##suffix for some user prefix.
 * -- DOUBLIST_MALLOC(size): as the malloc function to use
 * -- DOUBLIST_FREE(size, ptr): as the free function to use
 * -- DOUBLIST_VAL_TYPEDEF: to be the type of the elements in the doublist.
 * -- DOUBLIST_HEAPHEAD: to be 1 if the head of the list should be
 *    heap-allocated, 0 otherwise.
 */

/* Setup default symbol-gen macro */
#ifndef DOUBLIST
#define DOUBLIST(suffix) doublist##suffix
#endif

/* Setup default element type (void*): */
#ifndef DOUBLIST_VAL_TYPEDEF
#define DOUBLIST_VAL_TYPEDEF void*
#endif
typedef DOUBLIST_VAL_TYPEDEF DOUBLIST(_val_t);

/* Abstract-types */
ABSTYPE(DOUBLIST(_s),         DOUBLIST(_t));
ABSTYPE(DOUBLIST(_node_s),    DOUBLIST(_node_t));
ABSTYPE(DOUBLIST(_headptr_s), DOUBLIST(_headptr_t));
ABSTYPE(DOUBLIST(_iter_s),    DOUBLIST(_iter_t));

/* Handles are just node-pointers: */
typedef DOUBLIST(_node_t)* DOUBLIST(_handle_t);

#define DOUBLIST_NULL_HANDLE NULL

/*
  DOUBLIST -- doubly-linked lists.

  =-=-=-=-=-=-=-=-=-=
  Micro-optimization:

  A doublist can be instantiated two ways -- the key difference is how
  much space is required for an empty list and whether the first
  inserted element will require an allocation of a list-node. (i) and
  (ii) summarize the choices:

  (i) empty-list is just a sizeof(void*)-byte head pointer -- good for
  lists which may often be empty, (e.g., lists hanging off slots of a
  hash-table).
  
  (ii) empty-list is a sizeof(struct doublist_node_s)-byte head node
  -- good for lists which may often contain a single element, (e.g.,
  lists holding the readers of a modref), since no allocation is
  necessary to insert the first element into the empty list.

  The directive DOUBLIST_HEAPHEAD controls this.  When it's nonzero
  option (i) is used, otherwise (ii) is used.
*/

/* -- Basic API -- */
DOUBLIST_FUN void                DOUBLIST(_init)(DOUBLIST(_t)* l);
DOUBLIST_FUN DOUBLIST(_val_t)    DOUBLIST(_val)(DOUBLIST(_handle_t) h);
DOUBLIST_FUN DOUBLIST(_handle_t) DOUBLIST(_push)(DOUBLIST(_t)* l, DOUBLIST(_val_t) val);
DOUBLIST_FUN DOUBLIST(_handle_t) DOUBLIST(_find)(DOUBLIST(_t)* l, DOUBLIST(_val_t) val);
DOUBLIST_FUN DOUBLIST(_handle_t) DOUBLIST(_peek)(DOUBLIST(_t)* l);
DOUBLIST_FUN DOUBLIST(_val_t)    DOUBLIST(_remove)(DOUBLIST(_handle_t) h);
DOUBLIST_FUN int                 DOUBLIST(_isempty)(DOUBLIST(_t)* l);
DOUBLIST_FUN DOUBLIST(_val_t)    DOUBLIST(_pop)(DOUBLIST(_t)* l);


/* --- Iteration API --- */
DOUBLIST_FUN void                DOUBLIST(_iter)(DOUBLIST(_t)* l, DOUBLIST(_iter_t)* i);
DOUBLIST_FUN int                 DOUBLIST(_isend)(DOUBLIST(_iter_t)* i);
DOUBLIST_FUN DOUBLIST(_handle_t) DOUBLIST(_currenthandle)(DOUBLIST(_iter_t)* i);
DOUBLIST_FUN DOUBLIST(_val_t)    DOUBLIST(_currentval)(DOUBLIST(_iter_t)* i);
DOUBLIST_FUN void                DOUBLIST(_next)(DOUBLIST(_iter_t)* i);

/* -- "Special" API -- (Used by hashtble impl) */
DOUBLIST_FUN void                DOUBLIST(_pushhandle)(DOUBLIST(_t)* l, DOUBLIST(_handle_t) h);

#include "doublist_impl.h"
