/* Matthew Hammer <hammer@tti-c.org> */
/* Tue Nov  7 13:14:24 CST 2006 */

#ifndef __HASH_H__
#define __HASH_H__

#include <stdlib.h>
#include <stdint.h>

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
from: http://www.burtleburtle.net/bob/hash/doobs.html
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bits set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() was built out of 36 single-cycle latency instructions in a 
  structure that could supported 2x parallelism, like so:
      a -= b; 
      a -= c; x = (c>>13);
      b -= c; a ^= x;
      b -= a; x = (a<<8);
      c -= a; b ^= x;
      c -= b; x = (b>>13);
      ...
  Unfortunately, superscalar Pentiums and Sparcs can't take advantage 
  of that parallelism.  They've also turned some of those single-cycle
  latency instructions into multi-cycle latency instructions.  Still,
  this is the fastest good hash I could find.  There were about 2^^68
  to choose from.  I only looked at a billion or so.
--------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

typedef unsigned int hash_val_t;

typedef struct actk_hash_s {
  hash_val_t a;
  hash_val_t b;
  hash_val_t m;
} actk_hash_t;

static actk_hash_t*
actk_hash_init(actk_hash_t* h, int m) {
  h->a = rand();
  h->b = rand();
  h->m = m;
  return h;
}

static int
actk_hash_ptr(void* dummy /* actk_state_t* */,
              actk_hash_t* h, void* p) {
  hash_val_t a = h->a;
  hash_val_t b = h->b;
  hash_val_t c = (hash_val_t) p;
  mix(a,b,c);
  return c % h->m;

  /*
    OLD METHOD: This old method was taken from the SML library; its
    based on arithmetic mod p and emperically wasn't working very well
    for pointers (perhaps because they're always 4-byte aligned and
    the lower bits weren't getting mixed up enough?)  I've replaced
    the method with an explicit bit-mixing funtion.
    
    const int prime = 1073741789;
    uint32 x = (h->a * ((uint32)(p)) + h->b);
    uint32 x_mod_p = x % prime;
    return (x_mod_p % h->m);
  */
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Jenkin's one-at-a-time hash, for hashing that needn't be seeded by
   This was taken from Wikipedia, Feb 26 2007.*/
static hash_val_t
hash_joaat(hash_val_t start, unsigned char* key, int len) {
  hash_val_t hash = start;
  int i;
  for(i = 0; i < len; i++) {
    hash += key[i];
    hash += (hash << 10);
    hash ^= (hash >> 6);    
  }
  hash += (hash << 3);
  hash ^= (hash >> 11);
  hash += (hash << 15);
  return hash;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*
  Paul Hsieh's hash function, 32-bit only.  This operates a word at a
  time, and is often reported to be faster than Jenkin's function.
  (from: http://idefix.uchicago.edu/~bhudson/src/hash-analysis/hash.cpp)
*/
inline static hash_val_t
hash_hsieh_1step(hash_val_t hash, void* word_) {
  hash_val_t word = (hash_val_t) word_;
  hash_val_t hi = word >> 16;
  hash_val_t lo = word & ( (1<<16) - 1);
  hash_val_t tmp;
  
  hash += hi;
  tmp   = (lo << 11) ^ hash;
  hash  = (hash << 16) ^ tmp;
  hash += hash >> 11;
  return hash;
}

inline static hash_val_t
hash_hsieh_avalanche(hash_val_t hash) {
  
  /* Force "avalanching" of final 127 bits */
  hash ^= hash << 3;
  hash += hash >> 5;
  hash ^= hash << 4;
  hash += hash >> 17;
  hash ^= hash << 25;
  hash += hash >> 6;

  return hash;
}

static hash_val_t
hash_hsieh_array(hash_val_t hash0, int keyc, void** keyv) {
  hash_val_t hash = hash0;
  int i;  
  for(i = 0; i < keyc; i++)
    hash = hash_hsieh_1step(hash, keyv[i]);  
  return hash_hsieh_avalanche(hash);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Bob Jenkins -- From:
 * http://burtleburtle.net/bob/c/lookup3.c
 * Original name: hashword().
 */

/*
  This works on all machines.  To be useful, it requires
  -- that the key be an array of uint32_t's, and
  -- that the length be the number of uint32_t's in the key
  
  The function hashword() is identical to hashlittle() on little-endian
  machines, and identical to hashbig() on big-endian machines,
  except that the length has to be measured in uint32_ts rather than in
  bytes.  hashlittle() is more complicated than hashword() only because
  hashlittle() has to dance around fitting the key bytes into registers.
*/
#define rot(x,k) (((x)<<(k)) | ((x)>>(32-(k))))
#define final(a,b,c) \
{ \
  c ^= b; c -= rot(b,14); \
  a ^= c; a -= rot(c,11); \
  b ^= a; b -= rot(a,25); \
  c ^= b; c -= rot(b,16); \
  a ^= c; a -= rot(c,4);  \
  b ^= a; b -= rot(a,14); \
  c ^= b; c -= rot(b,24); \
}
static
uint32_t hash_jenkins_array(uint32_t initval,
                            size_t length,
                            const uint32_t *k)
{
  uint32_t a,b,c;

  /* Set up the internal state */
  a = b = c = 0xdeadbeef + (((uint32_t)length)<<2) + initval;

  /* handle most of the key */
  while (length > 3)
  {
    a += k[0];
    b += k[1];
    c += k[2];
    mix(a,b,c);
    length -= 3;
    k += 3;
  }

  /* handle the last 3 uint32_t's */
  /* (all the case statements fall through) */
  switch(length)                     
  { 
  case 3 : c+=k[2];
  case 2 : b+=k[1];
  case 1 : a+=k[0];
    final(a,b,c);
  case 0:
    /* nothing left to add */
    break;
  }
  return c;
}
#endif
