/* Matthew Hammer <hammer@tti-c.org> */

/* inline-funtions (macros, sort of) to stuff flag bits in and out of
   the LSBs of pointers.  The point of this little module is to try to
   keep all the dirty bit-stuffing stuff somewhat abstract so that as
   we try to compile for different machines or different environments
   (with different alignments) we can keep all the dirty stuff in
   (somewhat) higher-level terms: a pointer with flags is to be cast
   and bit-juggled only though this "API".
   
   -- ptr_get(p) -- give the pointer-part of p (with any flags masked out)

   -- ptr_set(p) -- set the pointer-part of p (with flags left unchanged)

   -- ptr_getflags(p) -- give the flags stored in p (all higher pointer bits masked out)

   -- ptr_setflags(p, bits) -- set the flag bits of the pointer,

     -- To toggle a single bit on:
             ptr_setflags(p, ptr_getflags(p) | bit)

     -- To toggle a single bit off:
             ptr_setflags(p, ptr_getflags(p) & ~bit)
       (or)  ptr_setflags(p, ptr_getflags(p) - bit)      
*/

#ifndef __POINTER_H__
#define __POINTER_H__

#include "adt_directives.h"
#include <assert.h>

typedef void* adt_ptr_t;
typedef unsigned int adt_ptrbits_t;

PTR_FUN adt_ptr_t ptr_get(adt_ptr_t p);
PTR_FUN void ptr_set(adt_ptr_t* p, adt_ptr_t p_);

PTR_FUN adt_ptrbits_t ptr_getflags(adt_ptr_t p);
PTR_FUN void ptr_setflags(adt_ptr_t* p, adt_ptrbits_t bits);

PTR_FUN adt_ptr_t ptr_addflags(adt_ptr_t p, adt_ptrbits_t bits);

/* --- IMPLEMENTATION -- */  

#define ADT_PTR_MASK_PTR   ((adt_ptrbits_t)0x3)
#define ADT_PTR_MASK_FLAGS (~ADT_PTR_MASK_PTR)

PTR_FUN
adt_ptr_t ptr_get(adt_ptr_t p) {
  return (adt_ptr_t) (((adt_ptrbits_t) p) & ADT_PTR_MASK_FLAGS);
}

PTR_FUN
adt_ptrbits_t ptr_getflags(adt_ptr_t p) {
  return ((adt_ptrbits_t) p) & ADT_PTR_MASK_PTR;
}


PTR_FUN
adt_ptr_t ptr_addflags(adt_ptr_t p, adt_ptrbits_t bits) {
  return (adt_ptr_t) (((adt_ptrbits_t) p & ADT_PTR_MASK_FLAGS) | bits);
}

PTR_FUN
void ptr_setflags(adt_ptr_t* p, adt_ptrbits_t bits) {
  assert(bits <= ADT_PTR_MASK_PTR);
  *p = ptr_addflags(*p, bits);
}

PTR_FUN
void ptr_set(adt_ptr_t* p, adt_ptr_t p_) {
  adt_ptrbits_t flags = ptr_getflags(*p);
  *p = p_;
  ptr_setflags(p, flags);
}

#endif
