/* Matthew A Hammer <hammer@tti-c.org> */

/* Get the header if we don't already have it. */
#ifndef __PQUEUE_H__
#define __PQUEUE_C__
#include "pqueue.h"
#undef __PQUEUE_C__
#endif

#include <string.h>
#include <stdlib.h>

PQUEUE_FUN void PQUEUE(__grow)(PQUEUE(_t)* pq);
PQUEUE_FUN void PQUEUE(__bubble_up)(PQUEUE(_t)* pq, int pos);
PQUEUE_FUN void PQUEUE(__bubble_down)(PQUEUE(_t)* pq, int pos);

#define UP(pos) ((pos - 1) >> 1)
#define DOWN(pos) ((pos << 1) + 1)

/* Setup COMP as the comparision function for elements */
#ifndef PQUEUE_ELT_COMP
#define COMP(pq, a, b) ((int)(a) - (int)(b))
#else
#define COMP(pq, a, b) ((PQUEUE_ELT_COMP)(a, b))
#endif

/* Setup SETPOS as the "storing" function for elements -- it keeps the
   elements' position in the heap-array consistent with the position
   maintained by the element itself. */
#ifndef PQUEUE_ELT_SETPOS
#define SETPOS(pq, a, pos) ((pq)->data[pos] = a)
#else
#define SETPOS(pq, a, pos)                                \
  (__extension__({                                        \
     PQUEUE(_pos_t) __pos = pos;                          \
     PQUEUE(_elt_t) __b = (PQUEUE_ELT_SETPOS)(a, __pos);  \
     if (__pos >= 0) (pq)->data[__pos] = __b;             \
  }))
#endif

/* SWAP is a useful macro for swaping elements' positions (via SETPOS) */
#define SWAP(pq, pos1, pos2)                      \
  (__extension__({                                \
    PQUEUE(_pos_t) __pos1 = pos1;                 \
    PQUEUE(_pos_t) __pos2 = pos2;                 \
    PQUEUE(_elt_t) __elt1 = (pq)->data[__pos1];   \
    PQUEUE(_elt_t) __elt2 = (pq)->data[__pos2];   \
    SETPOS(pq, __elt1, __pos2);                   \
    SETPOS(pq, __elt2, __pos1);                   \
  }))
   
     
PQUEUE_FUN
PQUEUE(_t)* PQUEUE(_init)(PQUEUE(_t)* pq) {
  pq->size = 10;
  pq->num = 0;
  pq->data = (PQUEUE(_elt_t)*) PQUEUE_MALLOC(pq->size * sizeof(PQUEUE(_elt_t)));
  return pq;
}

PQUEUE_FUN
int PQUEUE(_isempty)(const PQUEUE(_t)* pq) {
  return pq->num == 0;
}

PQUEUE_FUN
void PQUEUE(__bubble_up)(PQUEUE(_t)* pq, int pos) {
  while(pos > 0) {
    int up = UP(pos);
    if (COMP(pq, pq->data[pos], pq->data[up]) < 0) {
      SWAP(pq, pos, up);
      pos = up;
    }
    else
      return;
  }
}

PQUEUE_FUN
int PQUEUE(_push)(PQUEUE(_t)* pq, PQUEUE(_elt_t) elt) {
  if(pq->num == pq->size)
    PQUEUE(__grow)(pq);
  SETPOS(pq, elt, pq->num ++);
  PQUEUE(__bubble_up)(pq, pq->num - 1);
  return 0;
}

PQUEUE_FUN
PQUEUE(_elt_t) PQUEUE(_peek)(const PQUEUE(_t)* pq) {
  return pq->data[0];
}


PQUEUE_FUN
void PQUEUE(__bubble_down)(PQUEUE(_t)* pq, int pos) {
  int down;

  while((down = DOWN(pos)) < pq->num) {
    int tmp = down + 1;
    
    if ((tmp < pq->num) &&
        COMP(pq, pq->data[tmp], pq->data[down]) < 0) {
      down = tmp;
    }
    
    if (COMP(pq, pq->data[pos], pq->data[down]) > 0) {
      SWAP(pq, pos, down);
      pos = down;
    }
    else
      break;
  }
}

PQUEUE_FUN
PQUEUE(_elt_t) PQUEUE(_pop)(PQUEUE(_t)* pq) {
  PQUEUE(_elt_t) elt_out = pq->data[0];
  SWAP(pq, 0, -- pq->num);
  PQUEUE(__bubble_down)(pq, 0);

  /* -2 symbolizes: popped-from-queue */
  SETPOS(pq, elt_out, -2); 
  return elt_out;
}

PQUEUE_FUN
PQUEUE(_elt_t) PQUEUE(_remove)(PQUEUE(_t)* pq, PQUEUE(_pos_t) pos) {
  PQUEUE(_elt_t) elt_out = pq->data[pos];
  SWAP(pq, pos, -- pq->num);

  /* At most one of these will do anything effectful: */
  PQUEUE(__bubble_down)(pq, pos);
  PQUEUE(__bubble_up)(pq, pos);

  /* -1 symbolizes: removed-from-queue */
  SETPOS(pq, elt_out, -1);
  return elt_out;
}

PQUEUE_FUN
void PQUEUE(_cleanup)(PQUEUE(_t*) pq) {
  PQUEUE_FREE(pq->data);
  pq->data = NULL;
}

PQUEUE_FUN
void PQUEUE(__grow)(PQUEUE(_t*) pq) {
  PQUEUE(_elt_t)* tmp;
  pq->size *= 2;
  tmp = PQUEUE_MALLOC(pq->size * sizeof(void*));
  memcpy(tmp, pq->data, (sizeof(void*)) * pq->num);
  PQUEUE_FREE(pq->data);
  pq->data = tmp;
}

/* Move these out of scope: */
#undef UP
#undef DOWN
#undef SETPOS
#undef COMP
#undef SWAP
