/* Matthew A Hammer <hammer@tti-c.org> */

#define __PQUEUE_H__
#include "adt_directives.h"

/*
 * Priority Queue - a priority queue implemention based on
 * heap-arrays.  The user provides memory for the structure, but the
 * queue's array is malloc-allocated.  Automatically grows by doubling
 * when capacity is reached.  The cleanup function will free this
 * memory.
 *
 * Has standard ops:
 *   push, peek, pop, isempty.
 *
 */

/* The user defines the following before including this header:
 *
 * -- PQUEUE(suffix): as prefix##suffix for some user prefix.
 *
 * -- PQUEUE_ELT_TYPEDEF: to be the type of the elements in the queue.
 *
 * -- PQUEUE_ELT_COMP: some funtion of type PQUEUE(comp_fun_t), this
 * funtion will be called statically (instead of through a funtion
 * pointer, etc).
 *
 * -- PQUEUE_ELT_SETPOS: some function of type PQUEUE(setpos_fun_t),
 * this function will be called each time the element changes position
 * in the queue to store the new position within the element (the
 * element stored should be returned by the this function, so it's
 * allowed to change value as it changes position).  In order to
 * delete an element from the queue when the element isn't at the
 * head, this position is needed (there is no find() operation).
 *
 * 
 */

/* Setup default symbol-gen macro */
#ifndef PQUEUE
#define PQUEUE(suffix) pqueue##suffix
#endif

/* Setup default element type (void*): */
#ifndef PQUEUE_ELT_TYPEDEF
#define PQUEUE_ELT_TYPEDEF
typedef void* PQUEUE(_elt_t);
#else
typedef PQUEUE_ELT_TYPEDEF PQUEUE(_elt_t);
#endif

typedef int            PQUEUE(_pos_t);
typedef int            (*PQUEUE(_comp_fun_t))   (PQUEUE(_elt_t) p1, PQUEUE(_elt_t) p2);
typedef PQUEUE(_elt_t) (*PQUEUE(_setpos_fun_t)) (PQUEUE(_elt_t), PQUEUE(_pos_t));

ABSTYPE(PQUEUE(_s), PQUEUE(_t));

PQUEUE_FUN PQUEUE(_t)*    PQUEUE(_init)(PQUEUE(_t)* pq);
PQUEUE_FUN int            PQUEUE(_isempty)(const PQUEUE(_t)* pq);
PQUEUE_FUN int            PQUEUE(_push)(PQUEUE(_t)* pq,  PQUEUE(_elt_t) elt);
PQUEUE_FUN PQUEUE(_elt_t) PQUEUE(_pop)(PQUEUE(_t)* pq);
PQUEUE_FUN PQUEUE(_elt_t) PQUEUE(_remove)(PQUEUE(_t)* pq, PQUEUE(_pos_t) pos);
PQUEUE_FUN PQUEUE(_elt_t) PQUEUE(_peek)(const PQUEUE(_t)* pq);
PQUEUE_FUN void           PQUEUE(_cleanup)(PQUEUE(_t*) pq);

#include "pqueue_impl.h"
