#include <time.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>

#include "actk_types.h"
#include "actk_debug.h"
#include "timestamp.h"
#include "closure_impl.h"
#include "logging.h"
#include "state_impl.h"

typedef struct print_interval_args_s {
  FILE* file;
  int timestampc;
  int closurec;
} print_interval_args_t;

void actk__print_indent(FILE* f, int indent) {
  int j;
  for(j = 0; j < indent; j++)
    fprintf(f, " . ");
}

void
actk__verify_interval_closures(actk_state_t* st,
                               actk_interval_t interval) {
  actk_interval_t i = interval;
  int count_closures_ok = 0;
  int count_closures_null = 0;
  int count_closures_error = 0;
  int count_closures = 0;
  int count_total_errors = 0;
  
  while(i.start && totalorder_compare(i.start, i.end) < 0) {
    actk_closure_t* closure = NULL;
    void* data = (void*) totalorder_getdata(i.start);

    if(ptr_getflags(data) == ACTK_TIMESTAMP_CLOSURE_START) {
      closure = (actk_closure_t*) ptr_get(data);
    }

    if(closure) {
      int errors = actk__closure_verify(st, closure);
      if(!errors)
        count_closures_ok ++;
      else
        count_closures_error ++;
      
      count_total_errors += errors;
      count_closures ++;
    }
    else
      count_closures_null ++;
    
    i.start = totalorder_next(i.start);
  }
  
  logcatf(st,"closures OK        :%10d", count_closures_ok);
  logcatf(st,"closures MALFORMED :%10d", count_closures_error);
  logcatf(st,"closures NULL      :%10d", count_closures_null);
  logcatf(st,"total closures     :%10d", count_closures);
  logcatf(st,"total errors       :%10d", count_total_errors);
}

actk_timestamp_t*
actk__print_interval_rec(actk_state_t* st,
                         print_interval_args_t* args, int indent,
                         actk_interval_t interval,
                         actk_closure_t* prev_closure)
{
  actk_interval_t i = interval;
  
  while(i.start && totalorder_compare(i.start, i.end) < 0) {
    actk_closure_t* closure = NULL;
    void* data = (void*) totalorder_getdata(i.start);

    if(ptr_getflags(data) == ACTK_TIMESTAMP_CLOSURE_START) {
      closure = (actk_closure_t*) ptr_get(data);
    }

    if(closure) args->closurec ++;

    fprintf(args->file, "%03d|%016llx|%10p|",
                      indent, i.start->label, i.start);
    actk__print_indent(args->file, indent);

    fprintf(args->file,  "[" DBSTR_CLOSURE " ends@%p\n",
            DBVAS_CLOSURE(closure),
            closure ? actk__closure_get_end(st, closure) : NULL);

    i.start = totalorder_next(i.start);
    args->timestampc ++;
    
    if(closure) {
      actk_interval_t inner_interval = {i.start, NULL};

      if(!actk__closure_get_end(st, closure) ||
         totalorder_compare(i.end, actk__closure_get_end(st, closure)) < 0)
        inner_interval.end = i.end;
      else
        inner_interval.end = actk__closure_get_end(st, closure);
      
      i.start = actk__print_interval_rec (st, args, indent + 1, inner_interval, closure);
    }
  }

  if(prev_closure) {
    fprintf(args->file, "---|%016llx|%10p|", i.start->label, i.start);
    actk__print_indent(args->file, indent - 1);    
    fprintf(args->file, "%s End:" DBSTR_CLOSURE ") \n",
            (actk__closure_get_end(st, prev_closure) == i.start)
            ? "(ok)" : "(ERROR)",
            DBVAS_CLOSURE(prev_closure));
  }
  return i.start;
}

#define ACTK_INTERVAL_FILE_PREFIX "actk.interval"

void actk__print_interval(actk_state_t* st, actk_interval_t interval, const char* fmt, ...) {
  static int next_num = 0;
  const int MAX = 255;
  char filename[MAX];
  char timestring[MAX];
  print_interval_args_t args;

  /* compute the filename */ {
    va_list ap;
    int offset = 0;
    va_start(ap, fmt);  
    offset += snprintf(&filename[offset],MAX-offset-1, ACTK_INTERVAL_FILE_PREFIX "%03d-", next_num++);
    offset += vsnprintf(&filename[offset],MAX-offset-1, fmt, ap);
    offset += snprintf(&filename[offset],MAX-offset-1, ".out");
    logcatf(st, "writing interval to `%s'", filename);
    va_end(ap);
  }
    
  /* get a string of the current time */ {
    time_t t;
    struct tm *tmp;
    t = time(NULL);
    tmp = localtime(&t);
    strftime(timestring, MAX-1, "%s == %F %X", tmp);
  }
  
  args.file = fopen(filename, "w+");
  args.timestampc = 1;
  args.closurec = 0;
  
  fprintf(args.file, "Interval File   : `%s'\n", filename);
  fprintf(args.file, "Generation Time : %s\n", timestring);
  fprintf(args.file, "---+----------------+----------+-------->\n");
  fprintf(args.file, "%-3s|%-16s|%10s|%s\n", "dp", "timestamp-label", "timestamp", " closure");
  fprintf(args.file, "---+----------------+----------+-------->\n");
  
  actk__print_interval_rec(st, &args, 0, interval, NULL);
  
  fprintf(args.file, "---+----------------+----------+-------->\n");
  fprintf(args.file, "   Prgctx-Count : %d\n", args.closurec);
  fprintf(args.file, "Timestamp-Count : %d\n", args.timestampc);
  fclose(args.file);
}

void actk__remove_interval_files(actk_state_t* st) {
  if(st->flags & ACTK_RM_INTERVALS) 
    system("rm " ACTK_INTERVAL_FILE_PREFIX "*");
}
