/* Matthew Hammer <hammer@tti-c.org>
 *
 * ACTK-Macros : we use some macros for brevity and to help hide
 * implementation details of the API.
 */

#include <assert.h>

/* BEGIN -- local macros ------------------------------------------ */
#define TERM_HIGHLIGHT_1 "\x1b[1;32m"
#define TERM_HIGHLIGHT_2 "\x1b[1;33m"
#define TERM_HIGHLIGHT_E "\x1b[1;30m"
#define TERM_NORM        "\x1b[0m"
#define ACTK_UNUSED       __attribute__((unused))
#define UNUSED           ACTK_UNUSED

/** \defgroup actk_public_fundecl Function Declaration
    \ingroup actk_public
    @{
*/
#define ACTK_FUN_SIG(fun_name, name_string, description_string, ...) \
  extern actk_fun_t*    fun_name;                                    \
  UNUSED static actk_funarg_t  fun_name##__ARGS[] = {__VA_ARGS__};   \
  UNUSED static actk_funsig_t  fun_name##__SIG =                     \
     { name_string,                                                  \
       description_string,                                           \
       sizeof(fun_name##__ARGS) / sizeof(actk_funarg_t),             \
       fun_name##__ARGS,                                             \
       /* don't know the closure and recctx sizes yet.               \
          If the C macro system was more powerful we could compute   \
          it here, but alas, we can't. */                            \
       -1, -1,                                                       \
     };                                                              \

#define ACTK_FUN_ARG(name, type, desc)                               \
  {name, ACTK_FUN_ARG_##desc}
/* @} */

/** \defgroup actk_public_fundef Function Definition
    \ingroup actk_public
    @{ */

#define ACTK_FUN(fun_name, c_function_name, fun_flags)               \
  void c_function_name(actk_state_t* st);                            \
  actk_fun_t     fun_name##__FUN   =                                 \
     { &fun_name##__SIG,                                             \
       c_function_name,                                              \
       fun_flags                                                     \
       IF_DEBUG(,1,1)                                                \
     };                                                              \
  actk_fun_t*   fun_name = &fun_name##__FUN;                         \

/** To be placed at the top of every actk function body's definition,
    where "top" means before any other operations are performed
    (including initialization of locals-- this is necessary to setup
    the argument unpacking code as well as setup a place for
    "tail-recursions" to loop to). */
#define actk_fun_begin(st)                                           \
  {while(st->fun_loop_on) {                                          \
    actk_closure_t*    __closure = st->closure;                      \
    actk_recctx_t*      __recctx = __closure->recctx;                \
    actk_fun_t*            __fun = __recctx->fun;                    \
    actk_funsig_t*      __funsig = __fun->sig;                       \
    actk_funarg_t*      __funarg = __funsig->args;                   \
    actk_val_t*          __arg_p = (actk_val_t*) (__closure + 1);    \
    actk_val_t*    __envir_arg_p =                                   \
                         (actk_val_t*) (__recctx + 1);               \
    int UNUSED __dummy_1[] =                                         \
     {logcatf(st, TERM_HIGHLIGHT_1 "BEGIN: "                         \
                  DBSTR_CLOSURE TERM_NORM,                           \
                  DBVAS_CLOSURE(__closure))};                        \
    int UNUSED __dummy_2[] =                                         \
     {logcatf(st, "  in " DBSTR_STATE, DBVAS_STATE(st)) };           \
    int UNUSED __dummy_3[] = {st->fun_loop_on = 0};                  \

/** Get the next argument from the currently-running closure. */ 
#define actk_fun_arg(st)                                             \
 (__extension__({                                                    \
   actk_val_t     __argval = (actk_val_t) 0xdeadbeef;                \
   actk_funarg_t* __arg = __funarg ++;                               \
   assert(__arg - __funsig->args < __funsig->argc);                  \
   switch(__arg->desc){                                              \
   case ACTK_FUN_ARG_READ_LOC_OR_VAL:                                \
     assert(!__arg_p[0] ||                                           \
            (actk_deref(st, (actk_modref_t*)__arg_p[0])              \
            == __arg_p[1]));                                         \
     __arg_p  ++;                                                    \
     __argval = *__arg_p;                                            \
     __arg_p += 2;                                                   \
     break;                                                          \
   case ACTK_FUN_ARG_LIFTD_DST:                                      \
      /* TEMP: use actk_closure_slot_* API eventually: */            \
     __argval =   __arg_p;                                           \
     __arg_p += 3;                                                   \
     break;                                                          \
   case ACTK_FUN_ARG_CONST_DST:                                      \
   case ACTK_FUN_ARG_CONST_VAL:                                      \
     __argval = *(__arg_p ++);                                       \
     break;                                                          \
   case ACTK_FUN_ARG_ENVIR_VAL:                                      \
     __argval = *(__envir_arg_p ++);                                 \
     break;                                                          \
   };                                                                \
   logcatf(st," %10s = %p / %d", __arg->name, __argval, __argval);   \
   (void*)__argval;}))                                               \

/** To be placed at the end of every actk function body's definition,
    where "end" means that no other statements syntactically come
    afterward. */ 
#define actk_fun_end(st)                                             \
  }}                                                                 \
  logcatf(st, "END: " DBSTR_CLOSURE "", DBVAS_CLOSURE(st->closure)); \
/*                                                                  */
/* @}                                                               */

/*                                                                  */
#define actk_alloc(_st, _size, ...)                                  \
  (__extension__({                                                   \
    actk_state_t*       __st       = (_st);                          \
    unsigned int        __size     = (_size);                        \
    actk_key_t          __keyv[]   = {__VA_ARGS__};                  \
    int                 __keyc     = sizeof(__keyv) /                \
                                     sizeof(actk_key_t);             \
    actk_alloc_new(__st, __size, __keyc, __keyv);                    \
    }))                                                              \
/*                                                                  */
#define actk_modref(_st, ...)                                        \
  (__extension__({                                                   \
    actk_state_t*       __st       = (_st);                          \
    actk_key_t          __keyv[]   = {__VA_ARGS__};                  \
    int                 __keyc     = sizeof(__keyv) /                \
                                     sizeof(actk_val_t);             \
    actk_modref_new(__st, __keyc, __keyv);}))                        \
/*                                                                  */

/* Closures --                                                      */
/*                                                                  */
#define actk_closure(_st, _fun, ...)                                 \
  (__extension__({                                                   \
    actk_state_t*       __st       = (_st);                          \
    actk_fun_t*         __fun      = (_fun);                         \
    unsigned int UNUSED __argindex = 0;                              \
    actk_argvmask_t     __readmask = 0;                              \
    actk_argvmask_t     __rkeymask = 0;                              \
    actk_val_t          __args[]   = {__VA_ARGS__};                  \
    actk_recctx_t*      __recctx   =                                 \
     ((__st->closure == NULL)                                        \
       ? actk__recctx_new(__st, __fun, __rkeymask, __args)           \
       : ((__st->closure->recctx->fun == __fun)                      \
           ? (assert(__rkeymask == 0), __st->closure->recctx)        \
           : actk__recctx_new(__st, __fun, __rkeymask, __args)));    \
    actk__closure_new(__st, __fun, __recctx,                         \
                      __rkeymask, __readmask, __args);}))            \
/*                                                                  */
#define actk_read(arg)                                               \
  (__extension__({                                                   \
    __readmask |= (1 << __argindex++);                               \
    (actk_val_t)(arg);}))                                            \
/*                                                                  */
#define actk_pass(arg)                                               \
  (__extension__({                                                   \
    __argindex++;                                                    \
    (actk_val_t)(arg);}))                                            \
/*                                                                  */
#define actk_rkey(arg)                                               \
  (__extension__({                                                   \
    __rkeymask |= (1 << __argindex++);                               \
    (actk_val_t)(arg);}))                                            \
/*                                                                  */
#define actk_call(st, fun, ...)                                      \
  actk_run(st, actk_closure(st, fun, __VA_ARGS__))                   \
/*                                                                  */
