#include <string.h>
#include "allocation2.h"
#include "allocation2_impl.h"

#include "stats.h"
#include "modref_impl.h"
#include "closure_impl.h"
#include "state_impl.h"
#include "readqueue.h"
#include "logging.h"
#include "actk_debug.h"

#define USE_ALLOCSET 1

/* Use this to track (and abort) when the hashtables have too many
   collisions.  This debugging code dumps files containing hashes and
   keys to help determine the cause of the collisions.  */
#define TRACK_COLLISIONS 1
#define MAX_COLLISIONS_PER_SLOT 200

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* General utility functions to conceal implementation grime from
   subroutines later. */

static alloctbl_t*
current_alloctbl(actk_state_t* st) {  
  return (st->closure->recctx
          ? &st->closure->recctx->alloctbl
          : NULL);
}

static int
head_tag_from_keyc(actk_state_t* st, int keyc) {
  if(!(st->flags & ACTK_BOOKKEEPING)) {
    return 1; /* Just header 1 */
  }
  else if(keyc == 0) {
    return 2; /* Just headers 1 and 2 */
  }
  else if(current_alloctbl(st) == NULL) {
    return 2; /* Just headers 1 and 2 */
  }
  else {
    return 3; /* Headers 1, 2 and 3 */
  }
}

static int
head_size_from_head_tag(int last_hd, int keyc) {
  static const unsigned int sizes[] = {
    sizeof(actk_alloc_hd1_t),
    sizeof(actk_alloc_hd1_t) + sizeof(actk_alloc_hd2_t),
    sizeof(actk_alloc_hd1_t) + sizeof(actk_alloc_hd2_t) + sizeof(actk_alloc_hd3_t),
  };
  assert(last_hd >= 1 && last_hd <= 3);
  return sizes[last_hd - 1] + (last_hd == 3 ? keyc * sizeof(void*) : 0);
}

static hash_val_t
hash_from_keys(int keyc, void** keyv) {
  if(0) {
    /* Jenkin's One-At-A-Time. */
    hash_val_t hash = 0;
    int i;  
    for(i = 0; i < keyc; i++) {
      hash = hash_joaat(hash, (unsigned char*) &keyv[i], sizeof(void*));
    }
    return hash;
  }
  else if(0){
    /* Hsieh hash function. */
    return hash_hsieh_array(0, keyc, keyv);
  }
  else if(1) {
    /* Another Jenkin's hash function */
    return hash_jenkins_array(0, keyc, (const uint32_t*) keyv);
  }
}

static int
keys_are_equal(int keyc, void** keyv1, void** keyv2) {
  int i;
  for(i = 0; i < keyc; i++) {
    if(keyv1[i] != keyv2[i])
      return 0;
  }
  return 1;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Convert and retrieve various pointers, headers and bit-fields given
   other pointers or headers.  (these all essentially boil down to
   bumping pointers around by fixed offests and sometimes doing some
   bit-masking) */

static actk_alloc_hd1_t*
hd1_from_user_ptr(void* ptr) {
  actk_alloc_hd1_t* hd1 = (actk_alloc_hd1_t*) ptr;
  return hd1 - 1;
}

static actk_alloc_hd2_t*
hd2_from_hd1(actk_alloc_hd1_t* hd1) {
  actk_alloc_hd2_t* hd2 = (actk_alloc_hd2_t*) hd1;
  return hd2 - 1;
}

static actk_alloc_hd3_t*
hd3_from_hd2(actk_alloc_hd2_t* hd2) {
  actk_alloc_hd3_t* hd3 = (actk_alloc_hd3_t*) hd2;
  return hd3 - 1;
}

static actk_alloc_hd3_t*
hd3_from_hd1(actk_alloc_hd1_t* hd1) {
  return hd3_from_hd2(hd2_from_hd1(hd1));
}

static void**
keyv_from_hd3(actk_alloc_hd3_t* hd3) {
  return hd3->keyv_end - hd3->keyc;
}

static unsigned int
head_tag_from_hd1(actk_alloc_hd1_t* hd1) {
  return (unsigned int) ptr_getflags((void*)hd1->size);
}

static unsigned int
size_from_hd1(actk_alloc_hd1_t* hd1) {
  return (unsigned int) ptr_get((void*)hd1->size);
}

static unsigned int
head_size_from_hd1(actk_alloc_hd1_t* hd1) {
  unsigned int hd_tag = head_tag_from_hd1(hd1);
  int keyc = 0;
  if(hd_tag == 3) {
    keyc = hd3_from_hd1(hd1)->keyc;
  }
  return head_size_from_head_tag(hd_tag, keyc);
}
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Convenience functions for dealing with owners in Header-(3) */

static void
hd3_set_owner(actk_alloc_hd3_t* hd3,
              actk_alloc_ownertag_t ownertag,
              void* owner) {
  hd3->owner.ptr = owner;
  ptr_setflags(& hd3->owner.ptr, ownertag);
}

static void
hd2_disown_garbage(actk_alloc_hd2_t* hd2) {
  /* Test if there is a Header-(3), if so we assume it contains a
     owning closure, which we switch to a alloctable */
  if(head_tag_from_hd1(hd2->hd1) == 3) {
    actk_alloc_hd3_t* hd3 = hd3_from_hd2(hd2);
    hd3_set_owner(hd3, ACTK_OWNER_TABLE,
                  &hd3->owner.closure->recctx->alloctbl);
  }
}

static void
hd2_adopt_live(actk_alloc_hd2_t* hd2, actk_closure_t* closure1) {
  if(head_tag_from_hd1(hd2->hd1) == 3) {
    actk_alloc_hd3_t* hd3 = hd3_from_hd2(hd2);
    hd3_set_owner(hd3, ACTK_OWNER_CLOSURE,
                  closure1);
  }
}

static void*
owner_from_hd3(actk_alloc_hd3_t* hd3) {
  return ptr_get(hd3->owner.ptr);
}

static actk_alloc_ownertag_t
ownertag_from_hd3(actk_alloc_hd3_t* hd3) {
  return ptr_getflags(hd3->owner.ptr);
}

static actk_closure_t*
owning_closure_from_hd3(actk_alloc_hd3_t* hd3) {
  switch(ownertag_from_hd3(hd3)) {
  case ACTK_OWNER_CLOSURE: return owner_from_hd3(hd3);
  case ACTK_OWNER_TABLE: return NULL;
  default:
    abort();
  }
}

static alloctbl_t*
alloctbl_from_hd3(actk_alloc_hd3_t* hd3) {
  switch(ownertag_from_hd3(hd3)) {
  case ACTK_OWNER_CLOSURE:
    return &((actk_closure_t*) owner_from_hd3(hd3))->recctx->alloctbl;
  case ACTK_OWNER_TABLE:
    return owner_from_hd3(hd3);
  default:
    abort();
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* "Public" Interface */


hash_val_t
actk_alloc_hash(actk_alloc_hd3_t* hd3) {
  return hash_from_keys(hd3->keyc, keyv_from_hd3(hd3));
}

void*
actk_alloc_new(actk_state_t* st, unsigned int req_size, int keyc, void** keyv) {
  if(keyc == 0 ||
     !(st->flags & ACTK_BOOKKEEPING) ||
     !(st->closure->recctx->fun->flags & ACTK_USEALLOCTABLE)) {
    return actk_alloc_fresh(st, req_size, 0, NULL);
  }
  else if(!st->chpctx) {
    return actk_alloc_fresh(st, req_size, keyc, keyv);
  }
  else {
    /* Look-up for keys */
    alloctbl_t*    alloctbl = current_alloctbl(st);
    hash_val_t     hash_val = hash_from_keys(keyc, keyv);
    doublink_node_t*  first = alloctbl_lookup(alloctbl, hash_val);
    doublink_node_t*   node = first;
    actk_alloc_hd3_t* match = NULL;

#if TRACK_COLLISIONS
     const int max = MAX_COLLISIONS_PER_SLOT;
    int i = 0;
    doublink_node_t* nodes[max];
#endif
    
    /* Look for a match in the list of buckets. */
    while(node ) {
      actk_closure_t* owner = NULL;
      actk_alloc_hd3_t* hd3 = (actk_alloc_hd3_t*) node;
      actk_alloc_hd1_t* hd1 = hd3->hd2->hd1;
      unsigned int     size = size_from_hd1(hd1) - head_size_from_hd1(hd1);
      
#if TRACK_COLLISIONS
      if(i >= max) {
        FILE* file_hashes = fopen("hashes.out", "w+");
        FILE* file_keys   = fopen("keys.out", "w+");
        
        int j;
        for(j = 0; j < max; j++) {
          actk_alloc_hd3_t* hd3 = (actk_alloc_hd3_t*) nodes[j];
          int k;

          /* Write the hash */
          fprintf(file_hashes, "%x\n", actk_alloc_hash(hd3));

          /* Write the keys */
          for(k = 0; k < hd3->keyc; k++)
            fprintf(file_keys, "%p ", hd3->keyv_end[hd3->keyc - k]);
          fprintf(file_keys, "\n");

        }
        abort();
      }
      else {
        nodes[i++] = node;
      }  
#endif
      
      if(node->next != first)
        node = node->next;
      else
        node = NULL;

      /* Check: sizes must be equal */
      if(size != req_size) {
        continue;
      }      
      /* Check: keys must be equal */
      if(hd3->keyc != keyc ||
         !keys_are_equal(keyc, keyv_from_hd3(hd3), keyv)) {
        continue;
      }

      /* Check: if there is an owner, it must be in the right time
         interval */
      if( (owner = owning_closure_from_hd3(hd3)) ) {
        if( totalorder_compare(st->now, owner->interval.start) >= 0 ||
            totalorder_compare(st->chpctx->end, owner->interval.start) < 0 ) {
          continue;
        }            
      }
      
      /* Success! */
      match = hd3;
      break;
    }

    /* Either return what we found or make a fresh allocation */
    if(match) {
      actk_alloc_reuse(st, match);
      return match->hd2->hd1->user_ptr;
    }
    else {
      return actk_alloc_fresh(st, req_size,  keyc, keyv);
    }
  }
}

void
actk_alloc_garb(actk_state_t* st, void* ptr) {
  actk_alloc_hd1_t* hd1 = hd1_from_user_ptr(ptr);
  int            hd_tag = head_tag_from_hd1(hd1);

  /* TEMP: */
  /* doublink_verify(&st->garb_set); */
  
  if(hd_tag == 1) {
    void*     sys_ptr = hd1;
    unsigned int size = size_from_hd1(hd1);
    actk_freeset_free(size, sys_ptr);
  }
  else if(hd_tag >= 2) {
    actk_alloc_hd2_t* hd2 = hd2_from_hd1(hd1);
    doublink_remove(&hd2->allocset_node);
    doublink_push_last(&st->garb_set, &hd2->allocset_node);
    hd2_disown_garbage(hd2);
  }
  else {
    abort();
  }

  /* TEMP: */
  /* doublink_verify(&st->garb_set); */

}

void
actk_alloc_garball(actk_state_t* st, actk_closure_t* closure) {
  doublink_node_t* last;
  doublink_node_t* first = doublink_seq_remove(& closure->alloc_set, &last);
  doublink_node_t* node = first;

  /* TEMP: */
  /* doublink_verify(&st->garb_set); */
  
  while(node) {
    actk_alloc_hd2_t* hd2 = (actk_alloc_hd2_t*) node;
    hd2_disown_garbage(hd2);
    node = node->next;
  }

  if(first) {
    doublink_seq_append(& st->garb_set, first, last);
  }
  
  /* TEMP: */
  /* doublink_verify(&st->garb_set); */
}

void actk_alloc_adoptall(actk_state_t* st,
                         actk_closure_t* closure0,
                         actk_closure_t* closure1) {
  doublink_node_t* last;
  doublink_node_t* first = doublink_seq_remove(& closure0->alloc_set, &last);
  doublink_node_t* node = first;

  while(node) {
    actk_alloc_hd2_t* hd2 = (actk_alloc_hd2_t*) node;
    hd2_adopt_live(hd2, closure1);
    node = node->next;
  }

  if(first) {
    doublink_seq_append(& closure1->alloc_set, first, last);
  }  
}

void
actk_alloc_freeall(actk_state_t* st) {
  doublink_node_t* last;
  doublink_node_t* first;
  doublink_node_t* node;

  first = doublink_seq_remove(&st->garb_set, &last);
  node = last;
  
  while(node) {
    actk_alloc_hd2_t* hd2 = (actk_alloc_hd2_t*) node;
    unsigned int     size = size_from_hd1(hd2->hd1);
    unsigned int   hd_tag = head_tag_from_hd1(hd2->hd1);
    unsigned int    hd_sz = head_size_from_hd1(hd2->hd1);
    void*         sys_ptr = NULL;
    
    switch(hd_tag) {
    case 3: {
      actk_alloc_hd3_t*   hd3 = hd3_from_hd2(hd2);
      alloctbl_t*    alloctbl = alloctbl_from_hd3(hd3);

      /* TEMP: */
      /* fprintf(stderr, "freeing: node=%p\n", &hd3->alloctbl_node); */

      /* Remove from alloctbl */
      alloctbl_remove(alloctbl, &hd3->alloctbl_node);
      
      if(sys_ptr == NULL)
        sys_ptr = keyv_from_hd3(hd3);
    }
    case 2: {
      /* Note on Recctxs: Originally I had planned on having some
         special code here to free their resources (e.g., memory for
         alloc/memo-tables) but now I have a hashtable implementation
         which enforces the policy that when it is empty (no entries)
         it will consume no additional memory resources.  We still
         must guarrentee that all it's entries will be removed before
         it is itself freed.  Currently I believe this property
         holds.*/
      if(sys_ptr == NULL)
        sys_ptr = hd2;
      break;
    }
    default:
      abort();
    }
    
    /* push A into freeset */
    node = node->prev;
    actk_freeset_free(size, sys_ptr);
    actk__stats_incr(st->stats, ALLOC_FREED);
    actk__stats_add(st->stats, USER_FREE, size - hd_sz);
  }
}

/* Update the maxlive, if necessary */
void actk_alloc_dostats(actk_state_t* st) {
  actk_stats_val_t bytes_malloc = actk__stats_get(st->stats, TOTAL_MALLOC);
  actk_stats_val_t bytes_free   = actk__stats_get(st->stats, TOTAL_FREE);
  actk_stats_val_t bytes_live   = bytes_malloc - bytes_free;

  actk__stats_max(st->stats, TOTAL_MAXLIVE, bytes_live);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* if the owner is non-NULL, we steal it, otherwise it's garbage and
   we reuse it.  Both cases are handled here. In either case, we do
   _not_ zero the user space. */
void actk_alloc_reuse(actk_state_t* st, actk_alloc_hd3_t* hd3) {
  actk_alloc_ownertag_t ownertag = ownertag_from_hd3(hd3);

  logcatf(st, "reusing allocation %p", hd3->hd2->hd1->user_ptr);
  actk__stats_incr(st->stats, ALLOC_REUSE);

  /* if there is an owning closure, enqueue it. */
  if(ownertag == ACTK_OWNER_CLOSURE) {
    actk_closure_t* owner = owner_from_hd3(hd3);

    logcatf(st, " -- stole allocation %p from " DBSTR_CLOSURE,
            hd3->hd2->hd1->user_ptr,
            DBVAS_CLOSURE(owner));

    actk__closure_enqueue(st, owner);
  }

  
  /* move allocation into liveset for st->closure */
  doublink_remove(&(hd3->hd2->allocset_node));
  doublink_push_last(&(st->closure->alloc_set), &(hd3->hd2->allocset_node));
  hd3_set_owner(hd3, ACTK_OWNER_CLOSURE, st->closure);  
}

/* Returns size bytes of zeroed space, using given keys and associated
   with the currently-running closure */
void* actk_alloc_fresh(actk_state_t* st, unsigned int req_size, int keyc, void** keyv) {  
  int            hd_tag = head_tag_from_keyc(st, keyc);
  unsigned int    hd_sz = head_size_from_head_tag(hd_tag, keyc);
  unsigned int total_sz = req_size + hd_sz;
  
  /* We use this union to avoid casts where possible */
  union {
    void*             ptr;
    actk_alloc_hd1_t* hd1;
    actk_alloc_hd2_t* hd2;
    actk_alloc_hd3_t* hd3;
  } p;

  /* Allocate the memory for the user and our header(s) */
  p.ptr = actk_freeset_malloc(total_sz);
  actk__stats_incr(st->stats, ALLOC_FRESH);
  actk__stats_add(st->stats, USER_MALLOC, req_size);
  
  switch(hd_tag) {
  case 3: {
    alloctbl_t* alloctbl = current_alloctbl(st);
    void**      hd3_keyv = p.ptr;
    int         i;

    /* The first keyc words are for keys. */
    p.hd3 = (void*) ((void**)p.ptr + keyc);
    
    /* Copy the keys into Header-(3) */
    p.hd3->keyc = keyc;
    for(i = 0; i < keyc; i++) {
      hd3_keyv[i] = keyv[i];
    }

    /* Insert the allocation into the allocation table. */
    alloctbl_insert(alloctbl, &(p.hd3->alloctbl_node));

#if USE_ALLOCSET
    /* Setup a pointer to the owner */
    hd3_set_owner(p.hd3, ACTK_OWNER_CLOSURE, st->closure);
#endif
    /* Get Header-(2) */
    p.hd2 = p.hd3->hd2;
  }    
  case 2: {
#if USE_ALLOCSET
    /* Insert into the liveset for st->closure */
    doublink_push_last(&st->closure->alloc_set, &p.hd2->allocset_node);

    /* TEMP: */
    /* doublink_verify(&st->closure->alloc_set); */
#endif
    /* Get Header-(1) */
    p.hd1 = p.hd2->hd1;
  }
  case 1: {
    logcatf(st, "fresh allocation %p of %d bytes", p.hd1->user_ptr, req_size);
    
    /* Store the requested-size (for freeing later) and which headers
       are present. */
    p.hd1->size =  (unsigned int) ptr_addflags((void*) total_sz, hd_tag);
    memset(p.hd1->user_ptr, 0, req_size);
    
    return p.hd1->user_ptr;
  }
  default:
    abort();
    return NULL;
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
