#ifndef __ALLOCATION2_IMPL__
#define __ALLOCATION2_IMPL__

#include "allocation2.h"
#include "alloctbl.h"
#include "../adt/pointer.h"
#include "../adt/doublink.h"

/**
 * \addtogroup actk_private_allocation
 * @{
 */

/* Figure: 3-Layer Header Layout */
/* (The headers with higher numbers come earlier in memory). */
/*
  Header-(3)         ( >= 5 words)
     +-------------+
     | key_1       |
     +-------------+
     |  .          |
     |  .          |
     +-------------+
     | key_{keyc}  |
     +-------------+
     | keyc        |
     +-------------+
     | next        |
     +-------------+
     | prev        |
     +-------+-----+
     | owner | F_3 |
     +-------+-----+
  Header-(2)         ( == 2 words)
     +-------------+
     | next        |
     +-------+-----+
     | prev        |
     +-------+-----+
  Header-(1)         ( == 1 word)
     +-------+-----+
     | size  | F_1 |
     +-------+-----+
  (User data)
     +-------------+
     |  .          |
     |  .          | 
*/

/*
  Header Descriptions:

  Def: A _user-allocation_ is any memory-block whose addresses may be
  written to modrefs during some run of the program.
  
  Header-(1): All user allocations.  This is required in order to take
  advantage of the freeset system since we must tell the freeset how
  many bytes we are freeing.  Since our allocations are word-aligned,
  we use also use the spare LS-bits (labeled F_1 in the figure) to tag
  the presence of the other two (optional) headers.
  
  Header-(2): All user-allocations while bookkeeping is enabled. This
  header is required to preserve the soundness of modref-content
  equality and the allocations performed by closures, which are
  automatically collected when these closure's are invalidated.  

  Header-(3): All _keyed_ user-allocations.  These allocations are
  performed with the potential for stealing and resuse, and they are
  found via their keys.  The owner field either refers to a closure
  (in the case that the allocation is Live) or a recctx (in the case
  that the allocation is Garbage).  Any owning-closure must be
  enqueued when stolen-from, and upon freeing the allocation, the
  corresponding allocation-table must be updated.

  (User data): The remainder of the allocation's memory-block is for
  the user's use.  The user manages how they wish to potentially
  subdivide the available space (which is equal to the ammount they
  requested, not the ammount stored in hd1). The pointer returned from
  allocation requests points to the first byte here, and this pointer
  is used to make the allocation into garbage, should the user want
  to.  Concerning the collection of this memory, note that: (i)
  user-allocations become garbage automatically when their
  corresponding closure is invalidated; (2) the toplevel closure is
  never invalidated, so allocations done here must be made into
  garbage explicitly.
*/

struct actk_alloc_hd1_s {
  unsigned int size;
  void* user_ptr[0];
};

struct actk_alloc_hd2_s {
  doublink_node_t allocset_node;
  actk_alloc_hd1_t  hd1[0];
};

struct actk_alloc_hd3_s {
  void* keyv_end[0];
  doublink_node_t alloctbl_node;
  unsigned int keyc;
  union {
    adt_ptr_t ptr;
    actk_closure_t* closure;
    alloctbl_t* table;
  } owner;
  actk_alloc_hd2_t  hd2[0];
};

typedef enum actk_alloc_ownertag_e {
  ACTK_OWNER_CLOSURE,
  ACTK_OWNER_TABLE,
} actk_alloc_ownertag_t;

/* Steal the allocation from it's current owner, if any.  This takes
 * the allocation on one of the following transitions:
 *
 *  -- (Live --> Live)
 *  -- (Garb --> Live).
 */
void actk_alloc_reuse(actk_state_t* st, actk_alloc_hd3_t* hd3);

/* Make a fresh allocation from the freeset. This takes the allocation
 * on the following transition:
 *
 *  -- (Free -> Live)
 */
void* actk_alloc_fresh(actk_state_t* st, unsigned int size, int keyc, void** keyv);

/* @} */

#endif
