#include "actk_directives.h"
#include "state_impl.h"
#include "closure_impl.h"
#include "stats.h"
#include "logging.h"
#include "actk_debug.h"
#include "propagation.h"
#include "timestamp.h"

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Slot wrangling. */

static int
actk_closure_slot_size(actk_funarg_t* arg) {
  switch(arg->desc) {
  case ACTK_FUN_ARG_READ_LOC_OR_VAL:
    return sizeof(actk_closure_read_t);
  case ACTK_FUN_ARG_CONST_DST:
    return sizeof(actk_val_t);
  case ACTK_FUN_ARG_LIFTD_DST:
    return sizeof(actk_modref_t);    
  case ACTK_FUN_ARG_CONST_VAL:
    return sizeof(actk_val_t);
  case ACTK_FUN_ARG_ENVIR_VAL:
    return 0;
  default:
    abort();
  }
}

static int
actk_recctx_slot_size(actk_funarg_t* arg) {
  switch(arg->desc) {
  case ACTK_FUN_ARG_READ_LOC_OR_VAL: 
  case ACTK_FUN_ARG_LIFTD_DST: 
  case ACTK_FUN_ARG_CONST_DST:
  case ACTK_FUN_ARG_CONST_VAL:
    return 0;
  case ACTK_FUN_ARG_ENVIR_VAL:
    return sizeof(actk_val_t);
  default:
    abort();
  }
}


static actk_closure_slot_t*
actk_closure_slot_first(actk_closure_t* closure) {
  return (actk_closure_slot_t*) (closure + 1);
}

static actk_closure_slot_t*
actk_closure_slot_next(actk_funarg_t* arg,
                       actk_closure_slot_t* slot) {
  actk_byte_t* p = (actk_byte_t*) slot;
  p += actk_closure_slot_size(arg);
  return (actk_closure_slot_t*) p;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int actk__closure_compare(actk_closure_t* closure1, actk_closure_t* closure2) {
  return totalorder_compare(closure1->interval.start,
                            closure2->interval.start);
}

actk_closure_t* actk__closure_setpos(actk_closure_t* closure, int pos) {
  closure->readqueue_pos = pos;
  return closure;
}

int actk__closure_getpos(actk_closure_t* closure) {
  return closure->readqueue_pos;
}

void
actk__closure_sizes(actk_fun_t* fun, int* closure_bytec_, int* recctx_bytec_)
{  
  /* Used cached sizes */
  if(fun->sig->closure_bytec >= 0) {
    if(closure_bytec_) *closure_bytec_ = fun->sig->closure_bytec;
    if(recctx_bytec_)  *recctx_bytec_  = fun->sig->recctx_bytec;
    return;
  }

  /* Or compute from scratch: */ {
    int i;
    int closure_bytec = sizeof(actk_closure_t);
    int recctx_bytec = sizeof(actk_recctx_t);

    for(i = 0; i < fun->sig->argc; i++) {
      actk_funarg_t* arg = &fun->sig->args[i];
      closure_bytec += actk_closure_slot_size(arg);
      recctx_bytec += actk_recctx_slot_size(arg);
    }
    fun->sig->closure_bytec = closure_bytec;
    fun->sig->recctx_bytec = recctx_bytec;
    actk__closure_sizes(fun, closure_bytec_, recctx_bytec_);
  }
}

actk_closure_t* actk__closure_new(actk_state_t* st,
                                  actk_fun_t* fun,
                                  actk_recctx_t* recctx,
                                  actk_argvmask_t rkeymask,
                                  actk_argvmask_t readmask,
                                  actk_val_t* arg_vals) {
  int                  closure_bytec;
  actk_closure_t*      closure;
  actk_closure_slot_t* slot;
  int                  arg_i;
  actk_val_t*          arg_input;
  actk_funarg_t*       arg_fun;
  
  /* Allocation */
  actk__closure_sizes(fun, &closure_bytec, NULL);
  closure = actk_freeset_malloc(closure_bytec);

  /* Initialization of standard fields */
  closure->memotbl_node.prev = NULL;
  closure->memotbl_node.next = NULL;
  closure->recctx            = recctx;
  closure->color             = ACTK_CLOSURE_NONE;
  closure->interval.start    = NULL;
  closure->interval.end      = NULL;
  closure->readqueue_pos     = -3;
  doublink_init(&closure->alloc_set);
  
  /* NULL-check needed for toplevel */
  IF_DEBUG(closure->num_recur = recctx ? recctx->next_num_closure ++ : 0);
  IF_DEBUG(closure->num = fun->next_num_closure ++);


  /* Initialization of closure payload */
  logcatf(st,"closure for: %s", fun->sig->name);
  actk__stats_incr(st->stats, CLOSURE_NEW);  

  /* The payload consists of an array of slots, each of which
     corresponds to a formal argument (an actk_fun_arg_t); each slot
     is initialized according to the formal arguments descriptor and
     the given input (actual) arguments */
  slot = actk_closure_slot_first(closure);

  for(arg_fun = fun->sig->args, arg_i = 0, arg_input = arg_vals;
      arg_fun < fun->sig->args + fun->sig->argc;
      arg_fun ++, arg_i ++, arg_input ++)
  {

    if((1 << arg_i) & rkeymask) {
      /* Ignore the actk_rkey() arguments */
      arg_fun --;
      continue;
    }
    
    switch(arg_fun->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      {
        if((1 << arg_i) & readmask) {
          actk_modref_t* modref = (actk_modref_t*) *arg_input;
          slot->read.modref = modref;
          slot->read.val    = actk_deref(st, modref);
          slot->read.handle = NULL;
          
          logcatf(st," RL arg: %10s = !(" DBSTR_MODREF " <val %p=%d>)",
                  arg_fun->name,
                  DBVAS_MODREF(slot->read.modref),
                  slot->read.val,
                  slot->read.val);
        }
        else {
          slot->read.modref = NULL;
          slot->read.val    = *(arg_input);
          slot->read.handle = NULL;
          
          logcatf(st," RV arg: %10s = %p / %d",
                  arg_fun->name, *arg_input, *arg_input);
        }
      }
      break;
    case ACTK_FUN_ARG_LIFTD_DST:
      {
        actk_modref_t* m = *arg_input;        
        assert(!((1 << arg_i) & readmask));

        actk_modref_shadow_w_target(st, &slot->write, m);
        logcatf(st," WL arg: %10s = %p (shadow %p, target %p)",
                arg_fun->name, m, &slot->write, m);
      }
      break;
    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_CONST_VAL:
      {
        assert(!((1 << arg_i) & readmask));
        logcatf(st," CV arg: %10s = %p / %d",
                arg_fun->name, *arg_input, *arg_input);
        
        slot->const_val = *arg_input;
      }
      break;
    case ACTK_FUN_ARG_ENVIR_VAL:
      {
        assert(!((1 << arg_i) & rkeymask));
        assert(!((1 << arg_i) & readmask));
      }
      break;
    default:
      abort();
    }

    slot = actk_closure_slot_next(arg_fun, slot);
  }
  assert( (((char*)slot) - closure_bytec) == (char*) closure);
  return closure;  
}

void actk__closure_free(actk_state_t* st, actk_closure_t* closure) {
  int closure_size;
  actk__closure_sizes(closure->recctx->fun, &closure_size, NULL);
  actk_freeset_free(closure_size, closure);
}

actk_recctx_t* actk__recctx_new(actk_state_t* st,
                                actk_fun_t* fun,
                                actk_argvmask_t rkeymask,
                                actk_val_t* args) {
  const int            MAX = 8;
  int                  recctx_bytec;
  actk_recctx_t*       recctx;
  actk_val_t*          arg_recctx;
  actk_val_t*          arg_input;
  int                  arg_i;
  actk_val_t           rkeyv[MAX];
  int                  rkeyc = 0;
  actk_funarg_t*       arg_fun;

  actk__stats_incr(st->stats, RECCTX_NEW);

  /* Step 1: compute the size of the recctx */
  actk__closure_sizes(fun, NULL, &recctx_bytec);

  /* Step 2: extract the actk_rkey() arguments and fill an array with
     them (to prepare for the keyed-allocation) */
  for(arg_fun = fun->sig->args, arg_i = 0, arg_input = args;
      arg_fun < fun->sig->args + fun->sig->argc;
      arg_fun ++, arg_i ++, arg_input ++)
  {
    if((1 << arg_i) & rkeymask) {
      arg_fun --;
      rkeyv[rkeyc ++] = *arg_input;
    }
  }

  logcatf(st, "allocating recctx ...");
  
  /* Step 3: perform the keyed allocation */
  recctx = actk_alloc_new(st, recctx_bytec, rkeyc, rkeyv);
  
  /* Step 4A: Reuse. */
  if(recctx->fun) {
    actk__stats_incr(st->stats, RECCTX_REUSE);
  }
  /* Step 4B: if the allocation is fresh, we must initialize it */
  else {
    actk__stats_incr(st->stats, RECCTX_FRESH);
    
    recctx->fun = fun;
    recctx->end = NULL;
    
    IF_DEBUG( recctx->num = fun->next_num_recctx ++);
    IF_DEBUG( recctx->next_num_closure = 1);

    /* Memo-table */
    if((st->flags & ACTK_BOOKKEEPING) &&
       (fun->flags & ACTK_USEMEMOTABLE)) {
      memotbl_init(&recctx->memotbl, 0);
    }
    else {
      memotbl_init(&recctx->memotbl, 0);
    }

    /* Alloc-table */
    if((st->flags & ACTK_BOOKKEEPING) &&
       (fun->flags & ACTK_USEALLOCTABLE)) {
      alloctbl_init(&recctx->alloctbl, 0);
    }
    else {
      alloctbl_init(&recctx->alloctbl, 0);
    }
    
    arg_recctx = (actk_val_t*)(recctx + 1);
    
    logcatf(st,"recctx for: %s", fun->sig->name);

    /* Pack the corresponding arguments into the allocated block */
    for(arg_fun = fun->sig->args, arg_i = 0, arg_input = args;
        arg_fun < fun->sig->args + fun->sig->argc;
        arg_fun ++, arg_i ++, arg_input ++)
    {
      
      /* Here we can ignore the actk_rkey() arguments */
      if((1 << arg_i) & rkeymask) {
        arg_fun --;
        continue;
      }
    
      switch(arg_fun->desc) {
      case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      case ACTK_FUN_ARG_LIFTD_DST:
      case ACTK_FUN_ARG_CONST_DST:
      case ACTK_FUN_ARG_CONST_VAL:
        /* nothing to do */
        break;
      case ACTK_FUN_ARG_ENVIR_VAL:
        *(arg_recctx++) = *(arg_input);
        logcatf(st," EV arg: %10s = %p / %d", arg_fun->name, *arg_input, *arg_input);
        break;
      default:
        abort();
      }
    }  
    assert( (((char*)arg_recctx) - recctx_bytec) == (char*) recctx);
  }
  
  return recctx;  
}


void actk__closure_subscribe(actk_state_t* st, actk_closure_t* closure) {
  actk_funsig_t* sig = closure->recctx->fun->sig;
  actk_closure_slot_t* slot = actk_closure_slot_first(closure);
  int arg_i;
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL: {      
      if(slot->read.modref) {
        slot->read.handle = actk_modref_read(st, slot->read.modref, closure);
        logcatf(st,"read "DBSTR_MODREF", handle is %p",
                DBVAS_MODREF(slot->read.modref),
                slot->read.handle);
      }
      break;
    }
    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_LIFTD_DST:
    case ACTK_FUN_ARG_CONST_VAL:
    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    }

    slot = actk_closure_slot_next(arg, slot);
  }    
}

void actk__closure_unsubscribe(actk_state_t* st, actk_closure_t* closure) {
  actk_funsig_t* sig = closure->recctx->fun->sig;
  actk_closure_slot_t* slot = actk_closure_slot_first(closure);
  int arg_i;
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      if(slot->read.modref) {
        actk_closure_t* closure_ __attribute((unused))
          = (actk_closure_t*) readlist_remove(slot->read.handle);
        assert(closure == closure_);
      }
      break;
    case ACTK_FUN_ARG_LIFTD_DST:
    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_CONST_VAL:
    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    default:
      abort();
    }

    slot = actk_closure_slot_next(arg, slot);
  }
}

void
actk__closure_enqueue(actk_state_t* st, actk_closure_t* closure) {
  assert(closure->color == ACTK_CLOSURE_WHITE ||
         closure->color == ACTK_CLOSURE_GREY );
  
  if(closure->color == ACTK_CLOSURE_WHITE) {
    logcatf(st, "-- pushing closure " DBSTR_CLOSURE " onto queue", DBVAS_CLOSURE(closure));
    readqueue_push(st->chpctx->queue, closure);
    closure->color = ACTK_CLOSURE_GREY;
  }
  else {
    logcatf(st, "-- (closure " DBSTR_CLOSURE " already on queue)", DBVAS_CLOSURE(closure));
  }
}

static void
actk__closure_set_start(actk_state_t* st, actk_closure_t* closure) {

  /* Tag the closure as being a closure (rather than a recctx, etc) */
  to_data_t timestamp_data = (to_data_t)
    ptr_addflags(closure, ACTK_TIMESTAMP_CLOSURE_START);
  
  /* Start-time is a fresh successor to now */
  closure->interval.start = totalorder_insert(st->now, timestamp_data);
  
}

static void
actk__closure_set_end(actk_state_t* st, actk_closure_t* closure) {
  assert(!closure->interval.end);

  if(closure->recctx &&
     closure->recctx->fun->flags & ACTK_TAILRECURSIVE &&
     st->flags & ACTK_TAILRECURSION) {
    
    to_data_t timestamp_data = (to_data_t)
      ptr_addflags(closure->recctx, ACTK_TIMESTAMP_RECCTX_END);
    
    /* CASE: No end-time yet => fresh run. */
    if(!closure->recctx->end) {
      closure->recctx->end = totalorder_insert(st->now, timestamp_data);
      logcatf(st, "(new) end-time for recctx : %p", closure->recctx->end);
    }

    /* CASE: The end of the recctx is in the future (as the result of
       stealing it) -- so pick a new one after "now". */    
    else if(totalorder_compare(st->now, closure->recctx->end) < 0) {
      
      logcatf(st, "(stale) end-time for recctx : %p", closure->recctx->end);
      closure->recctx->end = totalorder_insert(st->now, timestamp_data);
      logcatf(st, "(fresh) end-time for recctx : %p", closure->recctx->end);
    }

    /* CASE: The end of the recctx is "now" */
    else if(totalorder_compare(st->now, closure->recctx->end) == 0) {
      logcatf(st, "end-time for recctx is 'now': %p", closure->recctx->end);
    }

    /* CASE: ERROR: The end of the recctx is in the past. */
    else {
      abort();
    }    
  }
  else {
    closure->interval.end = totalorder_insert(st->now, actk_timestamp_nulldata);
    logcatf(st, "(new) end-time for closure : %p", closure->interval.end);
  }

  st->now = actk__closure_get_end(st, closure);
}

actk_timestamp_t*
actk__closure_get_end(actk_state_t* st, actk_closure_t* closure) {
  if(closure->recctx &&
     closure->recctx->fun->flags & ACTK_TAILRECURSIVE &&
     st->flags & ACTK_TAILRECURSION ) {
    return closure->recctx->end;
  }
  return closure->interval.end;
}

/* Sets-up a closure, once we are certain it will be executed (there
   are no memo-matches for it, etc).  The expense of this setup is in
   (1) adding the closure to the readlists of the modrefs in it's
   read-set, (2) creating fresh timestamps for the closure's
   start-time / end-time. */
void actk__closure_init(actk_state_t* st,
                        actk_closure_t* closure) {

  actk__stats_incr(st->stats, CLOSURE_INIT);
  
  logcatf(st, "initializing closure " DBSTR_CLOSURE "", DBVAS_CLOSURE(closure));

  if(st->flags & ACTK_BOOKKEEPING) {
    /* Add the closure to the readlist of each modref it reads-from, and
       save each new readlist reference in the closure */ 
    actk__closure_subscribe(st, closure);
    
    /* Setup closure's interval */
    actk__closure_set_start(st, closure);
    
  }

  assert(closure->color == ACTK_CLOSURE_NONE);
  closure->color = ACTK_CLOSURE_WHITE;
}

void actk__closure_enter(actk_state_t* st,
                         actk_closure_t* closure,
                         actk_closure_enter_flags_t flags)
{
  actk_closure_t* caller_closure = st->closure;

  logcatf(st,"entering read " DBSTR_CLOSURE "", DBVAS_CLOSURE(closure));
  logcatf(st," closure interval is [%p, %p]",
          closure->interval.start,
          actk__closure_get_end(st, closure));
  
  assert(closure->color == ACTK_CLOSURE_WHITE);

  /* adjust state: */
  st->now = closure->interval.start;  
  st->closure = closure;
  st->fun_loop_on = 1;

  /* -- do (or skip) function call
     -- setup end-time afterwards (if call is taken). */
  switch(flags) {
  case ACTK_CLOSURE_ENTER_NEVER:    
    break;
  case ACTK_CLOSURE_ENTER_AUTO:
    /* CHECK: are we doing a tail-call ? */
    if((st->flags & ACTK_TAILRECURSION) &&
       (closure->recctx->fun->flags & ACTK_TAILRECURSIVE) &&
       (caller_closure->recctx == closure->recctx)) {
      /* TODO: if not bookkeeping, free the caller-closure */
      break;
    }
    /* OTHERWISE: we are doing a regular stack-call */
    else {
      /* Do function call: */
      closure->recctx->fun->funptr(st);

      if(st->flags & ACTK_BOOKKEEPING) {
        /* Set the closure's end-time, move ahead "now": */
        actk__closure_set_end(st, closure);
      }
      else {
        /* TODO: free the closure */
      }
      
      /* Restore the caller_closure as the active closure: */
      st->closure = caller_closure;
    }
    break;
  case ACTK_CLOSURE_ENTER_PROPAGATE:
    closure->recctx->fun->funptr(st);
    if(!actk__closure_get_end(st, closure)) {
      actk__closure_set_end(st, closure);
    }
    break;
  }
  
  /* increment the counter for AUTO and ALWAYS */
  switch(flags) {
  case ACTK_CLOSURE_ENTER_NEVER:
    break;
  case ACTK_CLOSURE_ENTER_AUTO:
  case ACTK_CLOSURE_ENTER_PROPAGATE:
    actk__stats_incr(st->stats, CLOSURE_ENTER);
    break;
  }  
}



actk_closure_t*
actk__closure_wakeup(actk_state_t* st, actk_closure_t* closure0)
{
  actk_recctx_t*      recctx = closure0->recctx;
  actk_fun_t*            fun = recctx->fun;
  actk_funsig_t*         sig = fun->sig;
  memotbl_t*         memotbl = NULL;
  actk_closure_t*   closure1 = NULL;
  int          closure_bytec;  
  
  memotbl = (((st->flags & ACTK_BOOKKEEPING) &&
              (fun->flags & ACTK_USEMEMOTABLE))
             ?  &recctx->memotbl
             : NULL
             );
  
  /* Does closure1 use a memotbl? */
  if(0 &&
     memotbl &&
     (closure0->memotbl_node.next || closure0->memotbl_node.prev))
  {
    /* If yes, then create closure1 to be keyed in the memotbl using
       the old arguments (read values, allocations, etc.) of
       closure1 */
    
    /* Allocation of closure1 */
    actk__closure_sizes(fun, &closure_bytec, NULL);
    closure1 = actk_freeset_malloc(closure_bytec);

    /* Initialization of standard fields */
    doublink_replace(&closure0->memotbl_node, &closure1->memotbl_node);  
    closure1->recctx         = recctx;
    closure1->color          = ACTK_CLOSURE_WHITE;
    closure1->interval.start = totalorder_insert(closure0->interval.start, (to_data_t) closure1);
    closure1->interval.end   = closure0->interval.end;
    closure1->readqueue_pos  = -3;
    doublink_init(&closure1->alloc_set);
  
    IF_DEBUG(closure1->num_recur = recctx->next_num_closure);
    IF_DEBUG(closure1->num = fun->next_num_closure ++);
  
    /* Transfer all of closure0's allocations to closure1 */
    actk_alloc_adoptall(st, closure0, closure1);
  }
  else {
    /* If no, then make all of closure0's allocations into garbage. */
    actk_alloc_garball(st, closure0);

    if(memotbl)
      memotbl_remove(memotbl, (doublink_node_t*) closure0);
  }

  /* Either way, closure0 is now live (and not in the queue any longer) */
  closure0->color = ACTK_CLOSURE_WHITE;  
  
  /* Process the closure's payload: */
  
  /* (1) Update (old) closure0                */
  /*    -- new read values put into closure0  */

  /* (2) Initialize (new) closure1                                     */
  /*     -- old read values in closure0 become read values in closure1 */
  /*     -- constant values are copied from closure0 to closure1       */
  /*     -- old write values in closure0 are copied into closure1      */
  {
    actk_closure_slot_t* slot0 = actk_closure_slot_first(closure0);
    actk_closure_slot_t* slot1 = closure1 ? actk_closure_slot_first(closure1) : NULL;
    int arg_i;
    
    for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
      actk_funarg_t* arg = &sig->args[arg_i];
      
      switch(arg->desc) {
      case ACTK_FUN_ARG_READ_LOC_OR_VAL:
        /* First, copy (old) read values from closure0 to closure1 */
        if(slot1) {
          slot1->read.val    = slot0->read.val;
          slot1->read.modref = NULL;
          slot1->read.handle = NULL;
        }
        
        /* Then, update closure0 with current values */
        if(slot0->read.modref) {
          slot0->read.val = actk_deref(st, slot0->read.modref);
        }        
        break;
      case ACTK_FUN_ARG_LIFTD_DST:
        if(slot1){
          actk_modref_shadow_w_val(st, &slot1->write, actk_deref(st, &slot0->write));
        }
        break;
      case ACTK_FUN_ARG_CONST_DST:
      case ACTK_FUN_ARG_CONST_VAL:
        if(slot1) {
          slot1->const_val = slot0->const_val;
        }
        break;
      case ACTK_FUN_ARG_ENVIR_VAL:
        break;
      }
      
      slot0 = actk_closure_slot_next(arg, slot0);
      if(slot1) slot1 = actk_closure_slot_next(arg, slot1);
    }
  }    
  
  /* (re-)insert into memotable */
  if(memotbl) {
    memotbl_insert(memotbl, (doublink_node_t*) closure0);
  }

  logcatf(st, "waking up: " DBSTR_CLOSURE, DBVAS_CLOSURE(closure0));
  logcatf(st, "  created: " DBSTR_CLOSURE, DBVAS_CLOSURE(closure1));
  
  return closure1;
}

hash_val_t actk__closure_hashval(actk_closure_t* closure) {
  actk_funsig_t* sig = closure->recctx->fun->sig;
  actk_closure_slot_t* slot = actk_closure_slot_first(closure);
  hash_val_t hashval = 0;
  int arg_i;
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:      
      hashval = hash_hsieh_1step(hashval, (void*) slot->read.val);
      break;

    case ACTK_FUN_ARG_LIFTD_DST:
      break;

    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_CONST_VAL:
      hashval = hash_hsieh_1step(hashval, (void*) slot->const_val);
      break;

    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    default:
      abort();
    }

    slot = actk_closure_slot_next(arg, slot);
  }
  
  return hash_hsieh_avalanche(hashval);
}

int actk__closure_equals(actk_closure_t* closure_1,
                         actk_closure_t* closure_2) {
  actk_funsig_t* sig = closure_1->recctx->fun->sig;
  actk_closure_slot_t* slot1 = actk_closure_slot_first(closure_1);
  actk_closure_slot_t* slot2 = actk_closure_slot_first(closure_2);

  int arg_i;
  
  if(closure_1->recctx != closure_2->recctx)
    return 0;
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      if(slot1->read.val != slot2->read.val)
        return 0;      
      break;
      
    case ACTK_FUN_ARG_LIFTD_DST:
      break;

    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_CONST_VAL:
      if(slot1->const_val != slot2->const_val)
        return 0;      
      break;

    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    }

    slot1 = actk_closure_slot_next(arg, slot1);
    slot2 = actk_closure_slot_next(arg, slot2);
  }
  return 1;
}


void actk__closure_replace(actk_state_t* st,
                           actk_closure_t* closure_old,
                           actk_closure_t* closure_new) {
  actk_funsig_t*  sig = closure_old->recctx->fun->sig;
  actk_closure_slot_t* slot_old = actk_closure_slot_first(closure_old);
  actk_closure_slot_t* slot_new = actk_closure_slot_first(closure_new);
  int arg_i;

  logcatf(st, "replacing (new): " DBSTR_CLOSURE, DBVAS_CLOSURE(closure_new));
  logcatf(st, "     with (old): " DBSTR_CLOSURE, DBVAS_CLOSURE(closure_old));
  
  assert(closure_new->color == ACTK_CLOSURE_NONE);

  assert(closure_old->color == ACTK_CLOSURE_GREY ||
         closure_old->color == ACTK_CLOSURE_WHITE );
  
  actk__stats_incr(st->stats, CLOSURE_REPLACE);
  assert(actk__closure_equals(closure_old, closure_new));

  /* Unsubscribe the old closure */
  actk__closure_unsubscribe(st, closure_old);

  /* Process the slots of the old closure, copying slot fields from
     new to old:
     
     -- read slots: copy read values and modrefs
     -- const vals: copy.
     -- write locs: copy.
     -- envir vals: (n/a).
  */
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      /* 0: copy modref pointer , if any */
      /* 1: copy value from modref */
      /* 2: clear readlist handle, also, neither new nor old closure
         shouldn't have one (at this point). */      
      slot_old->read.modref = slot_new->read.modref;
      slot_old->read.val    = slot_new->read.val;
      slot_old->read.handle = NULL;     
      break;

    case ACTK_FUN_ARG_LIFTD_DST:
      /* Replace shadow modref: */
      actk_modref_shadow_replace(st, &slot_old->write, &slot_new->write);
      break;
      
    case ACTK_FUN_ARG_CONST_VAL:
    case ACTK_FUN_ARG_CONST_DST:
      /* copy const */
      slot_old->const_val = slot_new->const_val;
      break;

    case ACTK_FUN_ARG_ENVIR_VAL:
      break;

    default:
      abort();
    }

    slot_new = actk_closure_slot_next(arg, slot_new);
    slot_old = actk_closure_slot_next(arg, slot_old);
  }
  
  actk__closure_subscribe(st, closure_old);
  actk__closure_free(st, closure_new);

  logcatf(st, "done.");
}

int actk__closure_isuptodate(actk_state_t* st, actk_closure_t* closure) {
  actk_funsig_t* sig = closure->recctx->fun->sig;
  actk_closure_slot_t* slot = actk_closure_slot_first(closure);
  int arg_i;
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL:
      if( slot->read.modref &&
          (slot->read.val != actk_deref(st, slot->read.modref)) )
        return 0;
      break;      
    case ACTK_FUN_ARG_CONST_VAL:
    case ACTK_FUN_ARG_LIFTD_DST:
    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    }

    slot = actk_closure_slot_next(arg, slot);
  }
  return 1;
}


int actk__closure_verify(actk_state_t* st, actk_closure_t* closure) {
  actk_funsig_t* sig = closure->recctx->fun->sig;
  actk_val_t* arg_closure = (actk_val_t*)(closure + 1);
  int arg_i;
  int count_errors = 0;

  assert(closure->color == ACTK_CLOSURE_WHITE ||
         closure->color == ACTK_CLOSURE_GREY );
  
  for(arg_i = 0; arg_i < sig->argc; arg_i++)  {
    actk_funarg_t* arg = &sig->args[arg_i];
    
    switch(arg->desc) {
    case ACTK_FUN_ARG_READ_LOC_OR_VAL: {
      actk_modref_t*      modref = (actk_modref_t*) arg_closure[0];
      /* actk_val_t          argval = arg_closure[1]; */
      /* doublist_handle_t   handle = (doublist_handle_t) arg_closure[2]; */

      if(modref) {
        /*
        actk_allochdr_t*  allochdr =
          actk__hdr_of_block ((actk_val_t*)modref);

        if(! actk__alloc_hasowner(allochdr)) {
          logcatf(st, "ERROR: " DBSTR_CLOSURE " reads ownerless " DBSTR_MODREF,
                  DBVAS_CLOSURE(closure),
                  DBVAS_MODREF(modref));
          count_errors ++;
        }

        if( actk_deref(st, modref) != argval ) {
          logcatf(st,DBSTR_CLOSURE " is inconsistent with " DBSTR_MODREF,
                  DBVAS_CLOSURE(closure),
                  DBVAS_MODREF(modref));
          count_errors ++;
        }
        */
      }
        
      arg_closure += 3;
      break;
    }
    case ACTK_FUN_ARG_LIFTD_DST:
    case ACTK_FUN_ARG_CONST_DST:
    case ACTK_FUN_ARG_CONST_VAL:
      arg_closure += 1;
      break;
    case ACTK_FUN_ARG_ENVIR_VAL:
      break;
    }
  }
  return count_errors;
}


actk_state_t*
actk__closure_invalidate(actk_state_t* st,
                         actk_timestamp_t* timestamp,
                         void* data)
{
  actk_timestamp_type_t timestamp_type = ptr_getflags(data);

  switch(timestamp_type) {
  case ACTK_TIMESTAMP_CLOSURE_START: {
    actk_closure_t* closure = ptr_get(data);
    actk__stats_incr(st->stats, CLOSURE_INVALIDATE);
    
    assert(closure->color == ACTK_CLOSURE_WHITE ||
           closure->color == ACTK_CLOSURE_GREY ||
           closure->color == ACTK_CLOSURE_BLACK );
    
    closure->color = ACTK_CLOSURE_BLACK;
    
    logcatf(st,"invalidating " DBSTR_CLOSURE "", DBVAS_CLOSURE(closure));

    logcatf(st,"    with interval: [%p, %p)",
            closure->interval.start,
            actk__closure_get_end(st, closure));
    
    logcatf(st,"     timestamp is:  %p", timestamp);
    
    /* remove from memotable */
    if(closure->memotbl_node.prev != NULL) {
      memotbl_t* memotbl = &closure->recctx->memotbl;
      memotbl_remove(memotbl, & closure->memotbl_node);
    }
    
    /* remove from priority queue */
    actk__chpctx_remove_closure(st, closure);
    
    /* clear the readlist, which unhooks the closure from it's read locations */
    actk__closure_unsubscribe(st, closure);
    
    /* move allocations to garbage */
    actk_alloc_garball(st, closure);
    
    /* cleanup other malloc's */
    actk__closure_free(st, closure);
    break;
  }
  case ACTK_TIMESTAMP_RECCTX_END: {
    actk_recctx_t* recctx = ptr_get(data);

    if(recctx->end == timestamp) {
      /*
      logcatf(st,"timestamp %p is end-time for recctx : " DBSTR_ALLOC,
              timestamp, DBVAS_ALLOC(actk__hdr_of_block((actk_val_t*)recctx)));
      */
      logcatf(st,"   clearing end-time for recctx.");
      recctx->end = NULL;      
    }
    else {
      /*
      logcatf(st,"timestamp %p is was (but no-longer is) end-time for recctx : " DBSTR_ALLOC,
              timestamp, DBVAS_ALLOC(actk__hdr_of_block((actk_val_t*)recctx)));
      */
      logcatf(st, "   ignoring ...");
    }
    break;
  }
  case ACTK_TIMESTAMP_OTHER: {
    logcatf(st,"timestamp is unclaimed -- nothing to do");
    logcatf(st,"     timestamp is:  %p", timestamp);
    break;
  }
  }
  return st;
}

