#include "modref_impl.h"
#include "state_impl.h"
#include "readlist.h"
#include "closure_impl.h"
#include "logging.h"
#include "propagation.h"
#include "actk_debug.h"
#include "allocation2.h"
#include "stats.h"

actk_val_t actk_deref(actk_state_t* st, actk_modref_t* m) {
  assert(ACTK_MODREF_STATE(m) == MODREF_STATE_WRITTEN);
  return m->u.normal.val;
}

void actk_write(actk_state_t* st, actk_modref_t* m, actk_val_t v) {
  while(1) {
    switch(ACTK_MODREF_BEHAV(m)) {    
    case MODREF_BEHAV_SHADOW: {  
      actk_modref_t* target = m->u.shadow.target;
      
      logcatf(st, "writing: " DBSTR_MODREF " (a shadow with target: " DBSTR_MODREF ")",
              DBVAS_MODREF(m), DBVAS_MODREF(target));
      
      /* Shadow modref becomes a written modref */
      ACTK_MODREF_STATE(m) = MODREF_STATE_WRITTEN;
      m->u.shadow.val = v;
      
      /* Also, write the shadowed target */
      /*assert(ACTK_MODREF_BEHAV(target) == MODREF_BEHAV_NORMAL); */
      m = target;
      continue;
    }
      
    case MODREF_BEHAV_NORMAL: {
      logcatf(st, "writing: " DBSTR_MODREF, DBVAS_MODREF(m));
      
      if(!(st->flags & ACTK_BOOKKEEPING))
        goto update_val;
      
      if(ACTK_MODREF_STATE(m) == MODREF_STATE_WRITTEN) {
        assert(st->flags & ACTK_BOOKKEEPING);
        
        if(!st->chpctx) {
          actk__chpctx_init(st);
        }
        
        /* Is the modref changing value? */
        if(v != m->u.normal.val) {
          /* Handle a change: */
          readlist_iter_t i;
          logcatf(st, "adding readers to change-propagation queue:");
          readlist_iter(&m->u.normal.readlist, &i);
          
          while(!readlist_isend(&i)) {
            actk_closure_t* closure = (actk_closure_t*) readlist_currentval(&i);
            actk__closure_enqueue(st, closure);
            readlist_next(&i);
          }
          logcatf(st, "all readers added.");
        }
        else {
          /* No change. */
          logcatf(st, "no change for %p", m);
        }
        
      }
      else if(ACTK_MODREF_STATE(m) == MODREF_STATE_EMPTY) {
        ACTK_MODREF_STATE(m) = MODREF_STATE_WRITTEN;
      }
      
    update_val:
      logcatf(st, "wrote: " DBSTR_MODREF, DBVAS_MODREF(m));
      m->u.normal.val = v;
      return;
    }
    } /* end switch. */
  } /* end while. */
}

actk_modref_t* actk_modref_init(actk_state_t* st, actk_modref_t* m) {

  /*assert(ACTK_MODREF_BEHAV(m) == MODREF_BEHAV_NORMAL);*/
  
  switch(ACTK_MODREF_STATE(m))
  {
  case MODREF_STATE_EMPTY:
    logcatf(st, "fresh: " DBSTR_MODREF, DBVAS_MODREF(m));
    ACTK_MODREF_STATE(m) = MODREF_STATE_EMPTY;
    if(st->flags & ACTK_BOOKKEEPING) {
      readlist_init(&m->u.normal.readlist);
    }
    m->u.normal.val = (actk_val_t) 0xDEADBEEF;
    break;

  case MODREF_STATE_WRITTEN:
    logcatf(st, "reused: " DBSTR_MODREF, DBVAS_MODREF(m));
    break;

  default:
    logcatf(st, "error: we got a pointer to bad memory: %p", m);
    abort();    
  }
  return m;
}

actk_modref_t* actk_modref_new(actk_state_t* st, int keyc, void** keyv) {  
  actk_modref_t* m = actk_alloc_new(st, sizeof(actk_modref_t), keyc, keyv);  
  actk_modref_init(st, m);
  return m;
}

readlist_handle_t
actk_modref_read(actk_state_t* st,
                 actk_modref_t* m,
                 actk_closure_t* closure) {
  assert(ACTK_MODREF_BEHAV(m) == MODREF_BEHAV_NORMAL);
  assert(ACTK_MODREF_STATE(m) == MODREF_STATE_WRITTEN);  
  return readlist_push(& m->u.normal.readlist, closure);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void
actk_modref_shadow_w_target(actk_state_t* st,
                            actk_modref_t* m,
                            actk_modref_t* m_target)
{
  ACTK_MODREF_BEHAV(m) = MODREF_BEHAV_SHADOW;
  ACTK_MODREF_STATE(m) = MODREF_STATE_EMPTY;
  m->u.shadow.target   = m_target;
  
  logcatf(st, "setup shadow modref %p with target %p",
          m, m_target);
}


void
actk_modref_shadow_w_val(actk_state_t* st,
                         actk_modref_t* m,
                         actk_val_t val)
{
  ACTK_MODREF_BEHAV(m) = MODREF_BEHAV_SHADOW;
  ACTK_MODREF_STATE(m) = MODREF_STATE_WRITTEN;
  m->u.shadow.target   = NULL;
  m->u.shadow.val      = val;
  
  logcatf(st, "setup shadow modref %p with value %p",
          m, val);
}


void
actk_modref_shadow_replace(actk_state_t* st,
                           actk_modref_t* m_old,
                           actk_modref_t* m_new)
{
  logcatf(st, "replacing: " DBSTR_MODREF, DBVAS_MODREF(m_old));
  logcatf(st, "     with: " DBSTR_MODREF, DBVAS_MODREF(m_new));

  /* Assert: m_old is a written shadow modref. */
  /* BUG: this fails: */
  assert(ACTK_MODREF_BEHAV(m_old) == MODREF_BEHAV_SHADOW);
  assert(ACTK_MODREF_STATE(m_old) == MODREF_STATE_WRITTEN);

  /* Assert: m_new is a shadow modref. */
  assert(ACTK_MODREF_BEHAV(m_new) == MODREF_BEHAV_SHADOW);

  /* Replacement semantics: */
  /* -- update old shadow with new target */
  /* -- write new target with old value */
  {
    actk_modref_t* new_target = m_new->u.shadow.target;
    actk_val_t        old_val = m_old->u.shadow.val;
    
    actk_write(st, new_target, old_val);
    m_old->u.shadow.target = new_target;
  }
}
