/* Matthew Hammer <hammer@tti-c.org> */

/* inline-funtions (macros, sort of) to stuff flag bits in and out of
   the LSBs of pointers.  The point of this little module is to try to
   keep all the dirty bit-stuffing stuff somewhat abstract so that as
   we try to compile for different machines or different environments
   (with different alignments) we can keep all the dirty stuff in
   (somewhat) higher-level terms: a pointer with flags is to be cast
   and bit-juggled only though this "API".
   
   -- ptr_get(p) -- give the pointer-part of p (with any flags masked out)

   -- ptr_set(p) -- set the pointer-part of p (with flags left unchanged)

   -- ptr_getflags(p) -- give the flags stored in p (all higher pointer bits masked out)

   -- ptr_setflags(p, bits) -- set the flag bits of the pointer,

     -- To toggle a single bit on:
             ptr_setflags(p, ptr_getflags(p) | bit)

     -- To toggle a single bit off:
             ptr_setflags(p, ptr_getflags(p) & ~bit)
       (or)  ptr_setflags(p, ptr_getflags(p) - bit)      
*/

#ifndef __POINTER_H__
#define __POINTER_H__

#include "actk_directives.h"
#include <assert.h>

typedef void* actk_ptr_t;
typedef unsigned int actk_ptrbits_t;

ACTK_INLINE static actk_ptr_t ptr_get(actk_ptr_t p);
ACTK_INLINE static void ptr_set(actk_ptr_t* p, actk_ptr_t p_);

ACTK_INLINE static actk_ptrbits_t ptr_getflags(actk_ptr_t p);
ACTK_INLINE static void ptr_setflags(actk_ptr_t* p, actk_ptrbits_t bits);

ACTK_INLINE static actk_ptr_t ptr_addflags(actk_ptr_t p, actk_ptrbits_t bits);

/* --- IMPLEMENTATION -- */  

#define ACTK_PTR_MASK_PTR   ((actk_ptrbits_t)0x3)
#define ACTK_PTR_MASK_FLAGS (~ACTK_PTR_MASK_PTR)

ACTK_INLINE static
actk_ptr_t ptr_get(actk_ptr_t p) {
  return (actk_ptr_t) (((actk_ptrbits_t) p) & ACTK_PTR_MASK_FLAGS);
}

ACTK_INLINE static
actk_ptrbits_t ptr_getflags(actk_ptr_t p) {
  return ((actk_ptrbits_t) p) & ACTK_PTR_MASK_PTR;
}


ACTK_INLINE static
actk_ptr_t ptr_addflags(actk_ptr_t p, actk_ptrbits_t bits) {
  return (actk_ptr_t) (((actk_ptrbits_t) p & ACTK_PTR_MASK_FLAGS) | bits);
}

ACTK_INLINE static
void ptr_setflags(actk_ptr_t* p, actk_ptrbits_t bits) {
  assert(bits <= ACTK_PTR_MASK_PTR);
  *p = ptr_addflags(*p, bits);
}

ACTK_INLINE static
void ptr_set(actk_ptr_t* p, actk_ptr_t p_) {
  actk_ptrbits_t flags = ptr_getflags(*p);
  *p = p_;
  ptr_setflags(p, flags);
}

#endif
