#include "actk_types.h"
#include "closure_impl.h"
#include "state_impl.h"
#include "logging.h"
#include "actk_debug.h"
#include "allocation2.h"
#include "propagation.h"
#include "stats.h"
#include "actk_debug.h"
#include "actk_macros.h"

#include <assert.h>

static void actk__commit_and_jump(actk_state_t* st, actk_timestamp_t* to);
static void actk__propagate_rec(actk_state_t* st);

void actk__chpctx_init(actk_state_t* st) {
  assert(!st->chpctx);
  assert(st->flags & ACTK_BOOKKEEPING);
  logcatf(st,"change-propagation: initialization");
  actk_print_interval(st, st->toplevel->interval, "initial-run");
  st->chpctx = (actk_chpctx_t*) actk_libc_malloc(sizeof(actk_chpctx_t));
  st->chpctx->queue = (readqueue_t*) actk_libc_malloc(sizeof(readqueue_t));
  readqueue_init(st->chpctx->queue);

  st->chpctx->end = NULL;
}

void actk__chpctx_clear(actk_state_t* st) {
  assert(st->chpctx);
  assert(st->flags & ACTK_BOOKKEEPING);
  actk_alloc_freeall(st);
  totalorder_remove(st->toplevel->interval.start,
                    st->toplevel->interval.end,
                    (to_data_t) st,
                    (to_callback_t) actk__closure_invalidate);
  actk_alloc_freeall(st);
  
  readqueue_cleanup(st->chpctx->queue);
  actk_libc_free(st->chpctx->queue);
  actk_libc_free(st->chpctx);
  st->chpctx = NULL;
}

void actk_freeall(actk_state_t* st) {
  assert(st->flags & ACTK_BOOKKEEPING);
  assert(st->closure == st->toplevel);

  if(!st->chpctx) {
    /* No changes yet */
    return;
  }
  else {  
    assert(readqueue_isempty(st->chpctx->queue));
    actk_alloc_dostats(st);
    actk_alloc_freeall(st);
  }
}


void actk__chpctx_remove_closure(actk_state_t* st, actk_closure_t* closure) {
  assert(st->chpctx);
  if(closure->readqueue_pos >= 0) {
    logcatf(st,"removing " DBSTR_CLOSURE " from queue", DBVAS_CLOSURE(closure));
    readqueue_remove(st->chpctx->queue, closure->readqueue_pos);
  }
}

void actk_propagate(actk_state_t* st) {
  IF_DEBUG( static int next_num = 1);
  IF_DEBUG( int num = next_num++ );
  actk_stats_t stats;
  
  assert(st->flags & ACTK_BOOKKEEPING);
  assert(st->closure == st->toplevel);

  /* Update the maxlive stats before each propagation. */
  actk_alloc_dostats(st);

  if(!st->chpctx) {
    /* No changes yet */
    return;
  }
  
  log_inc_major_num();
  actk__stats_incr(st->stats, PROP);
  if(st->flags & ACTK_PRINT_PROP_STATS)
    actk__stats_save(&stats, st->stats);
  
  /* actk_verify_interval_closures(st, st->toplevel->interval); */

  /* Setup the change-propagation interval as the top-level interval. */
  st->now         = st->toplevel->interval.start;
  st->chpctx->end = st->toplevel->interval.end;

  /* Do the change-propagation */
  actk__propagate_rec(st);

  /* Update the maxlive stats after each propagation. */
  actk_alloc_dostats(st);

  actk_print_interval(st, st->toplevel->interval, "propagation-%d", num);
  /* actk_verify_interval_closures(st, st->toplevel->interval); */
  
  log_inc_major_num();
  actk_print_interval(st, st->toplevel->interval, "propagation-%d", num);
  if(st->flags & ACTK_PRINT_PROP_STATS) {
    actk__stats_diff(&stats, st->stats, &stats);
    actk__stats_print(&stats, stderr);
  }  
  actk__closure_enter(st, st->toplevel, ACTK_CLOSURE_ENTER_NEVER);
}

void actk__propagate_rec(actk_state_t* st) {
  readqueue_t* q;
  IF_DEBUG( static int next_num = 1);
  IF_DEBUG( int num = next_num++ );
  
  assert(st->chpctx);
  assert(st->chpctx->queue);
  
  q = st->chpctx->queue;
  log_inc_minor_num();
  actk__stats_incr(st->stats, PROPREC);
  logcatf(st,"starting propagation (%d)", num);
  logcatf(st, " in " DBSTR_STATE, DBVAS_STATE(st));

  while(!readqueue_isempty(q) &&
        (totalorder_compare( readqueue_peek(q)->interval.start,
                             st->chpctx->end) < 0) )
  {
    actk_closure_t* closure = readqueue_pop(q);
    
    actk__stats_incr(st->stats, PROP_QUEUEPOP);
    logcatf(st,"next closure off queue is " DBSTR_CLOSURE "", DBVAS_CLOSURE(closure));
    
    /* FACT: the next closure is in our change-propagation context */
    assert(totalorder_compare(st->now, closure->interval.start) <= 0);
    assert(totalorder_compare(closure->interval.start, st->chpctx->end) < 0);
    assert(closure->color == ACTK_CLOSURE_GREY);

    actk__closure_wakeup(st, closure);
    
    actk__closure_enter(st, closure, ACTK_CLOSURE_ENTER_PROPAGATE);
    actk__commit_and_jump(st, actk__closure_get_end(st, closure));
    
    logcatf(st, "now in " DBSTR_STATE, DBVAS_STATE(st));
  }
  logcatf(st,"propagation complete (%d)", num);
}


void actk__reuse_run(actk_state_t* st, actk_closure_t* closure) {
  /* Save the incoming chpctx interval for later restoration. */
  /* actk_interval_t current_interval_0 = st->current_interval; */
  /* actk_interval_t chpctx_interval_0  = st->chpctx->interval; */
  actk_closure_t* caller_closure = st->closure;
  actk_timestamp_t* chpctx_end_0 = st->chpctx->end;
  int print_interval = 1;
  
  /* logging: */
  logcatf(st,TERM_HIGHLIGHT_2 "RE-USING: " DBSTR_CLOSURE TERM_NORM "",
          DBVAS_CLOSURE(closure));

  logcatf(st, "  it's interval is [%p,%p)",
          closure->interval.start,
          actk__closure_get_end(st, closure));

  logcatf(st, "  it's recctx is %p", closure->recctx);
  
  if(print_interval)
    actk_print_interval(st, st->toplevel->interval, "reuse-a");

  /* 1. drop everything between "now" and the start of the reused closure */
  actk__commit_and_jump(st, closure->interval.start);

  /* logging: */
  if(print_interval)
    actk_print_interval(st, st->toplevel->interval, "reuse-b");
  
  /* 2. current chpctx interval := closure's interval */
  st->chpctx->end = actk__closure_get_end(st, closure);
  assert(st->chpctx->end);
  
  /* 3. recursively run changepropagation */
  actk__propagate_rec(st);

  /* logging: */
  if(print_interval)
    actk_print_interval(st, st->toplevel->interval, "reuse-c");
  
  /* 4. Restore old interval's end-time */
  st->chpctx->end = chpctx_end_0;

  /* 5. Restore the caller closure */
  st->closure     = caller_closure;
  st->now         = actk__closure_get_end(st, closure);
  st->fun_loop_on = 0;

  if(print_interval)
    actk_print_interval(st, st->toplevel->interval, "reuse-d");
  logcatf(st,"END: REUSE of: " DBSTR_CLOSURE TERM_NORM "", DBVAS_CLOSURE(closure));
  logcatf(st,"caller was : " DBSTR_CLOSURE, DBVAS_CLOSURE(caller_closure));
}

void actk__commit_and_jump(actk_state_t* st, actk_timestamp_t* jump_to)
{
  actk_interval_t J = {st->now, jump_to};
  actk_interval_t __attribute__((unused)) C = {jump_to, st->chpctx->end};

  logcatf(st,"         jump = " DBSTR_INTERVAL, DBVAS_INTERVAL(J));
  logcatf(st,"  changes rem = " DBSTR_INTERVAL, DBVAS_INTERVAL(C));
    
  /* Splice-out items being jumped over (non-inclusive) */
  totalorder_remove(J.start, J.end,
                    (to_data_t) st,
                    (to_callback_t) actk__closure_invalidate);

  /* Move ahead "now" to jump target */
  st->now = jump_to;
}


/* Antiquated. */
void actk__trim_queue(actk_state_t* st, actk_timestamp_t* jump_time) {
  /* Sept 24 2007 -- This function should no longer be needed since we
     now actively remove closures from the change-propagation queue
     when they are invalidated. */
  
/*   /\* Update chpctx queue: remove all reads from the queue that come */
/*      before or at the given timestamp -- this ensures that the queue */
/*      contains no dangling pointers to invalidated reads, which are */
/*      otherwise freed. *\/ */
/*   { */
/*     readqueue_t* q = st->chpctx->queue; */
/*     while(!readqueue_isempty(q) && */
/*           (totalorder_compare(readqueue_peek(q)->interval.start, jump_time) < 0)) { */
/*       logcatf(st,"popping " DBSTR_CLOSURE " from chpctx queue", */
/*               DBVAS_CLOSURE(readqueue_peek(q))); */
/*       actk_closure_t* closure = readqueue_pop(q); */
/*       assert(closure->color == ACTK_CLOSURE_GREY); */
/*       closure->color = ACTK_CLOSURE_BLACK; */
/*     } */
/*   } */
}
