#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "actk.h"
#include "actk_types.h"
#include "state_impl.h"
#include "actk_macros.h"
#include "stats.h"
#include "propagation.h"
#include "closure_impl.h"
#include "actk_debug.h"
#include "logging.h"

static void actk_args(actk_state_t* st, int argc, char** argv) {
  {int i;
  for(i = 1; i < argc; i++) {
    if(!argv[i]) {
      fprintf(stderr, "actk: WARNING: arg %d is NULL\n", i);
    }
    else if(!strcmp(argv[i], "-actk-log-stderr"))    { st->flags |=  ACTK_LOG_STDERR;  }
    else if(!strcmp(argv[i], "+actk-log-stderr"))    { st->flags &= ~ACTK_LOG_STDERR;  }
    else if(!strcmp(argv[i], "-actk-log-intervals")) { st->flags |=  ACTK_LOG_INTERVALS | ACTK_RM_INTERVALS; }
    else if(!strcmp(argv[i], "+actk-log-intervals")) { st->flags &= ~ACTK_LOG_INTERVALS;  }
    else if(!strcmp(argv[i], "-actk-log-file"))      { st->flags |=  ACTK_LOG_FILE; }
    else if(!strcmp(argv[i], "+actk-log-file"))      { st->flags &= ~ACTK_LOG_FILE; }
    else if(!strcmp(argv[i], "-actk-prop-stats"))    { st->flags |=  ACTK_PRINT_PROP_STATS; }
    else if(!strcmp(argv[i], "+actk-prop-stats"))    { st->flags &= ~ACTK_PRINT_PROP_STATS; }
    else if(!strcmp(argv[i], "-actk-rm-intervals"))  { st->flags |=  ACTK_RM_INTERVALS; }
    else if(!strcmp(argv[i], "+actk-rm-intervals"))  { st->flags &= ~ACTK_RM_INTERVALS; }
    else if(!strcmp(argv[i], "-actk-tailrecursion")) { st->flags |=  ACTK_TAILRECURSION; }
    else if(!strcmp(argv[i], "+actk-tailrecursion")) { st->flags &= ~ACTK_TAILRECURSION; }
    else if(!strcmp(argv[i], "-actk-bookkeeping"))   { st->flags |=  ACTK_BOOKKEEPING; }
    else if(!strcmp(argv[i], "+actk-bookkeeping"))   { st->flags &= ~ACTK_BOOKKEEPING; }
    else if(!strcmp(argv[i], "-actk-test-assert"))   {
      int asserts_active = 0;
      assert((asserts_active = 1, fprintf(stderr, "actk: NOTE: asserts are active\n")));
      if(!asserts_active) {
        fprintf(stderr, "actk: ERROR: asserts are not active in this binary\n");
        abort();
      }
    }
    else {
      fprintf(stderr, "actk: WARNING: argument not recognized: %s\n", argv[i]);
      exit(EXIT_FAILURE);
    }
  }}
}

ACTK_FUN_SIG(actk__toplevel, "toplevel", "");
ACTK_FUN(actk__toplevel, actk__toplevel_, 0);

void actk__toplevel_(actk_state_t* st) {
  assert(!("should never actually be called."));
}

void actk_init(actk_state_t* st, int argc, char** argv) {  
  /* Flags: setup the flags from defaults and user-given arguments */ {
    st->flags = ( ACTK_TAILRECURSION |
                  ACTK_BOOKKEEPING );
    actk_args(st, argc, argv);
  }
  
  actk__remove_interval_files(st);
  st->stats = actk__stats_init(actk_libc_malloc(sizeof(actk_stats_t)));  
  actk_freeset_init();
  actk_heap_set_stats(st->stats);
  st->chpctx = NULL;
  doublink_init(&st->garb_set);
  
  if(st->flags & ACTK_BOOKKEEPING) {
    /* Toplevel: Create a default ("top-level") closure for
       initialization -- that way no 'owner' fields are ever NULL and
       the actk state always has a non-NULL closure. Toplevel also
       records the absolute starting and ending times of the
       totalorder. */
    st->now = totalorder_create((to_data_t) NULL);
    st->toplevel = actk__closure_new(st, actk__toplevel, NULL, 0, 0, 0);
    st->closure = st->toplevel;
    st->toplevel->recctx = actk__recctx_new(st, actk__toplevel, 0, 0);
    actk__closure_init(st, st->toplevel);
    actk__closure_enter(st, st->toplevel, ACTK_CLOSURE_ENTER_NEVER);
    st->toplevel->interval.end =
      totalorder_insert(st->toplevel->interval.start, actk_timestamp_nulldata);    
  } else {
    st->toplevel = actk__closure_new(st, actk__toplevel, NULL, 0, 0, 0);
    st->closure = st->toplevel;
    st->toplevel->recctx = actk__recctx_new(st, actk__toplevel, 0, 0);
  }
}

void actk_clear(actk_state_t* st) {
  if(st->flags & ACTK_BOOKKEEPING) {
    if(st->chpctx)
      actk__chpctx_clear(st);    
    
    /* restart -- look at actk_init() for most of the initial values we
       reuse here: */
    st->closure = NULL;
    actk__closure_enter(st, st->toplevel, ACTK_CLOSURE_ENTER_NEVER);
  }
}

void actk_cleanup(actk_state_t* st) {
  actk_clear(st);
}

void actk_garbage(actk_state_t* st, void* p) {
  actk_alloc_garb(st, p);
}

void actk_run(actk_state_t* st, actk_closure_t* closure) {
  /* First, if we have a memotable available and we are in
     change-propagation mode try to do a memo-match.  If we aren't in
     change-prop mode but we have a memotable, insert a new entry for
     this read. */  
  if((st->flags & ACTK_BOOKKEEPING) &&
     (closure->recctx->fun->flags & ACTK_USEMEMOTABLE))
  {
    memotbl_t* memotbl = &closure->recctx->memotbl;
    
    if(st->chpctx) {
      actk_closure_t* closure0 = actk_memo_match(st, memotbl, closure);
      
      if(closure0) {
        actk__closure_replace(st, closure0, closure);
        actk__reuse_run(st, closure0);
        return;
      }
    }

    /* If no match, insert it */
    memotbl_insert(memotbl, (doublink_node_t*) closure);
  }
  
  /* In the ordinary case, initialize and enter the read. */
  actk__closure_init(st, closure);
  actk__closure_enter(st, closure, ACTK_CLOSURE_ENTER_AUTO);
  return;
}
