#ifndef __EXPRTREE_H__
#define __EXPRTREE_H__

#include "actk.h"
#include <stdio.h>

/**
 * \defgroup actk_stdlib_exprtree Expression Trees
 * \ingroup actk_stdlib
 * @{
 */

typedef actk_modref_t exprtree_t;

typedef enum exprtree_op_e {
  EXPRTREE_OP_1, /* Unary ops */
  EXPRTREE_OP_INC,
  EXPRTREE_OP_DEC,
  EXPRTREE_OP_NEG,
  EXPRTREE_OP_2, /* Binary ops */
  EXPRTREE_OP_ADD,
  EXPRTREE_OP_DIV,
  EXPRTREE_OP_SUB,
  EXPRTREE_OP_MUL,
  EXPRTREE_OP_3, /* Ternery ops */

  /* Note: EXPRTREE_OP_N where N is a positive integer are all only
     used internally and aren't actually operators. */
} exprtree_op_t;

#define EXPRTREE_OP_1_COUNT (EXPRTREE_OP_2 - EXPRTREE_OP_1 - 1)
#define EXPRTREE_OP_2_COUNT (EXPRTREE_OP_3 - EXPRTREE_OP_2 - 1)

#define EXPRTREE_OP_1_TEST(op) (op > EXPRTREE_OP_1 && op < EXPRTREE_OP_2)
#define EXPRTREE_OP_2_TEST(op) (op > EXPRTREE_OP_2 && op < EXPRTREE_OP_3)

typedef enum exprtree_tag_e {
  EXPRTREE_OP,
  EXPRTREE_INT,
  EXPRTREE_ERR,
} exprtree_tag_t;

typedef union exprtree_data_u {
  int           intval;
  exprtree_op_t op;
} exprtree_data_t;

typedef struct exprtree_node_s {
  exprtree_tag_t tag;
  exprtree_data_t u;
  exprtree_t* children[0];
} exprtree_node_t;

/* Node Introduction-forms: */
exprtree_node_t* exprtree_int(actk_state_t* st, int val);
exprtree_node_t* exprtree_err(actk_state_t* st);
exprtree_node_t* exprtree_op(actk_state_t* st, exprtree_op_t op);

exprtree_node_t* exprtree_leaf(actk_state_t* st,
                               exprtree_node_t* key);

exprtree_node_t* exprtree_node_empty(actk_state_t* st, int nchildren);

/* Tree generation: */
exprtree_t* exprtree_random(actk_state_t* st, int num_nodes);
int         exprtree_random_rec(actk_state_t* st, int num_nodes, actk_modref_t* dest);

/* Pretty-printing: */
void exprtree_print(actk_state_t* st, FILE* f, exprtree_t* e);

/* Utility functions: */
int         exprtree_nchildren(actk_state_t* st, exprtree_node_t* node);
int         exprtree_isnormalized(actk_state_t* st, exprtree_node_t* node);
int         exprtree_op_nchildren(actk_state_t* st, exprtree_op_t op);
const char* exprtree_op_string(actk_state_t* st, exprtree_op_t op);

ACTK_FUN_SIG(exprtree_eval, "eval",
             "evaluate (normalize) a given expression tree, write the normalized form to dst",
             ACTK_FUN_ARG("src", exprtree_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("dst", exprtree_t*,      CONST_VAL)
             );

/* @} */
             
#endif
