 /* Matthew Hammer  */
/* <hammer@tti-c.org> */

#ifndef __MODLIST_H__
#define __MODLIST_H__

#include "actk.h"
#include <stdio.h>

/**
 * \defgroup actk_stdlib_modlist Modifiable Lists
 * \ingroup actk_stdlib
 * @{
 */

typedef actk_modref_t modlist_t;

typedef struct modlist_cell_s {
  actk_val_t head;
  modlist_t* tail;
} modlist_cell_t;

typedef struct modlist_pair_s {
  actk_val_t fst;
  actk_val_t snd;
} modlist_pair_t;

typedef actk_val_t (*modlist_generator_t)      (actk_state_t* st, actk_val_t i, actk_val_t* next_i);
typedef actk_val_t (*modlist_cell_mapf_t)      (actk_state_t* st, actk_val_t env, modlist_cell_t* c);
typedef int        (*modlist_cell_pred_t)      (actk_state_t* st, actk_val_t env, modlist_cell_t* c);
typedef int        (*modlist_cell_bin_pred_t)  (actk_state_t* st, modlist_cell_t* c1, modlist_cell_t* c2);
typedef actk_val_t (*modlist_val_combine_op_t) (actk_state_t* st, actk_val_t env, actk_val_t v1, actk_val_t v2);
typedef void       (*modlist_val_printf_t)     (actk_state_t* st, FILE* f, actk_val_t v);


#define modlist_nil_cell NULL

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Allocation functions */
modlist_cell_t* modlist_cell(actk_state_t* st, actk_val_t head);
modlist_cell_t* modlist_cell_keyed(actk_state_t* st, actk_val_t head);
modlist_cell_t* modlist_cell_empty(actk_state_t* st);
modlist_t*      modlist_nil(actk_state_t* st);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Top-Level API */
modlist_t*  modlist_generate(actk_state_t* st, modlist_generator_t g, actk_val_t start, actk_val_t stop);
void        modlist_print(actk_state_t* st, FILE* f, modlist_t* l, modlist_val_printf_t val_printf);
modlist_t*  modlist_get_nth(actk_state_t* st, modlist_t* l, int index);
actk_val_t  modlist_replace(actk_state_t* st, modlist_t* l, actk_val_t newval);
void        modlist_insert(actk_state_t* st, modlist_t* l, actk_val_t newval);
actk_val_t  modlist_remove(actk_state_t* st, modlist_t* l);

void modlist_print_hi(actk_state_t* st, FILE* out, modlist_t* l,
                      modlist_val_printf_t val_printf,
                      int hi_begin, int hi_end);

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Adaptive API */
ACTK_FUN_SIG(modlist_cons, "cons",
             "create a cell (with an empty tail) given a head",
             ACTK_FUN_ARG("head", actk_val_t,  READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",    modlist_t*,  CONST_DST)
             );
             
ACTK_FUN_SIG(modlist_one, "one",
             "create a cell with nil-tail given a head",
             ACTK_FUN_ARG("head", actk_val_t,  READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",    modlist_t*,  CONST_DST)
             );

ACTK_FUN_SIG(modlist_car, "car",
             "put the car (the head) of list l into the given destination",
             ACTK_FUN_ARG("l", actk_val_t,      READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d", modlist_cell_t*, CONST_DST)
             );

/* Note on CONS and CAR: these operations are traditionally thought of
 * as duals (though CAR is usually paired with CDR, ignore that for
 * now), in the sense that CONS adds a value to a list at it's head
 * and CAR gets the head of a list -- here the definitions are altered
 * for the setting when the head value being CONS'd or the list being
 * CAR'd are subject to certain changes.
 *
 * CONS is used for "self-adjusting initialization" of a list cell
 * where the new head must be read before it can be included in the
 * list -- likewise, CAR is used to "adaptively" move the head of a
 * list into a modref.
*/

ACTK_FUN_SIG(modlist_map, "map",
             "map a list l, using map function (mapf mapf_env), into d",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          LIFTD_DST),
             ACTK_FUN_ARG("mapf",     modlist_cell_mapf_t,     ENVIR_VAL),
             ACTK_FUN_ARG("mapf_env", actk_val_t,              ENVIR_VAL)
             );

ACTK_FUN_SIG(modlist_combine, "combine",
             "combine a list into a single element using (associative) op",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("op",       actk_val_combine_op_t,   ENVIR_VAL),
             ACTK_FUN_ARG("op_env",   actk_val_t,              ENVIR_VAL),
             );

ACTK_FUN_SIG(modlist_combine_rec, "combine_rec",
             "combine a list into a single element using (associative) op. test is 1"
             "iff the list has 1 element, zero otherwise.  The length of the list"
             "is assumed to be >= 1.",
             ACTK_FUN_ARG("test",     int,                     READ_LOC_OR_VAL),
             ACTK_FUN_ARG("l",        modlist_cell_t*,         CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("op",       actk_val_combine_op_t,   ENVIR_VAL),
             ACTK_FUN_ARG("op_env",   actk_val_t,              ENVIR_VAL),
             );

ACTK_FUN_SIG(modlist_fold, "fold",
             "fold a list into a single element using (associative) op",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("op",       actk_val_combine_op_t,   ENVIR_VAL),
             ACTK_FUN_ARG("op_env",   actk_val_t,              ENVIR_VAL)
             );
             
ACTK_FUN_SIG(modlist_reverse, "reverse",
             "reverse list l into d",
             ACTK_FUN_ARG("l",        ackt_modref_t*,          CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             );

ACTK_FUN_SIG(modlist_reverse_rec, "reverse_rec",
             "reverse list l into d using z as the reversed list tail",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("z",        modlist_cell_t*,         CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             );

ACTK_FUN_SIG(modlist_filter, "filter",
             "filter list l, using unary predicate (pred pred_env), into d",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          LIFTD_DST),
             ACTK_FUN_ARG("pred",     modlist_cell_pred_t,     ENVIR_VAL),
             ACTK_FUN_ARG("pred_env", actk_val_t,              ENVIR_VAL)
             );

ACTK_FUN_SIG(modlist_filter_sparse, "filter_sparse",
             "filter list l, using unary predicate (pred pred_env), into d",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("pred",     modlist_cell_pred_t,     ENVIR_VAL),
             ACTK_FUN_ARG("pred_env", actk_val_t,              ENVIR_VAL)
             );
             
ACTK_FUN_SIG(modlist_split, "split",
             "split list l, using unary predicate (pred pred_env), into d1 and d2",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d1",       actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("d2",       actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("pred",     modlist_cell_pred_t,     ENVIR_VAL),
             ACTK_FUN_ARG("pred_env", actk_val_t,              ENVIR_VAL)
             );

ACTK_FUN_SIG(modlist_merge, "merge",
             "merge lists l1 and l2, using binary predicate pred, into d",
             ACTK_FUN_ARG("l1",       modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("l2",       modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST),
             ACTK_FUN_ARG("pred",     modlist_cell_bin_pred_t, ENVIR_VAL)
             );
                          
ACTK_FUN_SIG(modlist_cross, "cross",
             "create of list of pairs (x,y) -- foreach x in l1, y in l2",
             ACTK_FUN_ARG("l1",       modlist_cell_t*,         CONST_VAL),
             ACTK_FUN_ARG("l2",       modlist_t*,              CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_VAL),
             );

ACTK_FUN_SIG(modlist_testlen, "testlen",
             "if length(l) >= n then write 1 into d, else write 0 into d",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("n",        int,                     CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          CONST_DST)
             );

ACTK_FUN_SIG(modlist_traverse, "traverse",
             "has no output (it isn't useful), used as a base-line in benchmarks",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL)
             );

/* @} */
#endif
