#include "bst.h"
#include "modutil.h"

/* Private stuff */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Since all the node-pointers are kept in modrefs, we use these
   accessors to help make the rest of the code more readable. */

static bst_node_t* root(bst_t* bst)
{ return (bst_node_t*) actk_deref(bst->actk_state, bst->root); }

static bst_node_t* parent(bst_t* bst, bst_node_t* n)
{ return (bst_node_t*) actk_deref(bst->actk_state, n->parent); }

static bst_node_t* left(bst_t* bst, bst_node_t* n)
{ return (bst_node_t*) actk_deref(bst->actk_state, n->left); }

static bst_node_t* right(bst_t* bst, bst_node_t* n)
{ return (bst_node_t*) actk_deref(bst->actk_state, n->right); }

/* Given a node (finger) get the modref that points to the finger from
   finger's parent (either the left or right pointer of the parent). */
static actk_modref_t* pointer_from_parent(bst_t* bst, bst_node_t* finger) {
  bst_node_t* p = parent(bst, finger);
  return (p ?
          (left(bst, p) == finger
           ? p->left
           : p->right )
          : bst->root);
}

/* Primary API Implementation */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void bst_init(actk_state_t* st, bst_t* bst, bst_compare_t comp) {
  bst->actk_state = st;
  bst->root       = actk_modref(st);
  bst->comp       = comp;
  bst->initverif  = 0;
  actk_write(st, bst->root, (actk_val_t) NULL);
}

/* Either find a given key in the tree and return it, or return NULL if not found. */
void* bst_find(bst_t* bst, void* key) {
  bst_node_t* finger = bst_finger(bst, key);
  if(bst->comp(finger->key, key) == 0)
    return finger->key;
  else
    return NULL;
}

/* Return 0 or 1 depending on whether the given key is in the tree */
int bst_ismemb(bst_t* bst, void* key) {
  bst_node_t* finger = bst_finger(bst, key);
  return (bst->comp(finger->key, key) == 0);
}

/* Given a destination (either the left or right pointer for the given
   parent) and a key, create and install a new leaf. */
static void
bst_leaf_insert(bst_t* bst, bst_node_t* parent, actk_modref_t* dest, void* key) {
  actk_state_t* st = bst->actk_state;
  bst_node_t* leaf;
  
  /* Insert a new leaf under the given dest: */
  leaf = actk_alloc(st, sizeof(bst_node_t));
  actk_write(st, dest, (actk_val_t) leaf);
  
  /* Initialize the leaf: */
  leaf->key    = key;
  leaf->parent = actk_modref(st); actk_write(st, leaf->parent, (actk_val_t) parent);
  leaf->left   = actk_modref(st); actk_write(st, leaf->left,   (actk_val_t) NULL);
  leaf->right  = actk_modref(st); actk_write(st, leaf->right,  (actk_val_t) NULL);
}

/* Insert a given key into the tree.  If the key exists, the tree will
   be unchanged and 0 will be returned, otherwise the new key will be
   installed as a leaf and 1 will be returned. */
int bst_insert(bst_t* bst, void* key) {
  int return_val = 0;
  bst_verify(bst);
  {
    bst_node_t* finger = bst_finger(bst, key);

    if(!finger) {
      /* No finger => no nodes at all, so make a root node */
      bst_leaf_insert(bst, NULL, bst->root, key);
    }
    else {
      int comp = bst->comp(key, finger->key);
      actk_modref_t* childptr = NULL;

      /* Determine if the leaf goes to the right or left */
      if(comp < 0) {
        childptr = finger->left;
      }
      else if(comp > 0) {
        childptr = finger->right;
      }
      else {
        /* nothing: key already in tree */
        goto done;        
      }
      
      bst_leaf_insert(bst, finger, childptr, key);
      return_val = 1;
    }
  }
 done:
  bst_verify(bst);
  return return_val;
}

/* Remove a given key from the tree.  Returns 1 when this is
   successful and 0 when the key isn't found. */
int bst_remove(bst_t* bst, void* key) {
  bst_node_t* finger = bst_finger(bst, key);
  if(finger && bst->comp(finger->key, key) == 0)
    return bst_fing_remove(bst, finger);
  else
    return 0;
}

/* Printing */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static void
bst_print_rec(bst_t* bst, bst_node_t* node, void* env, bst_print_t print, FILE* file) {
  fprintf(file, "(");
  if(node) {
    print(bst, env, node->key, file);
    if(left(bst,node) || right(bst,node)) {
      fprintf(file, " ");
      bst_print_rec(bst, left(bst, node), env, print, file);
      fprintf(file, " ");
      bst_print_rec(bst, right(bst, node), env, print, file);
    }
  }
  fprintf(file, ")");
}

void bst_print(bst_t* bst, void* env, bst_print_t print, FILE* file) {  
  bst_node_t* node = root(bst);
  bst_print_rec(bst, node, env, print, file);
}

/* Finger operations */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static bst_node_t*
bst_finger_rec(bst_t* bst, bst_node_t* finger, void* key) {
  bst_node_t* child = NULL;
  int comp;

  if(!finger)
    return NULL;

  comp = bst->comp(key, finger->key);
  if(comp < 0) {
    child = left(bst, finger);
  }
  else if(comp > 0) {
    child = right(bst, finger);
  }
  else {
    return finger;
  }

  if(!child) {
    return finger;
  }
  else {
    return bst_finger_rec(bst, child, key);
  }
}

bst_node_t*
bst_finger(bst_t* bst, void* key) {
  bst_verify(bst);
  return bst_finger_rec(bst, root(bst), key);
}

bst_node_t*
bst_fing_pred(bst_t* bst, bst_node_t* finger) {
  bst_verify(bst);
  {
    bst_node_t* pred_ = left(bst, finger);
    bst_node_t* pred  = NULL;
    
    /* initially, pred' := finger->left */
    /* Now move to the right as far as possible */
    while(pred_) {
      pred = pred_;
      pred_ = right(bst, pred_);
    }
    
    return pred;
  }
}

bst_node_t*
bst_fing_succ(bst_t* bst, bst_node_t* finger) {
  bst_verify(bst);
  {
    bst_node_t* succ_ = right(bst, finger);
    bst_node_t* succ  = NULL;
    
    while(succ_) {
      succ = succ_;
      succ_ = left(bst, succ_);
    }
    
    return succ;
  }
}


static int
bst_fing_remove_simple(bst_t* bst, bst_node_t* finger) {
  actk_state_t* st = bst->actk_state;
  bst_node_t*    p = parent(bst, finger);
  bst_node_t*    l = left(bst, finger);
  bst_node_t*    r = right(bst, finger);  
  actk_modref_t* z = pointer_from_parent(bst, finger);
  
  if(!l) {
    actk_write(st, z, (actk_val_t) r);
    if(r) {
      actk_write(st, r->parent, (actk_val_t) p);
    }
    return 1;
  }
  else if(!r) {
    actk_write(st, z, (actk_val_t) l);
    if(l) {
      actk_write(st, l->parent, (actk_val_t) p);
    }
    return 1;
  }
  else {
    return 0;
  }
}

/* Replace the given node (old_node) with a new node from the tree.
   The given node to replace may be the root. */
static void
bst_fing_replace(bst_t* bst, bst_node_t* old_node, bst_node_t* new_node) {
  actk_state_t* st = bst->actk_state;
  bst_node_t*    p = parent(bst, old_node);
  bst_node_t*    l = left(bst, old_node);
  bst_node_t*    r = right(bst, old_node);
  actk_modref_t* z = pointer_from_parent(bst, old_node);
  
  /* double-link with parent */
  actk_write(st, z, (actk_val_t) new_node);
  actk_write(st, new_node->parent, (actk_val_t) p);

  /* double-link with left-child */
  actk_write(st, new_node->left, (actk_val_t) l);
  if(l) actk_write(st, l->parent, (actk_val_t) new_node);

  /* double-link with right-child */
  actk_write(st, new_node->right, (actk_val_t) r);
  if(r) actk_write(st, r->parent, (actk_val_t) new_node);
}

/* Remove the given node (finger) from the tree.  Always succeeds and
   returns 1.  This works by trying to remove the node in a simple way
   if the node has less than two non-null children; otherwise, it
   finds a replacement with less than two non-null children (either
   succ or pred, a coin flip decides which) and replaces the node to
   remove (finger) with the replacement node, which is guarrenteed to
   have less than two non-null children.

   (See: http://en.wikipedia.org/wiki/Binary_search_tree#Deletion)
*/
int bst_fing_remove(bst_t* bst, bst_node_t* finger) {
  bst_verify(bst);
  {
    if(bst_fing_remove_simple(bst, finger))
    { /* done. */ }
    else
    {
      bst_node_t* replacement;
      int coin = rand() % 2;
      int __attribute__((unused)) removed = 0;
      
      if(coin) {
        replacement = bst_fing_pred(bst, finger);
      }
      else {
        replacement = bst_fing_succ(bst, finger);
      }
      
      removed = bst_fing_remove_simple(bst, replacement);
      assert(removed);      
      bst_fing_replace(bst, finger, replacement);
    }

    /* make the finger-node (removed) into garbage */
    {
      actk_state_t* st  = bst->actk_state;
      actk_garbage(st, finger);
      actk_garbage(st, finger->parent);
      actk_garbage(st, finger->left);
      actk_garbage(st, finger->right);
    }
  }
  bst_verify(bst);
  return 1;
}

/* Verify */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* Given a node, it's current parent, the comparison that should hold
   between the node and it's parent (either -1 or 1) and the bst
   itself, this will verify that the node and the node's descendents
   are in a valid configuartion. */
ACTK_FUN_SIG(bst_verify_rec, "verify",
             "verify that a given BST sub-tree is valid",
             ACTK_FUN_ARG("node",    bst_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("parent",  bst_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("comp",    int,         CONST_VAL),
             ACTK_FUN_ARG("bst",     bst_t*,      ENVIR_VAL),
             );

void bst_initverif(bst_t* bst) {

  /* This initial-run will verify the current tree.  All calls to
     bst_verify() will just "propagate" over this initial run.  i.e,
     when nodes are inserted or removed, this verification code will
     "wake-up" at those changed points in the tree. */
  actk_call(bst->actk_state, bst_verify_rec,
            actk_read(bst->root),
            actk_pass(NULL),
            actk_pass(0),
            actk_pass(bst),
            );
}

void bst_verify(bst_t* bst) {
  if(!bst->initverif)
    actk_propagate(bst->actk_state);
}

ACTK_FUN(bst_verify_rec,
         bst_verify_rec_,
         ACTK_USEMEMOTABLE |
         ACTK_USEALLOCTABLE
         );

void bst_verify_rec_(actk_state_t* st) {
  actk_fun_begin(st);
  bst_node_t* node   = actk_fun_arg(st);
  bst_node_t* parent = actk_fun_arg(st);
  int         comp   = (int) actk_fun_arg(st);
  bst_t*      bst    = actk_fun_arg(st);
  
  if(!node) {
    /* Ok -- empty-tree. */
  }
  else if(node) {    
    /* If there is a parent (no parent for root of tree) check invariants: */
    if(parent) {

      /* Ensure that the child is cyclically-linked to it's parent */
      actk_call(st, modutil_verifyeq,
                actk_read(node->parent),
                actk_pass(parent)
                );
      
      /* Check that the node compares to it's child in the proper way: */
      if( bst->comp(node->key, parent->key) != comp ) {
        abort();
      }
    }
    
    /* Recursively check each sub-tree: All nodes to left should
       compare to this node as -1, those to the right should compare
       as +1. */
    actk_call(st, bst_verify_rec,
              actk_read(node->left),
              actk_pass(node),
              actk_pass(-1)
              );

    actk_call(st, bst_verify_rec,
              actk_read(node->right),
              actk_pass(node),
              actk_pass(+1)
              );
  }
  actk_fun_end(st);
}
