#include "exprtree.h"

exprtree_node_t*
exprtree_node_empty(actk_state_t* st, int nchildren) {
  exprtree_node_t* node =
    actk_alloc(st, (sizeof(exprtree_node_t) +
                    sizeof(actk_modref_t*) * nchildren));  
  { int i; for(i = 0; i < nchildren; i++) {
    node->children[i] = actk_modref(st);
  }}
  return node;
}

exprtree_node_t* exprtree_int(actk_state_t* st, int val) {
  exprtree_node_t* node = exprtree_node_empty(st, 0);
  node->tag = EXPRTREE_INT;
  node->u.intval = val;
  return node;
}

exprtree_node_t* exprtree_err(actk_state_t* st) {
  exprtree_node_t* node = exprtree_node_empty(st, 0);
  node->tag = EXPRTREE_ERR;
  return node;
}

exprtree_node_t* exprtree_op(actk_state_t* st, exprtree_op_t op) {
  exprtree_node_t* node;    
  node = exprtree_node_empty(st, exprtree_op_nchildren(st, op));
  node->tag = EXPRTREE_OP;
  node->u.op = op;
  return node;
}

exprtree_node_t* exprtree_leaf(actk_state_t* st, exprtree_node_t* key) {
  exprtree_node_t* node =
    actk_alloc(st, sizeof(exprtree_node_t), key);
  return node;
}

int exprtree_random_rec(actk_state_t* st, int num_nodes, actk_modref_t* dest) {  
  if(num_nodes > 2) /* Make a binary node */ {
    int randomrange = num_nodes - 2;    
    int num_nodes_0 = 1 + (rand() % randomrange);
    int num_nodes_1 = 1 + (randomrange - num_nodes_0);
    exprtree_op_t op = (rand() % EXPRTREE_OP_2_COUNT) + EXPRTREE_OP_2 + 1;
    exprtree_node_t* node = exprtree_op(st, op);
    int leftover_0, leftover_1;
    
    actk_write(st, dest, (actk_val_t) node);
    leftover_0 = exprtree_random_rec(st, num_nodes_0, node->children[0]);
    leftover_1 = exprtree_random_rec(st, num_nodes_1 + leftover_0, node->children[1]);
    return leftover_1;
  }
  else if(num_nodes == 2) /* Make a unary node and give leftover of 1 */ {
    exprtree_random_rec(st, 1, dest);
    return 1;
  }
  else if(num_nodes == 1) /* Make a leaf */ {
    actk_write(st, dest, (actk_val_t) exprtree_int(st, rand() % 100));
    return 0;
  }
  else {
    abort();
  }
}

exprtree_t* exprtree_random(actk_state_t* st, int num_nodes) {
  actk_modref_t* modref = actk_modref(st);
  int __attribute__((unused)) leftover = exprtree_random_rec(st, num_nodes, modref);
  assert(leftover <= 1);
  return modref;
}

void exprtree_print(actk_state_t* st, FILE* f, exprtree_t* exprtree) {
  exprtree_node_t* node = (exprtree_node_t*) actk_deref(st, exprtree);

  switch(node->tag) {
  case EXPRTREE_OP: {
    const char* opstr = exprtree_op_string(st, node->u.op);
    int nchildren = exprtree_op_nchildren(st, node->u.op);    
    fprintf(f, "(%s ", opstr);
    {int i; for(i = 0; i < nchildren; i++) {
      exprtree_print(st, f, node->children[i]);
      if(i + 1 < nchildren) fprintf(f, " ");
    }}
    fprintf(f, ")");
    break;
  }
  case EXPRTREE_INT:
    fprintf(f, "%d", node->u.intval);
    break;
  case EXPRTREE_ERR:
    fprintf(f, "(error)");
    break;
  default:
    abort();

  }
}

int exprtree_nchildren(actk_state_t* st, exprtree_node_t* node) {
  switch(node->tag) {
  case EXPRTREE_INT:
  case EXPRTREE_ERR:
    return 0;
  case EXPRTREE_OP:
    return exprtree_op_nchildren(st, node->u.op);
  default:
    abort();
  }
}

int exprtree_op_nchildren(actk_state_t* st, exprtree_op_t op) {
  if(EXPRTREE_OP_1_TEST(op)) return 1;
  else if(EXPRTREE_OP_2_TEST(op)) return 2;
  else abort();
}

const char* exprtree_op_string(actk_state_t* st, exprtree_op_t op) {
  const char* opstr = NULL;
  switch(op) {
  case EXPRTREE_OP_INC: opstr = "++"; break;
  case EXPRTREE_OP_DEC: opstr = "--"; break;
  case EXPRTREE_OP_NEG: opstr = "~"; break;
  case EXPRTREE_OP_ADD: opstr = "+"; break;
  case EXPRTREE_OP_SUB: opstr = "-"; break;
  case EXPRTREE_OP_DIV: opstr = "/"; break;
  case EXPRTREE_OP_MUL: opstr = "*"; break;
  default:
    abort();
  }
  return opstr;
}

int exprtree_isnormalized(actk_state_t* st, exprtree_node_t* node) {
  switch(node->tag) {
  case EXPRTREE_OP:
    return 0;
  case EXPRTREE_INT:
  case EXPRTREE_ERR:
    return 1;
  default:
    abort();
  }
}


/* Self-adjusting functions */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ACTK_FUN_SIG(exprtree_op1, "op1",
             "perform a given unary operation, we assume that the operands have been normalized",
             ACTK_FUN_ARG("op",  exprtree_op_t,    CONST_VAL),
             ACTK_FUN_ARG("src", exprtree_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("dst", exprtree_t*,      CONST_DST)
             );

ACTK_FUN(exprtree_op1,
         exprtree_op1_,         
         0);

void exprtree_op1_(actk_state_t* st) {
  actk_fun_begin(st);
  exprtree_op_t    op = (exprtree_op_t) actk_fun_arg(st);
  exprtree_node_t* n1 = actk_fun_arg(st);
  exprtree_t*    dest = actk_fun_arg(st);

  exprtree_node_t* n2 =
    exprtree_leaf(st, n1);

  actk_write(st, dest, (actk_val_t)  n2);

  if(n1->tag == EXPRTREE_ERR) {
    n2->tag = EXPRTREE_ERR;
  }
  else {
    switch(op) {
    case EXPRTREE_OP_INC:
      n2->tag = EXPRTREE_INT;
      n2->u.intval = n1->u.intval + 1;
      break;
    case EXPRTREE_OP_DEC:
      n2->tag = EXPRTREE_INT;
      n2->u.intval = n1->u.intval - 1;
      break;
    case EXPRTREE_OP_NEG:
      n2->tag = EXPRTREE_INT;
      n2->u.intval = -(n1->u.intval);
      break;
    default:
      abort();
    }
  }
  actk_fun_end(st);
}

ACTK_FUN_SIG(exprtree_op2, "op2",
             "perform a given binary operation, we assume that the operands have been normalized",
             ACTK_FUN_ARG("op",   exprtree_op_t,    CONST_VAL),
             ACTK_FUN_ARG("src1", exprtree_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("src2", exprtree_node_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("dst",  exprtree_t*,      CONST_DST)
             );

ACTK_FUN(exprtree_op2,
         exprtree_op2_,
         0);

void exprtree_op2_(actk_state_t* st) {
  actk_fun_begin(st);
  exprtree_op_t    op = (exprtree_op_t) actk_fun_arg(st);
  exprtree_node_t* n1 = actk_fun_arg(st);
  exprtree_node_t* n2 = actk_fun_arg(st);
  exprtree_t*    dest = actk_fun_arg(st);

  exprtree_node_t* n3 =
    exprtree_leaf(st, n1);

  actk_write(st, dest, (actk_val_t) n3);
  
  if(n1->tag == EXPRTREE_ERR ||
     n2->tag == EXPRTREE_ERR) {
    n3->tag = EXPRTREE_ERR;
  }
  else {
    switch(op) {
    case EXPRTREE_OP_ADD:
      n3->tag = EXPRTREE_INT;
      n3->u.intval = n1->u.intval + n2->u.intval;
      break;
    case EXPRTREE_OP_SUB:
      n3->tag = EXPRTREE_INT;
      n3->u.intval = n1->u.intval - n2->u.intval;
      break;
    case EXPRTREE_OP_DIV:
      if(n2->u.intval) {
        n3->tag = EXPRTREE_INT;
        n3->u.intval = n1->u.intval / n2->u.intval;
      }
      else {
        n3->tag = EXPRTREE_ERR;
      }
      break;
    case EXPRTREE_OP_MUL:
      n3->tag = EXPRTREE_INT;
      n3->u.intval = n1->u.intval * n2->u.intval;
      break;
    default:
      abort();
    }
  }
  actk_fun_end(st);
}

ACTK_FUN(exprtree_eval,
         exprtree_eval_,
         ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void exprtree_eval_(actk_state_t* st) {
  actk_fun_begin(st);
  exprtree_node_t* node = actk_fun_arg(st);
  exprtree_t*      dest = actk_fun_arg(st);

  switch(node->tag) {
  case EXPRTREE_INT:
  case EXPRTREE_ERR:
    actk_write(st, dest, (actk_val_t) node);
    break;
  case EXPRTREE_OP: {
    int const max_children = 2;
    int nchildren = exprtree_op_nchildren(st, node->u.op);
    actk_modref_t* subtree_results[max_children];

    /* Step 1: recurse on each subtree, gather results in
       subtree_results[] */
    assert(nchildren <= max_children);
    {int i; for(i = 0; i < nchildren; i++) {
      subtree_results[i] = actk_modref(st);
      actk_call(st, exprtree_eval,
                actk_read(node->children[i]),
                actk_pass(subtree_results[i])
                );
    }}
    
    /* Step 2: use subtree_results[] to compute this result and have
       it written to dest */
    switch(nchildren) {
    case 1:
      actk_call(st, exprtree_op1,
                actk_pass(node->u.op),
                actk_read(subtree_results[0]),
                actk_pass(dest)
                );
      break;
    case 2:
      actk_call(st, exprtree_op2,
                actk_pass(node->u.op),
                actk_read(subtree_results[0]),
                actk_read(subtree_results[1]),
                actk_pass(dest)
                );
      break;
    default:
      abort();
    }
    break;}
  default:
    abort();
  }
  actk_fun_end(st);
}

