
/* CELL_SHARE: controls whether the space for cells (modlist_cell_t)
   and their tail (actk_modref_t) are allocated seperately or
   together */
#define CELL_SHARE 1

#include "modlist.h"
#include "modutil.h"

#if CELL_SHARE
#include "../core/modref_impl.h" /* <-- TODO: needed for sizeof(actk_modref_t) */
const int cell_share_size = sizeof(modlist_cell_t) + sizeof(actk_modref_t);
#endif

/* Introduction forms (allocation, etc.) */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

modlist_cell_t* modlist_cell_unkeyed(actk_state_t* st, actk_val_t head) {
#if CELL_SHARE
  modlist_cell_t* cell = actk_alloc(st, cell_share_size);
  cell->tail = actk_modref_init(st, (actk_modref_t*) (cell + 1));
  cell->head = head;
#else
  modlist_cell_t* cell = actk_alloc(st, sizeof(modlist_cell_t));
  actk_modref_t*  tail = actk_modref(st);
  cell->tail = tail;
  cell->head = head;
#endif  
  return cell;
}

modlist_cell_t* modlist_cell_keyed(actk_state_t* st, actk_val_t head) {
#if CELL_SHARE
  modlist_cell_t* cell = actk_alloc(st, cell_share_size, head);
  cell->tail = actk_modref_init(st, (actk_modref_t*) (cell + 1));
  cell->head = head;
#else
  modlist_cell_t* cell = actk_alloc(st, sizeof(modlist_cell_t));
  actk_modref_t*  tail = actk_modref(st,);
  cell->tail = tail;
  cell->head = head;
#endif  
  return cell;
}

modlist_cell_t* modlist_cell(actk_state_t* st, actk_val_t head) {
  return modlist_cell_keyed(st, head);
}

modlist_cell_t* modlist_cell_empty (actk_state_t* st) {
  return modlist_cell_unkeyed(st, 0);
}


modlist_t* modlist_nil(actk_state_t* st) {
  static modlist_t* nil = NULL;
  if(!nil) {
    nil = actk_modref(st);
    actk_write(st, nil, modlist_nil_cell);
  }
  return nil;
}


modlist_t* modlist_generate(actk_state_t* st,
                            modlist_generator_t g,
                            actk_val_t start, actk_val_t stop) {
  actk_val_t i = start;
  actk_modref_t* list = actk_modref(st);
  actk_modref_t* tail = list;
  
  while(i != stop) {
    actk_val_t head = g(st, i, &i);
    modlist_cell_t* cell = modlist_cell_empty(st);
    cell->head = head;
    actk_write(st, tail, cell);        
    tail = cell->tail;
  }
  actk_write(st, tail, modlist_nil_cell);
  return list;
}


modlist_t* modlist_get_nth(actk_state_t* st, modlist_t* l_, int index)
{
  modlist_t* l = l_;
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  while(index) {
    if(!c) { return NULL; }
    l = c->tail;
    c = (modlist_cell_t*) actk_deref(st, l);
    index --;    
  }
  return l;
}

/* Pretty-printing */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */


void modlist__print(actk_state_t* st, FILE* out, modlist_t* l, modlist_val_printf_t val_printf) {
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  fprintf(out, "[");
  while(c) {
    /* fprintf(out, "(%d=%x, %d=%p)", c->head, c->head, (actk_val_t) c->tail, c->tail); */
    /* fprintf(out, "%d=%x", c->head, c->head); */
    /*fprintf(out, "@%p: ", c);*/
    val_printf(st, out, c->head);
    c = (modlist_cell_t*) actk_deref(st, c->tail);
    if(c) fprintf(out, ", ");
  }
  fprintf(out, "]");
}

void modlist_print_hi(actk_state_t* st, FILE* out, modlist_t* l,
                       modlist_val_printf_t val_printf,
                      int hi_begin, int hi_end)
{
  int index = 0;
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  fprintf(out, "[");
  while(c) {
    int do_highlight = (index >= hi_begin && index <= hi_end);

    if(do_highlight)
      fprintf(out, "%s", TERM_HIGHLIGHT_1);

    val_printf(st, out, c->head);

    if(do_highlight)
      fprintf(out, "%s", TERM_NORM);

    c = (modlist_cell_t*) actk_deref(st, c->tail);
    if(c) fprintf(out, ", ");
    index ++;
  }
  fprintf(out, "]");
}

void modlist_print(actk_state_t* st, FILE* out, modlist_t* l, modlist_val_printf_t val_printf) {
#if 0
  char* buff = NULL;
  size_t buffsize = 0;
  FILE *mem = open_memstream(&buff, &buffsize);
#endif
  modlist__print(st, out, l, val_printf);
#if 0
  modlist__print(st, mem, l, val_printf);
  fclose(mem);
  
  logcatf(st, "list: %s", buff);
  free(buff);
#endif
}

/* Mutators */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

actk_val_t modlist_remove(actk_state_t* st, modlist_t* l) {
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  actk_val_t h = c->head;
  actk_write(st, l, actk_deref(st, c->tail));

#if CELL_SHARE
  actk_garbage(st, c);
#else
  actk_garbage(st, c->tail);
  actk_garbage(st, c);
#endif
  
  return h;
}

void modlist_insert(actk_state_t* st, modlist_t* l, actk_val_t v) {
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  modlist_cell_t* cc = modlist_cell(st, v);
  actk_write(st, l, cc);
  actk_write(st, cc->tail, c);
}

actk_val_t modlist_replace(actk_state_t* st, modlist_t* l, actk_val_t v) {
#if CELL_SHARE
  abort();
  return 0;
#else
  modlist_cell_t* c = (modlist_cell_t*) actk_deref(st, l);
  modlist_cell_t* cc = actk_alloc(st, sizeof(modlist_cell_t));
  actk_val_t oldhead = c->head;
  cc->head = v;
  cc->tail = c->tail;
  actk_garbage(st, c);
  actk_write(st, l, cc);
  return oldhead;
#endif
}

/* Self-adjusting functions */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

ACTK_FUN(modlist_cons, modlist_cons_, 0);

void modlist_cons_(actk_state_t* st) {
  actk_fun_begin(st);
  actk_val_t      head = actk_fun_arg(st);
  actk_modref_t*  dest = actk_fun_arg(st);
  modlist_cell_t* cell = modlist_cell(st, head);
  actk_write(st, dest, cell);
  actk_fun_end(st);
}

ACTK_FUN(modlist_one, modlist_one_, 0);

void modlist_one_(actk_state_t* st) {
  actk_fun_begin(st);
  actk_val_t      head = actk_fun_arg(st);
  actk_modref_t*  dest = actk_fun_arg(st);
  modlist_cell_t* cell = modlist_cell(st, head);
  actk_write(st, cell->tail, NULL);
  actk_write(st, dest, cell);
  actk_fun_end(st);
}

ACTK_FUN(modlist_car, modlist_car_, 0);

void modlist_car_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t* cell = actk_fun_arg(st);  
  actk_modref_t*  dest = actk_fun_arg(st);
  actk_write(st, dest, cell->head);
  actk_fun_end(st);
}

ACTK_FUN(modlist_map,
         modlist_map_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE
         );

void modlist_map_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell = actk_fun_arg(st);
  actk_modref_t*      dest = actk_fun_arg(st);
  modlist_cell_mapf_t mapf = actk_fun_arg(st);
  actk_val_t      mapf_env = (actk_val_t) actk_fun_arg(st);

  if(!cell) {
    actk_write(st, dest, modlist_nil_cell);
  }
  else {
    actk_val_t h_ = mapf(st, mapf_env, cell);
    modlist_cell_t* newcell = modlist_cell_unkeyed(st, h_);
    actk_write(st, dest, newcell);
    dest = newcell->tail;

    actk_call(st, modlist_map,
              actk_read(cell->tail),
              actk_pass(dest));    
  }  
  actk_fun_end(st);
}


ACTK_FUN_SIG(modlist_halve, "halve",
             "psuedo-randomly fold subsequences of list l using binop into d",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("mode",     actk_val_t,              CONST_VAL),
             ACTK_FUN_ARG("v",        actk_val_t,              CONST_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          LIFTD_DST),

             /* For combining elements: */
             ACTK_FUN_ARG("op",       actk_val_combine_op_t,   ENVIR_VAL),
             ACTK_FUN_ARG("op_env",   actk_val_t,              ENVIR_VAL),

             /* For "coin-flipping": */
             ACTK_FUN_ARG("pred",     actk_cell_pred_t,        ENVIR_VAL),
             ACTK_FUN_ARG("pred_env", actk_val_t,              ENVIR_VAL),
             );             

ACTK_FUN(modlist_halve,
         modlist_halve_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_halve_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell =              actk_fun_arg(st);
  actk_val_t          mode = (actk_val_t) actk_fun_arg(st);
  actk_val_t           val = (actk_val_t) actk_fun_arg(st);
  actk_modref_t*      dest =              actk_fun_arg(st);
  
  modlist_val_combine_op_t op =              actk_fun_arg(st);
  actk_val_t           op_env = (actk_val_t) actk_fun_arg(st);
  modlist_cell_pred_t    pred =              actk_fun_arg(st);
  actk_val_t         pred_env = (actk_val_t) actk_fun_arg(st);

  /* mode=1 => no val, so use cell->head and recur with mode=0. */
  /* mode=0 => have val, coin flip determines next mode */
  if(mode) {
    if(!cell) {
      actk_write(st, dest, modlist_nil_cell);
    }
    else {
      actk_call(st, modlist_halve,
                actk_read(cell->tail),
                actk_pass(0),
                actk_pass(cell->head),
                actk_pass(dest)
                );
    }
  }
  else {
    if(!cell) {
      modlist_cell_t* val_cell = modlist_cell(st, val);
      actk_write(st, val_cell->tail, modlist_nil_cell);
      actk_write(st, dest, val_cell);
    }
    else {
      actk_val_t new_val = op(st, op_env, val, cell->head);

      /* pred true  =>
           create a new cell with new val,
           dest <- new cell
           dest := tail of new cell,
           mode := 1
         pred false =>
           new val := op(val, head)
           mode := 0
      */
      if(pred(st, pred_env, cell) == 0) {
        modlist_cell_t* new_cell = modlist_cell(st, new_val);
        actk_write(st, dest, new_cell);
        
        actk_call(st, modlist_halve,
                  actk_read(cell->tail),
                  actk_pass(1), /* Toggle mode */
                  actk_pass(0), /* No val */
                  actk_pass(new_cell->tail)
                  );
      }
      else {
        actk_call(st, modlist_halve,
                  actk_read(cell->tail),
                  actk_pass(0),       /* Same mode */
                  actk_pass(new_val), /* New value */
                  actk_pass(dest)     /* Same dest */
                  );        
      }
    }
  }  
  actk_fun_end(st);
}

ACTK_FUN(modlist_combine_rec,
         modlist_combine_rec_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_combine_rec_(actk_state_t* st) {
  actk_fun_begin(st);
  int             length_test =        (int) actk_fun_arg(st);
  actk_modref_t*         list =              actk_fun_arg(st);
  actk_modref_t*         dest =              actk_fun_arg(st);
  modlist_val_combine_op_t op =              actk_fun_arg(st);
  actk_val_t           op_env = (actk_val_t) actk_fun_arg(st);
  
  if(length_test) {
    actk_call(st, modlist_car,
              actk_read(list),
              actk_pass(dest)
              );
  }
  else {
    actk_modref_t*  halved = actk_modref(st);
    actk_modref_t* lentest = actk_modref(st);
    actk_hash_t*      hash = actk_alloc(st, sizeof(actk_hash_t));
    
    actk_hash_init(hash, 4);
   
    actk_call(st, modlist_halve,
              actk_read(list),
              actk_pass(1),
              actk_pass(NULL),
              actk_pass(halved),
              actk_pass(op),
              actk_pass(op_env),
              actk_pass((modlist_cell_pred_t) actk_hash_ptr),
              actk_pass(hash)
              );

    actk_call(st, modlist_testlen,
              actk_read(halved),
              actk_pass(1),
              actk_pass(lentest)
              );

    actk_call(st, modlist_combine_rec,
              actk_read(lentest),
              actk_pass(halved),
              actk_pass(dest)
              );
  }
  actk_fun_end(st);
}

ACTK_FUN(modlist_combine,
         modlist_combine_,
         ACTK_USEALLOCTABLE | ACTK_USEMEMOTABLE
         );

void modlist_combine_(actk_state_t* st) {
  actk_fun_begin(st);
  actk_modref_t*         list = actk_fun_arg(st);
  actk_modref_t*         dest = actk_fun_arg(st);
  modlist_val_combine_op_t op = actk_fun_arg(st);
  actk_val_t           op_env = (actk_val_t) actk_fun_arg(st);
  
  actk_modref_t* lentest = actk_modref(st);

  actk_call(st, modlist_testlen,
            actk_read(list),
            actk_pass(1),
            actk_pass(lentest)
            );

  actk_call(st, modlist_combine_rec,
            actk_read(lentest),
            actk_pass(list),
            actk_pass(dest),
            actk_pass(op),
            actk_pass(op_env)
            );

  actk_fun_end(st);
}

ACTK_FUN(modlist_fold,
         modlist_fold_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_fold_(actk_state_t* st) {
  /* TODO */
}

ACTK_FUN(modlist_reverse_rec,
         modlist_reverse_rec_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_reverse_rec_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell = actk_fun_arg(st);
  modlist_cell_t*        z = actk_fun_arg(st);
  actk_modref_t*      dest = actk_fun_arg(st);

  if(!cell) {
    actk_write(st, dest, z);
  }
  else {
    modlist_cell_t* newcell = modlist_cell(st, cell->head);
    actk_write(st, newcell->tail, z);
    actk_call(st, modlist_reverse_rec,
              actk_read(cell->tail),
              actk_pass(newcell),
              actk_pass(dest));
  }
  actk_fun_end(st);
}


ACTK_FUN(modlist_reverse,
         modlist_reverse_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_reverse_(actk_state_t* st) {
  actk_fun_begin(st);
  actk_modref_t*      list = actk_fun_arg(st);
  actk_modref_t*      dest = actk_fun_arg(st);
  actk_call(st, modlist_reverse_rec,
            actk_read(list),
            actk_pass(modlist_nil_cell),
            actk_pass(dest)
            );
  actk_fun_end(st);
}


ACTK_FUN(modlist_filter,
         modlist_filter_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_filter_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell = actk_fun_arg(st);
  actk_modref_t*      dest = actk_fun_arg(st);
  modlist_cell_pred_t pred = actk_fun_arg(st);
  actk_val_t      pred_env = (actk_val_t) actk_fun_arg(st);

  if(!cell) {
    actk_write(st, dest, modlist_nil_cell);
  }
  else {
    if(pred(st, pred_env, cell)) {
      logcatf(st, "pred: true");
      modlist_cell_t* newcell = modlist_cell(st, cell->head);
      actk_write(st, dest, newcell);
      dest = newcell->tail;
    }
    else {
      logcatf(st, "pred: false");
      /* Nothing. */
    }
    
    actk_call(st, modlist_filter,
              actk_read(cell->tail),
              actk_pass(dest));
  }
  actk_fun_end(st);
}

ACTK_FUN(modlist_filter_sparse,
         modlist_filter_sparse_,
         ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_filter_sparse_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell = actk_fun_arg(st);
  actk_modref_t*      dest = actk_fun_arg(st);
  modlist_cell_pred_t pred = actk_fun_arg(st);
  actk_val_t      pred_env = (actk_val_t) actk_fun_arg(st);

  if(!cell) {
    actk_write(st, dest, modlist_nil_cell);
  }
  else {
    if(pred(st, pred_env, cell)) {
      modlist_cell_t* newcell = modlist_cell(st, cell->head);
      actk_write(st, dest, newcell);
            
      actk_call(st, modlist_filter_sparse,
                actk_read(cell->tail),
                actk_pass(newcell->tail));
    }
    else {
      actk_modref_t* filterd = actk_modref(st, cell->tail);
      
      actk_call(st, modlist_filter_sparse,
                actk_read(cell->tail),
                actk_pass(filterd));

      actk_call(st, modutil_copy,
                actk_read(filterd),
                actk_pass(dest));
    }
  }
  actk_fun_end(st);
}


ACTK_FUN(modlist_split,
         modlist_split_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_split_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*     cell = actk_fun_arg(st);
  actk_modref_t*     dest1 = actk_fun_arg(st);
  actk_modref_t*     dest2 = actk_fun_arg(st);
  modlist_cell_pred_t pred = actk_fun_arg(st);
  actk_val_t      pred_env = (actk_val_t) actk_fun_arg(st);
  
  if(!cell) {
    actk_write(st, dest1, modlist_nil_cell);
    actk_write(st, dest2, modlist_nil_cell);
  }
  else {
    modlist_cell_t* newcell = modlist_cell(st, cell->head);

    if(pred(st, pred_env, cell)) {
      actk_write(st, dest1, newcell);
      dest1 = newcell->tail;
    }
    else {
      actk_write(st, dest2, newcell);
      dest2 = newcell->tail;
    }
    
    actk_call(st, modlist_split,
              actk_read(cell->tail),
              actk_pass(dest1),
              actk_pass(dest2));
  }
  actk_fun_end(st);
}

ACTK_FUN(modlist_merge,
         modlist_merge_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_merge_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*        cell1 = actk_fun_arg(st);
  modlist_cell_t*        cell2 = actk_fun_arg(st);
  actk_modref_t*          dest = actk_fun_arg(st);
  modlist_cell_bin_pred_t pred = actk_fun_arg(st);

  if(!cell1) {
    actk_write(st, dest, cell2);
  }
  else if(!cell2) {
    actk_write(st, dest, cell1);
  }
  else {
    actk_closure_t* rec;
    modlist_cell_t* newcell;
    
    if(pred(st, cell1, cell2)) {
      newcell = modlist_cell(st, cell1->head);
      rec = actk_closure(st, modlist_merge,
                         actk_read(cell1->tail),
                         actk_pass(cell2),
                         actk_pass(newcell->tail));
    }
    else {
      newcell = modlist_cell(st, cell2->head);
      rec = actk_closure(st, modlist_merge,
                         actk_pass(cell1),
                         actk_read(cell2->tail),
                         actk_pass(newcell->tail));
    }
    actk_write(st, dest, newcell);
    actk_run(st, rec);
  }
  actk_fun_end(st);
}

ACTK_FUN_SIG(modlist_pairs, "pairs",
             "create of list of pairs (x,y) -- foreach y in l;"
             "the final modref is written to last",
             ACTK_FUN_ARG("l",        modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("d",        actk_modref_t*,          READ_LOC_OR_VAL),
             ACTK_FUN_ARG("x",        actk_val_t*,             ENVIR_VAL),
             ACTK_FUN_ARG("last",     actk_modref_t*,          ENVIR_VAL),
             );

ACTK_FUN(modlist_pairs,
         modlist_pairs_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_pairs_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*         cell = actk_fun_arg(st);
  actk_modref_t*          dest = actk_fun_arg(st);
  actk_val_t               val = (actk_val_t) actk_fun_arg(st);
  actk_modref_t*          last = actk_fun_arg(st);

  if(!cell) {
    actk_write(st, last, dest);
  }
  else {
    modlist_pair_t* pair    = actk_alloc(st, sizeof(modlist_pair_t), cell);
    modlist_cell_t* newcell = modlist_cell(st, (actk_val_t) pair);

    pair->fst = val;
    pair->snd = cell->head;
    
    actk_write(st, dest, newcell);
    
    actk_call(st, modlist_pairs,
              actk_read(cell->tail),
              actk_pass(newcell->tail)
              );
  }  
  actk_fun_end(st);
}

ACTK_FUN_SIG(modlist_cross_rec, "cross_rec",
             "",
             ACTK_FUN_ARG("l1",       modlist_cell_t*,         READ_LOC_OR_VAL),
             ACTK_FUN_ARG("l2",       modlist_t*,              CONST_VAL),
             ACTK_FUN_ARG("dd",       actk_modref_t*,          CONST_VAL),
             );

ACTK_FUN(modlist_cross_rec,
         modlist_cross_rec_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_cross_rec_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t*        cell1 = actk_fun_arg(st);
  actk_modref_t*         list2 = actk_fun_arg(st);
  actk_modref_t*      destdest = actk_fun_arg(st);

  if(!cell1) {
    actk_call(st, modutil_copy,
              actk_pass(modlist_nil_cell),
              actk_read(destdest),
              );
  }
  else {
    actk_modref_t* new_destdest = actk_modref(st, cell1->head);
    
    actk_call(st, modlist_pairs,
              actk_rkey(cell1->head),
              actk_read(list2),
              actk_read(destdest),
              actk_pass(cell1->head),
              actk_pass(new_destdest)
              );

    actk_call(st, modlist_cross_rec,
              actk_read(cell1->tail),
              actk_pass(list2),
              actk_pass(new_destdest)
              );
  }
  
  actk_fun_end(st);
}

ACTK_FUN(modlist_cross,
         modlist_cross_,
         ACTK_USEMEMOTABLE | ACTK_USEALLOCTABLE
         );

void modlist_cross_(actk_state_t* st) {
  actk_fun_begin(st);
  actk_modref_t*         list1 = actk_fun_arg(st);
  actk_modref_t*         list2 = actk_fun_arg(st);
  actk_modref_t*          dest = actk_fun_arg(st);

  actk_modref_t*      destdest = actk_modref(st);

  actk_write(st, destdest, dest);

  actk_call(st, modlist_cross_rec,
            actk_read(list1),
            actk_pass(list2),
            actk_pass(destdest)
            );
  
  actk_fun_end(st);
}

ACTK_FUN(modlist_traverse,
         modlist_traverse_,
         ACTK_TAILRECURSIVE | ACTK_USEMEMOTABLE
         );

void modlist_traverse_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t* cell = actk_fun_arg(st);
  if(!cell) {
    /* done. */
  }
  else {
    actk_call(st, modlist_traverse,
              actk_read(cell->tail)
              );
  }
  actk_fun_end(st);
}

ACTK_FUN(modlist_testlen,
         modlist_testlen_,
         ACTK_TAILRECURSIVE);

void modlist_testlen_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t* cell = actk_fun_arg(st);
  int                n = (int) actk_fun_arg(st);
  actk_modref_t*  dest = actk_fun_arg(st);

  if(!cell) {
    actk_write(st, dest, (actk_val_t) 1);
  }
  else if(!n) {
    actk_write(st, dest, (actk_val_t) 0);
  }
  else {
    actk_call(st, modlist_testlen,
              actk_read(cell->tail),                  
              actk_pass(n - 1),
              actk_pass(dest));
  }  
  actk_fun_end(st);
}
