#include "modlist.h"
#include "modutil.h"
#include "geom2d.h"
#include "quickhull.h"

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Wrappers for geometry functions */

static geom2d_point_t*
toleft(actk_state_t* st, void* _dummy,
       geom2d_point_t* p1,
       geom2d_point_t* p2) {
  return geom2d_toleft(p1,p2);
}

static geom2d_point_t*
toright(actk_state_t* st, void* _dummy,
        geom2d_point_t* p1,
        geom2d_point_t* p2) {
  return geom2d_toright(p1,p2);
}

static geom2d_point_t*
maxdist(actk_state_t* st, geom2d_line_t* line,
        geom2d_point_t* p1, geom2d_point_t* p2) {
  return geom2d_maxdist(line, p1, p2);
}

static int
isabove(actk_state_t* st, geom2d_line_t* line, modlist_cell_t* cell) {
  geom2d_point_t* p = (geom2d_point_t*) cell->head;
  int result = geom2d_isabove(line, p);
  
  logcatf(st, "(%.2f, %.2f) %s ((%.2f, %.2f), (%.2f, %.2f))",
          p->x, p->y,
          (result ? "is above" : "is _not_ above"),
          line->p1->x, line->p1->y,
          line->p2->x, line->p2->y
          );
  
  return result;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* Hidden prototypes for quickhull routines */

ACTK_FUN_SIG(quickhull_rec_1, "quickhull_rec_1", "",
             ACTK_FUN_ARG("cell_1",  modlist_cell_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("cell_2",  modlist_cell_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("list",    modlist_t*,      CONST_VAL),
             );

ACTK_FUN_SIG(quickhull_rec_2, "quickhull_rec_2", "",
             ACTK_FUN_ARG("lentest", int,             READ_LOC_OR_VAL),
             ACTK_FUN_ARG("cell_1",  modlist_cell_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("cell_2",  modlist_cell_t*, READ_LOC_OR_VAL),
             ACTK_FUN_ARG("list",    modlist_t*,      CONST_VAL),
             ACTK_FUN_ARG("line",    geom_line_t*,    CONST_VAL)
             );

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* Overview for Quickhull -- The algorithm is broken into three parts.
   The first part (see: quickhull_setup) finds the dividing-line which
   divides the upper and lower hulls.  Then the upper and lower hulls
   are found in the same way except that the dividing-line for the
   lower hull has it's end-points reversed [which is a trick that
   makes isabove(line,p) act like isbelow(line,p) for a given point p.
   In this way, the upper and lower hull routines are done with the
   same code].

   For both the upper and lower hulls we use the mutually-recursive
   steps (quickhull_rec_1 and quickhull_rec_2).  The first of these
   filters out points below the current line and tests the number of
   remaining points.  The second of these either terminates if no
   points are left or finds the point P furthest from the current line
   and recurses twice by using the end-points of the old line and the
   point P which is furthest from the old line.

   The resulting convex hull is a list of points [p0, ..., p1, ....,
   p0] where (p0, p1) is the line seperating the upper and lower hulls
   (recall this line is found in quickhull_setup).  The upper hull is
   made of the first portion of this list, while the lower hull is
   made from the second portion.  For each pair of recursive steps
   (_rec_1 and _rec_2) a new point is added to the hull, any
   inner-points are filtered away, and the procedure continues twice
   for each sub-sequence of the yet-to-be-found hull.  Note again that
   the points on the hull are stored in list cells.  At each base-case
   (no points are left to be considered) the two currently held points
   on the hull have their list cells joined.  The whole process is
   over when every pair of consecutive hull-points are joined.
*/

ACTK_FUN(quickhull, quickhull_setup_, 0);

void quickhull_setup_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_t*     list = actk_fun_arg(st);
  actk_modref_t* dest = actk_fun_arg(st);
  
  actk_modref_t* minx = actk_modref(st);
  actk_modref_t* maxx = actk_modref(st);

  /* Point furtherest left */
  actk_call(st, modlist_combine,
            actk_pass(list),
            actk_pass(minx),
            actk_pass(toleft),
            actk_pass(NULL)
            );

  /* Point furtherest right */
  actk_call(st, modlist_combine,
            actk_pass(list),
            actk_pass(maxx),
            actk_pass(toright),
            actk_pass(NULL)
            );

  /* The upper hull goes from cell c1 to cell c2, the lower hull goes
     from cell c2 to cell c3.  Therefore, the point with minimum x
     coordinate goes in c1 and c3, while the point with maximum
     coordinate goes in c2.  The final list will begin with c1 and end
     with c3 (which are the same point) and hold the entire hull (both
     lower and upper hulls), where the upper and lower hulls share
     cell c2. */
  actk_modref_t* c1 = actk_modref(st);
  actk_modref_t* c2 = actk_modref(st);
  actk_modref_t* c3 = actk_modref(st);

  actk_call(st, modlist_cons, actk_read(minx), actk_pass(c1));
  actk_call(st, modlist_cons, actk_read(maxx), actk_pass(c2));
  actk_call(st, modlist_one,  actk_read(minx), actk_pass(c3));
  
  /* Upper hull : fills in between c1 and c2 */
  actk_call(st, quickhull_rec_1,
            actk_read(c1),
            actk_read(c2),
            actk_pass(list)
            );

  /* Lower hull : fills in between c2 and c3 */
  actk_call(st, quickhull_rec_1,
            actk_read(c2),
            actk_read(c3),
            actk_pass(list)
            );

  actk_call(st, modutil_copy,
            actk_read(c1),
            actk_pass(dest),
            );
  
  actk_fun_end(st);
}

ACTK_FUN(quickhull_rec_1, quickhull_rec_1_, 0);

/* Recursive Step 1: filter-out all those points that are "below" the
   given line.  The line is given as two list cells which hold the
   points defining the line. */
void quickhull_rec_1_(actk_state_t* st) {
  actk_fun_begin(st);
  modlist_cell_t* cell_1 = actk_fun_arg(st); /* TODO: don't read */
  modlist_cell_t* cell_2 = actk_fun_arg(st); /* TODO: don't read */
  modlist_t*        list = actk_fun_arg(st);

  geom2d_line_t* line = actk_alloc(st, sizeof(geom2d_line_t),
                                   cell_1->head, cell_2->head);
  
  geom2d_lineinit (line,
                   (geom2d_point_t*) cell_1->head,
                   (geom2d_point_t*) cell_2->head);  
  
  actk_modref_t* lentest = actk_modref(st, line);
  actk_modref_t* filterd = actk_modref(st, line);
  
  actk_call(st, modlist_filter /* ALTERNATE: modlist_filter[_sparse] */,
            actk_rkey(line),
            actk_read(list),
            actk_pass(filterd),
            actk_pass(isabove),
            actk_pass(line)
            );

  actk_call(st, modlist_testlen,
            actk_read(filterd),
            actk_pass(0),
            actk_pass(lentest)
            );

  actk_call(st, quickhull_rec_2,
            actk_read(lentest),
            actk_pass(cell_1),
            actk_pass(cell_2),
            actk_pass(filterd),
            actk_pass(line)
            );
  
  actk_fun_end(st);
}

ACTK_FUN(quickhull_rec_2, quickhull_rec_2_, 0);

/* Recursive Step 2: assume points are filtered, either terminate if
   there are no points left, or find the point furthest from the given
   line.  Then define two new lines between the previous end-points
   and the furthest point.  For each of these two new lines, recur
   with step 1. */
void quickhull_rec_2_(actk_state_t* st) {
  actk_fun_begin(st);
  int            lentest = (int) actk_fun_arg(st);
  modlist_cell_t* cell_1 = actk_fun_arg(st);
  modlist_cell_t* cell_2 = actk_fun_arg(st);
  modlist_t*        list = actk_fun_arg(st);
  geom2d_line_t*    line = actk_fun_arg(st);

  /* Pre-condition: all the points in the given list are "above" the
   * given line.  Moreover, the given line's end-points are in cell_1
   * and cell_2, respectively.
   */
  
  if(lentest) {
    /* The point-list is empty, so connect cells we have consecutively
       (since no other points exist to go between them) */
    actk_write(st, cell_1->tail, (actk_val_t) cell_2);
  }
  else {
    actk_modref_t* furthest_point = actk_modref(st);
    actk_modref_t* cell_mid = actk_modref(st);
    
    /* The point-list has at least one point (which we know is above
       the given line); find which is furthest from the line. */
    actk_call(st, modlist_combine,
              actk_pass(list),
              actk_pass(furthest_point),
              actk_pass(maxdist),
              actk_pass(line)
              );

    /* Populate the cell for the mid-point */
    actk_call(st, modlist_cons,
              actk_read(furthest_point),
              actk_pass(cell_mid)
              );

    /* The hull from cell-1 to mid-point */
    actk_call(st, quickhull_rec_1,
              actk_pass(cell_1),
              actk_read(cell_mid),
              actk_pass(list)
              );

    /* The hull from mid-point to cell-2 */
    actk_call(st, quickhull_rec_1,
              actk_read(cell_mid),
              actk_pass(cell_2),
              actk_pass(list)
              );
  }
  actk_fun_end(st);  
}

