#include <stdlib.h>

#include "actk.h"
#include "../core/actk_debug.h"

#include "test_utils.h"
#include "test_params.h"
#include "test_state.h"
#include "test_modlist.h"
#include "test_apps.h"
#include "test_signalhandler.h"


static int change_limit_reached(test_params_t* test_params, test_state_t* test_st) {
  return (test_params->change_limit >= 0 &&
          test_params->change_limit <= test_st->change_count);
}

static
void test_do_1_change(actk_state_t* st,
                      test_params_t* test_params,
                      test_state_t* test_st,
                      test_change_t change) {
  actk_stats_t stats_temp;
  const char* change_name = test_change_name(change);

  if(change_limit_reached(test_params, test_st))
    return;
  
  switch(change) {
  case TEST_CHANGE_REPLACE:
    test_params->app->hooks->do_replace(st, test_params, test_st);
    break;
  case TEST_CHANGE_INSERT:
    test_params->app->hooks->do_insert(st, test_params, test_st);
    break;
  case TEST_CHANGE_REMOVE:
    test_params->app->hooks->do_remove(st, test_params, test_st);
    break;
  default:
    fprintf(stderr, "change-type not recognized: %d\n", change);
    exit(-1);
  }

  test_st->change_count ++;
  
  if(st->flags & ACTK_PRINT_PROP_STATS)
    fprintf(stderr, "%s @ %d:\n", change_name, test_st->change_index);

  /* print-in/out: print output */
  if(test_flagval(test_params->print_inout)) {
    mesg_begin("%3d %*s: ", test_st->change_index, LABEL_WIDTH, "changed-input");
    test_params->app->hooks->print_input(st, test_params, test_st, stderr);
    mesg_end;
  }

  /* stats: save stats beforehand. */
  if(test_flagval(test_params->write_stats))
    actk__stats_save(&stats_temp, st->stats);

  /* Change-propagation */
  actk_propagate(st);

  /* stats: compute and record the stats deltas */
  if(test_flagval(test_params->write_stats)) {
    actk_stats_t stats_net;    
    actk__stats_diff(&stats_net, st->stats, &stats_temp);
    test_stat_file_entry(st->stats, test_st->change_index, strapp2(change_name, "_1"), test_st->file_stats_cum);
    test_stat_file_entry(&stats_net, test_st->change_index, strapp2(change_name, "_1"), test_st->file_stats);
  }

  /* Free all the garbage */
  actk_freeall(st);

  /* stats: compute and record the stats deltas */
  if(test_flagval(test_params->write_stats)) {
    actk_stats_t stats_net;
    actk__stats_diff(&stats_net, st->stats, &stats_temp);
    test_stat_file_entry(st->stats, test_st->change_index, strapp2(change_name, "_2"), test_st->file_stats_cum);
    test_stat_file_entry(&stats_net, test_st->change_index, strapp2(change_name, "_2"), test_st->file_stats);
  }

  /* print-in/out: print output */
  if(test_flagval(test_params->print_inout)) {
    mesg_begin("%3d %*s: ", test_st->change_index, LABEL_WIDTH, "changed-output");
    test_params->app->hooks->print_output(st, test_params, test_st, stderr);
    mesg_end;
  }
}

void test_do(actk_state_t* st, test_params_t* test_params) {
  test_state_t test_st_; test_state_t* test_st = &test_st_;

  /* If no app has been specified, we are done. */
  if(!test_params->app) {
    fprintf(stderr, "No app given, giving up.\n");
    exit(0);
  }
  
  /* Initialize test state */
  test_state_init(st, test_params, test_st);
  
  /* Generate Input */
  test_params->app->hooks->do_gen_input(st, test_params, test_st);

  /* print-input/output: */
  if(test_flagval(test_params->print_inout)) {
    mesg_begin("%3s %*s: ", "---", LABEL_WIDTH, "starting list");
    test_params->app->hooks->print_input(st, test_params, test_st, stderr);
    mesg_end;
  }
  
  /* initialize: initial-run stats: */ {
    if(test_flagval(test_params->write_stats)) {
      stats_file_preamble(st->stats, test_st->file_stats);
      stats_file_preamble(st->stats, test_st->file_stats_cum);
    }
    actk__stats_save(&test_st->stats_begin_initial, st->stats);
  }
  
  /* initialize: initial-run time: */ {
    time_save(&test_st->time_begin_initial);
  }
                  
  /* Do initial Run */
  test_params->app->hooks->do_initial_run(st, test_params, test_st);
  fprintf(stderr, "initial-run complete.\n");

  /* initial-run time: */  {
    time_save(&test_st->time_end_initial);
    time_diff(&test_st->time_delta_initial,
              &test_st->time_end_initial, &test_st->time_begin_initial);
    printf("initial-time = %g\n", time_as_double(&test_st->time_delta_initial));
  }
  
  /* initial-run stats: */ {
    if(test_flagval(test_params->write_stats)) {
      test_stat_file_entry(st->stats, 0, "initial", test_st->file_stats);
      test_stat_file_entry(st->stats, 0, "initial", test_st->file_stats_cum);
    }
    actk__stats_diff(&test_st->stats_delta_initial,
                     st->stats, &test_st->stats_begin_initial);
  }

  /* print-input/output: */
  if(test_flagval(test_params->print_inout)) {
    mesg_begin("%3s %*s: ", "---", LABEL_WIDTH, "initial-output:");
    test_params->app->hooks->print_output(st, test_params, test_st, stderr);
    mesg_end;
  }
  
  if(test_params->change_skip > 0) {

    /* change-loop stats/time: */
    time_save(&test_st->time_begin_changes);
    actk__stats_save(&test_st->stats_begin_changes, st->stats);

    /* Change/Prop loop */
    {int j; for(j = 0; j < test_params->change_loops; j++) {
      test_st->change_modref = test_st->input[0];
      
      { int i;
      for(i = test_params->change_skip / 2;
          (i < test_params->input_size)
            && (!change_limit_reached(test_params, test_st));
          i+= test_params->change_skip)
      {        
        test_st->change_index = i;
        test_st->change_index_delta = min(i, test_params->change_skip);
        test_params->app->hooks->do_prechange(st, test_params, test_st);
                
        if(test_flagval(test_params->change_do_replace)) 
          test_do_1_change(st, test_params, test_st, TEST_CHANGE_REPLACE);
          
        if(test_flagval(test_params->change_do_ins_rem)) {
          test_do_1_change(st, test_params, test_st, TEST_CHANGE_INSERT);
          test_do_1_change(st, test_params, test_st, TEST_CHANGE_REMOVE);
        }        
      }}
    }}

    /* change-loop stats/time: */ {    
      time_save(&test_st->time_end_changes);

      time_diff(&test_st->time_delta_changes,
                &test_st->time_end_changes,
                &test_st->time_begin_changes);

      printf("sum-update-time = %g\n",
             time_as_double(&test_st->time_delta_changes));
      
      printf("ave-update-time = %g\n",
             time_as_double(&test_st->time_delta_changes) /
             (double) (test_st->change_count) );

      actk__stats_diff(&test_st->stats_delta_changes,
                       st->stats,
                       &test_st->stats_begin_changes);

      printf("ave-closures-initialized = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_CLOSURE_INIT].val /
             (double) (test_st->change_count) );

      printf("ave-closures-wokeup = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_PROP_QUEUEPOP].val /
             (double) (test_st->change_count) );

      printf("ave-closures-entered = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_CLOSURE_ENTER].val /
             (double) (test_st->change_count) );

      printf("ave-closures-invalidated = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_CLOSURE_INVALIDATE].val /
             (double) (test_st->change_count) );

      printf("ave-closure-cost = %g\n",
             (double) (test_st->stats_delta_changes.entries[ACTK_STATS_CLOSURE_ENTER].val +
                       test_st->stats_delta_changes.entries[ACTK_STATS_CLOSURE_INVALIDATE].val) /
             (double) (test_st->change_count) );

      printf("ave-prop-nop = %g\n",
             (double) (test_st->stats_delta_changes.entries[ACTK_STATS_PROP_NOP].val) /
             (double) (test_st->change_count) );
      
      printf("ave-alloc-fresh = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_ALLOC_FRESH].val /
             (double) test_st->change_count);

      printf("ave-alloc-reuse = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_ALLOC_REUSE].val /
             (double) test_st->change_count);

      printf("ave-alloc-freed = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_ALLOC_FREED].val /
             (double) test_st->change_count);
      
      printf("ave-gc-cost = %g\n",
             (double) test_st->stats_delta_changes.entries[ACTK_STATS_ALLOC_FREED].val /
             (double) test_st->change_count);

      print_size(stdout, "user-malloc", st->stats->entries[ACTK_STATS_USER_MALLOC].val);
      print_size(stdout, "user-free  ", st->stats->entries[ACTK_STATS_USER_FREE].val);
      print_size(stdout, "user-live  ", (st->stats->entries[ACTK_STATS_USER_MALLOC].val -
                                         st->stats->entries[ACTK_STATS_USER_FREE].val));
      print_size(stdout, "total-malloc", st->stats->entries[ACTK_STATS_TOTAL_MALLOC].val);
      print_size(stdout, "total-free  ", st->stats->entries[ACTK_STATS_TOTAL_FREE].val);
      print_size(stdout, "total-live  ", (st->stats->entries[ACTK_STATS_TOTAL_MALLOC].val -
                                          st->stats->entries[ACTK_STATS_TOTAL_FREE].val));
      print_size(stdout, "total-maxlive", st->stats->entries[ACTK_STATS_TOTAL_MAXLIVE].val);
      
      /*
        printf("heap-malloc-total = %llu (%s)\n", st->stats->entries[ACTK_STATS_HEAP_MALLOC].val);
        printf("heap-free-total   = %llu (%s)\n", st->stats->entries[ACTK_STATS_HEAP_FREE].val);
        printf("heap-bytes-total  = %llu (%s)\n", (st->stats->entries[ACTK_STATS_HEAP_MALLOC].val -
                                                 st->stats->entries[ACTK_STATS_HEAP_FREE].val));
      */
    }
  }

  /* timestamps: if they were requested, write them out */
  if(test_flagval(test_params->write_timestamps)) {
    actk__print_interval(st, st->toplevel->interval,
                         "%s", test_params->app->name);
  }
  
  /* Do cleanup */
  test_state_cleanup(st, test_params, test_st);
  fprintf(stderr, "done.\n");
}


int main(int argc, char** argv) {
  actk_state_t st;
  test_params_t test_params;  
  
  test_params_default(&test_params);
  int argi = test_params_from_argv(&test_params, argc, argv);
  /* Give the remaining arguments to actk_init() */ {
    int actk_argc = argc - argi + 1;
    char** actk_argv = argv + argi - 1;
    actk_init(&st, actk_argc, actk_argv);
  }
  test_params_print(&test_params, stderr);
  test_install_signalhandler(&test_params);
  test_do(&st, &test_params);
  actk_cleanup(&st);
  return 0;
}
