#include "exprtree.h"

#include "test_apps.h"
#include "test_params.h"
#include "test_state.h"
#include "test_hooks.h"

/* Exprtree test implementation */
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void test_exprtree_do_gen_input(actk_state_t* st,
                               test_params_t* test_params,
                               test_state_t* test_st)
{    
  test_st->input[0] = exprtree_random(st, test_params->input_size);
}

void test_exprtree_print_input(actk_state_t* st,
                              test_params_t* test_params,
                              test_state_t* test_st, FILE* file) {
  exprtree_print(st, file, test_st->input[0]);
}

void test_exprtree_print_output(actk_state_t* st,
                               test_params_t* test_params,
                               test_state_t* test_st, FILE* file) {
  exprtree_print(st, file, test_st->output[0]);
}

void test_exprtree_do_initial_run(actk_state_t* st,
                                 test_params_t* test_params,
                                 test_state_t* test_st) {
  actk_call(st, exprtree_eval,
            actk_read(test_st->input[0]),
            actk_pass(test_st->output[0]) );
}

#define MAX_STACK 128
void test_exprtree_do_prechange(actk_state_t* st,
                               test_params_t* test_params,
                               test_state_t* test_st)
{
  /* Using a pair of (static) stacks to keep track of what we've done
     so far, we move depth-first through the expression tree */
  static exprtree_t* exprtree_stack[MAX_STACK];
  static int childnum_stack[MAX_STACK];
  static int stack_top = -1;

  while(1) {
    if(stack_top < 0) {
      stack_top = 0;
      exprtree_stack[stack_top] = test_st->change_modref = test_st->input[0];
      childnum_stack[stack_top] = 0;
      break;
    }
    else {
      /* Figure out what to do next */
      exprtree_t* exprtree = exprtree_stack[stack_top];
      exprtree_node_t* node = (exprtree_node_t*) actk_deref(st, exprtree);
      int nchildren = exprtree_nchildren(st, node);
      
      /* Either move to the next child, or pop the stack and try again */
      if(childnum_stack[stack_top] < nchildren) {
        /* Go to the next child */
        int next_child = childnum_stack[stack_top] ++;
        
        /* Push the next child onto the stack */
        stack_top += 1;
        exprtree_stack[stack_top] = node->children[next_child];
        childnum_stack[stack_top] = 0;
        
        /* The modref to change is this child's */
        test_st->change_modref = exprtree_stack[stack_top];
        break;
      }
      else {
        stack_top --;
        continue;
      }
    }
  }
}


void test_exprtree_do_insert(actk_state_t* st,
                            test_params_t* test_params,
                            test_state_t* test_st)
{
  exprtree_node_t* new_node = exprtree_node_empty(st, 2);
  exprtree_node_t* old_node = (exprtree_node_t*)
    actk_deref(st, test_st->change_modref);
  exprtree_op_t new_op = (rand() % EXPRTREE_OP_2_COUNT) + EXPRTREE_OP_2 + 1;
  
  new_node->tag = EXPRTREE_OP;
  new_node->u.op = new_op;
    
  actk_write(st, new_node->children[0], (actk_val_t) old_node);
  exprtree_random_rec(st, 1, new_node->children[1]);
  actk_write(st, test_st->change_modref, (actk_val_t) new_node);
}

void test_exprtree_do_remove(actk_state_t* st,
                            test_params_t* test_params,
                            test_state_t* test_st)
{
  exprtree_node_t* new_node = (exprtree_node_t*)
    actk_deref(st, test_st->change_modref);
  exprtree_node_t* old_node = (exprtree_node_t*)
    actk_deref(st, new_node->children[0]);
  exprtree_node_t* new_leaf = (exprtree_node_t*)
    actk_deref(st, new_node->children[1]);  
  
  actk_write(st, test_st->change_modref, (actk_val_t) old_node);

  /* Identify the garbage: */
  actk_garbage(st, new_node);
  actk_garbage(st, new_node->children[0]);
  actk_garbage(st, new_node->children[1]);
  actk_garbage(st, new_leaf);
}

void test_exprtree_do_replace(actk_state_t* st,
                             test_params_t* test_params,
                             test_state_t* test_st)
{
  /* Replacement is not supported. */
  abort();
}

test_hooks_t test_hooks_exprtree = {
  test_exprtree_do_gen_input,
  test_exprtree_do_initial_run,
  test_exprtree_do_prechange,
  test_exprtree_do_replace,
  test_exprtree_do_insert,
  test_exprtree_do_remove,  
  test_exprtree_print_input,
  test_exprtree_print_output,
};
