open Util
open Printf

module Int = struct 
  type t = int 
  let compare a b = (b - a) 
end

module Table_map = Map.Make(Int)

type table_meta = { table_name        : string;
                    table_title       : string;
                    table_algo_name   : string;
                    table_num_samples : int;
                    table_checkptfile : string;
                    table_datafile    : string;
                    table_xlabel      : string;
                    table_ylabel      : string;                    
                  }

type table_data = { table_dom : int list;
                    table_map : Entry.t Table_map.t
                  }

type table = table_meta * table_data


let save table =
  let (meta, data) = table in
  let outch = open_out meta.table_checkptfile in
  let _ = Marshal.to_channel outch table [] in
  (close_out outch)

let load checkptfile =
  let inch = open_in checkptfile in
  let table = Marshal.from_channel inch in
  (close_in inch; table)

let fresh algoname dom num_samples =
  ({ table_name        = algoname;
     table_title       = algoname;
     table_algo_name   = algoname;
     table_checkptfile = (sprintf "%s.checkpt" algoname);
     table_datafile    = (sprintf "%s.data" algoname);
     table_num_samples = num_samples;
     table_xlabel      = "input size";
     table_ylabel      = "time";
   },
   { 
     table_dom = dom;
     table_map = Table_map.empty
   })

let resume table =
  let (meta, {table_dom=dom; table_map=map}) = table 
  in  
  let table_from_map map =
    (meta, {table_dom=dom; table_map=map})
  in
  let rec loop dom map =
    match dom with
        [] -> map
      | x::dom -> 
          let entry = 
            if Table_map.mem x map then 
              Table_map.find x map
            else
              Entry.fresh {Entry.input_size  = x; 
                           Entry.algo_name   = meta.table_algo_name;
                           Entry.num_samples = meta.table_num_samples
                          }
          in
          try
            let entry = Entry.resume entry in
            let map = Table_map.add x entry map in
            (loop dom map)
          with
              Sample.Time_out -> 
                (Printf.eprintf "%s: Timed out for entry at %d\n" myname x ;
                 loop dom map)

            | Entry.User_quit entry -> 
                (save (table_from_map (Table_map.add x entry map)); 
                 exit(0))
  in
  table_from_map(loop dom map)


let gnuplot table outch =
  let (meta, data) = table in
  let datach = open_out (sprintf "%s.data" meta.table_name) in
  let rec loop dom =
    match dom with
        [] -> close_out datach
      | x::dom -> 
          try
            let (entry_in, entry_out) = Table_map.find x data.table_map in
            let results = entry_out.Entry.sample_ave in 
            output_string datach 
              (sprintf "%d %f %f %f %f %f %f %f %f %f %f %f %f\n" x
                 (* Essential stuff *)
                 (Sample.get "static-time" results)
                 (Sample.get "initial-time" results)
                 (Sample.get "ave-update-time" results)
                 (Sample.get "total-maxlive" results)
                 (* Other stuff *)
                 (Sample.get "total-malloc" results)
                 (Sample.get "total-free" results)
                 (Sample.get "total-live" results)
                 (Sample.get "user-malloc" results)
                 (Sample.get "user-free" results)
                 (Sample.get "user-live" results)
                 (Sample.get "ave-closure-cost" results)
                 (Sample.get "ave-gc-cost" results)
              ) ;
            loop dom
          with
              Not_found -> 
                output_string datach (sprintf "# %d --- missing\n" x) ;
                loop dom
  in
  (loop data.table_dom)

let run table =
  try
    let (meta, data) as table = resume table in
    let _ = gnuplot table (open_out meta.table_datafile) in
    let _ = (save table) in
    table
  with
      User_quit -> (save table; exit(0))
