open Printf
exception User_quit

let myname = "testperf"

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* Utility functions *)
 
(* Get the last line of a file *)
let line_last path = 
  let file_in = open_in path in
  let last_line = ref "" in
  try 
    while true do
      (last_line := input_line file_in)
    done ; ""
  with
      End_of_file -> !last_line        

(* convert a string of floats seperated by whitespace to a list of floats *)
let line_to_floats line =
  let whitespace = Str.regexp "[ \t\n\r]+" in
  (List.map float_of_string (Str.split whitespace line))

(* for values that represent numbers of bytes *)
let format_float_space f =
  let megabyte = float_of_int (1 lsl 20) in
  (sprintf "%.1f" (f /. megabyte))

(* for values roughly between 1 and 100 *)
let format_float_norm f =
  (sprintf "%.1f" f)

(* to get from the printf sci-notation to a LaTeX form. *)
let format_float_sci f expsign include_mantissa =
  let pattern = Str.regexp ("\\([0-9]+\\)\\.\\([0-9]+\\)e\\" ^ expsign ^ "0+\\([0-9]+\\)") in
  let expsign = match expsign with 
      "+" -> ""
    | "-" -> "-"
    | _ -> invalid_arg expsign
  in
  let replacement = 
    if include_mantissa 
    then "$\\1.\\2 \\times 10 ^ {" ^ expsign ^ " \\3}$"
    else "$10 ^ {" ^ expsign ^ " \\3}$"
  in
  let ugly = sprintf "%.1e" f in
  (Str.replace_first pattern replacement ugly)

(* for values >> 100 *)
let format_float_large f = (format_float_sci f "+" true)

(* for values << 1 *)
let format_float_small f = (format_float_sci f "-" true)

(* for powers of ten *)
let format_float_pow10 f = (format_float_sci f "+" false)

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* ML numbers *)

let mldatadir = "./data_sap-toplas/"

let mlappname appname =
  match appname with
      "bst" -> None
    | "diameter" -> Some "diameter"
    | "distance" -> None
    | "exprtree" -> None
    | "filter" -> Some "filter"
    | "map" -> Some "map-2Million"
    | "mergesort" -> Some "msort-string"
    | "minimum" -> Some "minimum"
    | "quickhull" -> Some "quick-hull-300K"
    | "quicksort" -> Some "qsort-150K"
    | "reverse" -> Some "reverse"
    | "split" -> Some "split"
    | "sum" -> Some "fold-2Million"
    | "traverse" -> None
    | _ -> invalid_arg appname
        
let getmldata mlappname inputsize =

  let path1 = mldatadir ^ "/" ^ mlappname ^ "/InitialRun/app.dat" in
  let path2 = mldatadir ^ "/" ^ mlappname ^ "/InitialRun/full.dat" in
  let path3 = mldatadir ^ "/" ^ mlappname ^ "/AllInsertDelete/app.dat" in
  let path4 = mldatadir ^ "/" ^ mlappname ^ "/AllInsertDelete/full.dat" in
  let path5 = mldatadir ^ "/" ^ mlappname ^ "/AllInsertDelete/maxlive.dat" in

  let paths = [path1; path2; path3; path4; path5] in
  let last_lines = List.map line_last paths in
  let floatss = List.map line_to_floats last_lines in
  let floats = List.flatten floatss in
  
  match floats with
      i1::ir_app::
        i2::ir_full::
        i3::cp_app::
        i4::cp_full::
        maxlive::_ -> 
          (*
          if not (i1 = i2 && i2 = i3 && i3 = i4) then
            (eprintf "sizes don't match: %f, %f, %f, %f\n" i1 i2 i3 i4 ;
             invalid_arg mlappname)
          else if inputsize <> i1 then
            (eprintf "sizes don't match: %f, %f\n" inputsize i1 ;
             invalid_arg mlappname)              
          else
          *)
          (format_float_norm ir_app, 
           format_float_norm ir_full, 
           format_float_small cp_app,
           format_float_small cp_full,
           format_float_space maxlive)
    
    | _ -> invalid_arg "the list is the wrong length"
